/*
 * Copyright (C) 2003 Kristoffer Vinther.
 * For contact information, please visit
 * http://www.brics.dk/~kv/thesis/.
 * If you use this code in commercial applications, please notify me.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * Free Software Foundation, Inc.
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <exception>
#include <iostream>

#include "../sorted_stream.h"
#include "../elements.h"
#include "../stopwatch.h"
#include "../../SortGen/random.h"
#include "../../Sort/Sort.h"
#include "../../Sort/funnel/stack_allocator.h"
#include "../../Sort/funnel/merge_tree.h"
#include "../../Sort/funnel/merge_tree_types.h"

float d;
int alpha;

template<funnel::order_t order>
class my_splitter
{
public:
	static inline funnel::order_t prefered_order_output(_uint64_t N)
		{ return static_cast<funnel::order_t>(floor(pow(static_cast<double>(N), 1.0/d)/alpha+.5)); }
	static inline funnel::order_t prefered_order_input(_uint64_t N)
		{ return static_cast<funnel::order_t>(floor(pow(static_cast<double>(N), 1.0/d)/alpha+.5)); }
	static inline uint64_t switch_to_cache_aware()
		{ return static_cast<funnel::order_t>(floor(pow(alpha*(order+0.5f), d)+.5))+1; }
	static inline funnel::height_t split(funnel::height_t h)
		{ return h/2; }
	static inline size_t out_buffer_size(funnel::order_t k)
		{ return alpha*static_cast<size_t>(pow(static_cast<float>(k),d)); }
};

void run(size_t N)
{
	using namespace funnel;
	std::cout << "\t<run buildtime=\"" << sys_names::buildtime() << "\" compiler=\"" << sys_names::compiler() << "\" execstatus=\"0\">" << std::endl;
	std::cout << "\t\t<size>" << alpha << "</size>" << std::endl;
	std::cout << "\t\t<algorithm>Buffer size " << d << ',' << N << "</algorithm>" << std::endl;
	element *input = NULL;
	try
	{
		stopwatch w;
		miss_counter mc;
		element *input = new element[N];
		random_source rs;
		element chksum;
		rs >> chksum;
		*input = chksum;
		for( element *p=input+1; p!=input+N; ++p )
			rs >> *p, chksum += *p;
		sorted_stream<element> output(N);
		mc.start();
		w.start();
		iosort::merge_sort<merge_tree_types<element,2,buffer<element*>,simple_merger<2,buffer<element*>,element>,my_splitter<2> >::rec_stack_mveb, my_splitter<2> >(input, input+N, output.begin());
		w.stop();
		mc.stop();
		if( output.checksum() != chksum )
			std::cout << "\t\t<error>Checksum error: " << output.checksum() << " != " << chksum << "</error>" << std::endl;
		delete[] input; input = NULL;
		std::cout << w.xmltag(2) << mc.xmltag() << std::endl;
	}
	catch( std::exception& ex )
	{
		delete[] input;
		std::cout << "\t\t<error>Caught std::exception: " << ex.what() << "</error>" << std::endl;
	}
	std::cout << "\t</run>" << std::endl;
}

int main(int argc, char *argv[])
{
	local_time t;
	std::cout << "<?xml version=\"1.0\" encoding=\"UTF-8\"?>" << std::endl;
	std::cout << "<benchmark date=\"" << t << "\" os=\"" << sys_names::os() << "\" version=\"" << sys_names::release() << "\" host=\"" << sys_names::host() << "\">" << std::endl;
	int A[9] = { 1, 2, 4, 8, 12, 16, 24, 32, 40 };
	for( d=1.5f; d<=3.0f; d+=0.5f )
	{
		alpha = A[0];
		for( int i=0; i<9; alpha=A[++i] )
			if( sizeof(element) > 8 )
			{
				run(6700000);
				run(3700000);
				run(5100000);
			}
			else
			{
				run(26000000);
				run(16000000);
				run(22000000);
			}
	}
	std::cout << "</benchmark>" << std::endl;
}
