/*
 * Copyright (C) 2003 Kristoffer Vinther.
 * For contact information, please visit
 * http://www.brics.dk/~kv/thesis/.
 * If you use this code in commercial applications, please notify me.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * Free Software Foundation, Inc.
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <exception>
#include <iostream>

#include "../sorted_stream.h"
#include "../elements.h"
#include "../stopwatch.h"
#include "../../SortGen/random.h"
#define UNLIMITED_MERGER_OUTPUT
#include "../../Sort/Sort.h"
#include "../../Sort/funnel/stack_allocator.h"
#include "../../Sort/funnel/merge_tree.h"
#include "../../Sort/funnel/merge_tree_types.h"

float d;
int alpha;

template<funnel::order_t order>
class my_splitter
{
public:
	static inline funnel::order_t prefered_order_output(_uint64_t N)
		{ return static_cast<funnel::order_t>(floor(pow(static_cast<double>(N), 1.0/d)/alpha+.5)); }
	static inline funnel::order_t prefered_order_input(_uint64_t N)
		{ return static_cast<funnel::order_t>(floor(pow(static_cast<double>(N), 1.0/d)/alpha+.5)); }
	static inline uint64_t switch_to_cache_aware()
		{ return static_cast<funnel::order_t>(floor(pow(alpha*(order+0.5f), d)+.5))+1; }
	static inline funnel::height_t split(funnel::height_t h)
		{ return h/2; }
	static inline size_t out_buffer_size(funnel::order_t k)
		{ return alpha*static_cast<size_t>(pow(static_cast<float>(k),d)); }
};

namespace uncached
{

template<class Merger, class Splitter, class It, class OutIt, class T, class ItDiff>
void _merge_sort(It begin, It end, OutIt dest, typename Merger::allocator alloc)
{
	typedef typename Merger::stream Stream;
	size_t order = Splitter::prefered_order_output(end-begin);
	if( order < 2 )
		order = 2;
	ItDiff run_size = (end-begin)/order;
	size_t rem = static_cast<size_t>((end-begin) % order);
	run_size += rem/order;

	size_t big_runs = rem % order;
	It last = begin;
	size_t i;
	if( run_size > static_cast<ItDiff>(Splitter::switch_to_cache_aware()) )
	{
		if( big_runs )
			run_size++;
		T* tmp = alloc.allocate(static_cast<typename Merger::allocator::size_type>(run_size));
		if( big_runs )
		{
			_merge_sort<Merger,Splitter,It,T*,T,ItDiff>(last, last+run_size, tmp, alloc);
			for( i=1, last+=run_size; i!=big_runs; i++, last+=run_size )
				_merge_sort<Merger,Splitter,It,It,T,ItDiff>(last, last+run_size, last-run_size, alloc);
			run_size--;
			for( ; i!=order; i++, last+=run_size )
				_merge_sort<Merger,Splitter,It,It,T,ItDiff>(last, last+run_size, last-run_size-1, alloc);
			run_size++;
		}
		else
		{
			_merge_sort<Merger,Splitter,It,T*,T,ItDiff>(last, last+run_size, tmp, alloc);
			for( i=1, last+=run_size; i!=order; i++, last+=run_size )
				_merge_sort<Merger,Splitter,It,It,T,ItDiff>(last, last+run_size, last-run_size, alloc);
		}
		assert( last == end );
		std::copy(tmp, tmp+run_size, last-run_size);
		alloc.deallocate(tmp,static_cast<typename Merger::allocator::size_type>(run_size));
		Merger merger(order,alloc);
		if( big_runs )
		{
			for( i=0, last=begin; i!=big_runs-1; i++, last+=run_size )
				merger.add_stream(Stream(last, last+run_size));
			run_size--;
			for( ; i!=order-1; i++, last+=run_size )
				merger.add_stream(Stream(last, last+run_size));
			run_size++;
			merger.add_stream(Stream(last, last+run_size));
			last+=run_size;
		}
		else
			for( i=0, last=begin; i!=order; i++, last+=run_size )
				merger.add_stream(Stream(last, last+run_size));
		assert( last == end );
		merger(dest);
	}
	else
	{
		Merger merger(order, alloc);
		for( i=0; i!=order-big_runs; i++, last+=run_size )
		{
			iosort::base::__inplace_base_sort(last, last+run_size);
			merger.add_stream(Stream(last, last+run_size));
		}
		run_size++;
		for( ; i!=order; i++, last+=run_size )
		{
			iosort::base::__inplace_base_sort(last, last+run_size);
			merger.add_stream(Stream(last, last+run_size));
		}
		merger(dest);
	}
}
template<class Merger, class Splitter, class It, class OutIt, class T, class Diff>
inline void _merge_sort_a(It begin, It end, OutIt dest, T, Diff d, typename Merger::allocator& alloc)
{
	if( d > static_cast<Diff>(Splitter::switch_to_cache_aware()) )
		_merge_sort<Merger,Splitter,It,OutIt,T,Diff>(begin, end, dest, alloc);
	else
		iosort::base::__base_sort(begin, end, dest);
}
template<class Merger, class Splitter, class It, class OutIt>
inline void merge_sort(It begin, It end, OutIt dest)
{
	if( end-begin > static_cast<typename std::iterator_traits<It>::difference_type>(Splitter::switch_to_cache_aware()) )
	{
		iosort::allocator_factory<typename Merger::allocator> af;
		_merge_sort<Merger,Splitter,It,OutIt,typename std::iterator_traits<It>::value_type,typename std::iterator_traits<It>::difference_type>(begin, end, dest, af(iosort::tmp_space_needed<Merger,Splitter>(begin,end)));
	}
	else
		iosort::base::__base_sort(begin, end, dest);
}
template<class Merger, class Splitter, class It, class OutIt>
inline void merge_sort(It begin, It end, OutIt dest, typename Merger::allocator& alloc)
{
	_merge_sort_a<Merger,Splitter>(begin, end, dest, *begin, end-begin, alloc);
}

} // namespace uncached

template<class Merger>
void run(size_t N, const char *v)
{
	using namespace funnel;
	std::cout << "\t<run buildtime=\"" << sys_names::buildtime() << "\" compiler=\"" << sys_names::compiler() << "\" execstatus=\"0\">" << std::endl;
	std::cout << "\t\t<size>" << N << "</size>" << std::endl;
	std::cout << "\t\t<algorithm>Cached " << v << "</algorithm>" << std::endl;
	element *input = NULL;
	try
	{
		stopwatch w;
		miss_counter mc;
		element *input = new element[N];
		random_source rs;
		element chksum;
		rs >> chksum;
		*input = chksum;
		for( element *p=input+1; p!=input+N; ++p )
			rs >> *p, chksum += *p;
		sorted_stream<element> output(N);
		mc.start();
		w.start();
		iosort::merge_sort<Merger, my_splitter<2> >(input, input+N, output.begin());
		w.stop();
		mc.stop();
		if( output.checksum() != chksum )
			std::cout << "\t\t<error>Checksum error: " << output.checksum() << " != " << chksum << "</error>" << std::endl;
		delete[] input; input = NULL;
		std::cout << w.xmltag(2) << mc.xmltag() << std::endl;
	}
	catch( std::exception& ex )
	{
		delete[] input;
		std::cout << "\t\t<error>Caught std::exception: " << ex.what() << "</error>" << std::endl;
	}
	std::cout << "\t</run>" << std::endl;
}

template<class Merger>
void runuc(size_t N, const char *v)
{
	std::cout << "\t<run buildtime=\"" << sys_names::buildtime() << "\" compiler=\"" << sys_names::compiler() << "\" execstatus=\"0\">" << std::endl;
	std::cout << "\t\t<size>" << N << "</size>" << std::endl;
	std::cout << "\t\t<algorithm>Not cached " << v << "</algorithm>" << std::endl;
	element *input = NULL;
	try
	{
		stopwatch w;
		miss_counter mc;
		element *input = new element[N];
		random_source rs;
		element chksum;
		rs >> chksum;
		*input = chksum;
		for( element *p=input+1; p!=input+N; ++p )
			rs >> *p, chksum += *p;
		sorted_stream<element> output(N);
		mc.start();
		w.start();
		uncached::merge_sort<Merger, my_splitter<2> >(input, input+N, output.begin());
		w.stop();
		mc.stop();
		if( output.checksum() != chksum )
			std::cout << "\t\t<error>Checksum error: " << output.checksum() << " != " << chksum << "</error>" << std::endl;
		delete[] input; input = NULL;
		std::cout << w.xmltag(2) << mc.xmltag() << std::endl;
	}
	catch( std::exception& ex )
	{
		delete[] input;
		std::cout << "\t\t<error>Caught std::exception: " << ex.what() << "</error>" << std::endl;
	}
	std::cout << "\t</run>" << std::endl;
}

int main(int argc, char *argv[])
{
	using namespace funnel;
	local_time t;
	std::cout << "<?xml version=\"1.0\" encoding=\"UTF-8\"?>" << std::endl;
	std::cout << "<benchmark date=\"" << t << "\" os=\"" << sys_names::os() << "\" version=\"" << sys_names::release() << "\" host=\"" << sys_names::host() << "\">" << std::endl;
//	typedef merge_tree_types<element,4,buffer<element*>,four_merger<4,buffer<element*>,element>,my_splitter<4> >::rec_stack_mveb smerger;
//	typedef merge_tree_types<element,4,buffer<element*>,four_merger<4,buffer<element*>,element>,my_splitter<4> >::rec_heap_mveb hmerger;
	typedef merge_tree_types<element,2,buffer<element*>,simple_merger<2,buffer<element*>,element>,my_splitter<2> >::rec_stack_mveb smerger;
	typedef merge_tree_types<element,2,buffer<element*>,simple_merger<2,buffer<element*>,element>,my_splitter<2> >::rec_heap_mveb hmerger;
	alpha = 4; d = 2.5f;
	for( int k=4; k!=256; k*=2 )
	{
		if( sizeof(element) > 8 )
		{
			run<smerger>(120000*k,"stack");
			run<hmerger>(120000*k,"heap");
			runuc<smerger>(120000*k,"stack");
			runuc<hmerger>(120000*k,"heap");
		}
		else
		{
			run<smerger>(1050000*k,"stack");
			run<hmerger>(1050000*k,"heap");
			runuc<smerger>(1050000*k,"stack");
			runuc<hmerger>(1050000*k,"heap");
		}
	}
	std::cout << "</benchmark>" << std::endl;
}
