/*
 * Copyright (C) 2003 Kristoffer Vinther.
 * For contact information, please visit
 * http://www.brics.dk/~kv/thesis/.
 * If you use this code in commercial applications, please notify me.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * Free Software Foundation, Inc.
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <exception>
#include <iostream>

#include "../sorted_stream.h"
#include "../elements.h"
#include "../stopwatch.h"
#include "../../SortGen/random.h"
#include "../../Sort/Sort.h"
#include "../../Sort/funnel/stack_allocator.h"
#include "../../Sort/funnel/merge_tree.h"
#include "../../Sort/funnel/merge_tree_types.h"

float d;
int alpha;

template<funnel::order_t order>
class my_splitter
{
public:
	static inline funnel::order_t prefered_order_output(_uint64_t N)
		{ return (funnel::order_t)floor(pow((double)N, 1.0/d)/alpha+.5); }
	static inline funnel::order_t prefered_order_input(_uint64_t N)
		{ return (funnel::order_t)floor(pow((double)N, 1.0/d)/alpha+.5); }
	static inline uint64_t switch_to_cache_aware()
		{ return out_buffer_size(order+1); }
	static inline funnel::height_t split(funnel::height_t h)
			{ return h/2; }
	static inline size_t out_buffer_size(funnel::order_t k)
		{ return alpha*static_cast<size_t>(pow(static_cast<float>(k),d)); }
};

template<class Allocator>
class allocator_factory
{
public:
	inline Allocator operator()(typename Allocator::size_type size)
		{ return Allocator(); }
};
template<class T>
class allocator_factory<stack_allocator<T> >
{
public:
	inline stack_allocator<T> operator()(typename stack_allocator<T>::size_type size)
		{ return stack_allocator<T>(size); }
};

template<class It, class Diff>
class splitters
{
public:
	splitters(It b, It e, Diff order) : b(b), e(e), c(b)
	{
		run_size = (e-b)/order;
		Diff rem = (e-b)%order;
		run_size += rem/order;
		Diff big_runs = rem%order;
	}
	It operator*() const
		{ return c; }
	void operator++()
		{ }
private:
	It b, e, c;
	Diff run_size;
};

/*template<class Merger>
void run_merger(funnel::order_t k, const char* v)
{
	std::cout << "\t<run buildtime=\"" << sys_names::buildtime() << "\" compiler=\"" << sys_names::compiler() << "\" execstatus=\"0\">" << std::endl;
	std::cout << "\t\t<size>" << k << "</size>" << std::endl;
	std::cout << "\t\t<algorithm>" << v << "</algorithm>" << std::endl;
	typedef typename Merger::value_type T;
	typedef typename Merger::stream stream;
	try
	{
		stopwatch w;
		miss_counter mc;
		size_t N = k*k*k;
		k = Merger::splitter::prefered_order_output(N);
		size_t run_size = N/k;
		size_t rem = N%k;
		run_size += rem/k;
		size_t big_runs = rem%k;
		allocator_factory<typename Merger::allocator> af;
		typename Merger::allocator alloc = af(Merger::size_of(k));
		Merger merger(k,alloc);
		T *input = new T[N];
		sorted_stream<T> output(N);

		random_generator<T> rg;
		T chksum = rg();
		*input = chksum;
		for( T *p=input+1; p!=input+N; ++p )
			*p = rg(), chksum += *p;
		T *last = input;
		size_t i;
		run_size++;
		for( i=0; i!=big_runs; i++, last+=run_size )
			std::sort(last, last+run_size);
		run_size--;
		for( ; i!=k; i++, last+=run_size )
			std::sort(last, last+run_size);

		mc.start();
		w.start();
		for( size_t r=0; r<20000000/N; ++r )
		{
			if( big_runs )
			{
				run_size++;
				for( i=0, last=input; i!=big_runs; i++, last+=run_size )
					merger.add_stream(stream(last, last+run_size));
				run_size--;
				for( ; i!=k; i++, last+=run_size )
					merger.add_stream(stream(last, last+run_size));
				last+=run_size;
			}
			else
				for( i=0, last=input; i!=k; i++, last+=run_size )
					merger.add_stream(stream(last, last+run_size));
			merger(output.begin(), output.end());
			merger.reset();
		}
		w.stop();
		mc.stop();
		delete[] input;
		std::cout << w.xmltag(2) << mc.xmltag() << std::endl;
	}
	catch( std::exception& ex )
	{
		std::cerr << "Caught std::exception: " << ex.what() << std::endl;
	}
	std::cout << "\t</run>" << std::endl;
}
*/

template<class Merger>
void run_merger(funnel::order_t k, const char* v)
{
	std::cout << "\t<run buildtime=\"" << sys_names::buildtime() << "\" compiler=\"" << sys_names::compiler() << "\" execstatus=\"0\">" << std::endl;
	std::cout << "\t\t<size>" << k << "</size>" << std::endl;
	std::cout << "\t\t<algorithm>" << v << "</algorithm>" << std::endl;
	typedef typename Merger::value_type T;
	typedef typename Merger::stream stream;
	try
	{
		stopwatch w;
		miss_counter mc;
		allocator_factory<typename Merger::allocator> af;
		typename Merger::allocator alloc = af(Merger::size_of(k));
		Merger merger(k,alloc);
		T *input = new T[k*k*k];

		random_source rs;
		T chksum;
		rs >> chksum;
		*input = chksum;
		for( T *p=input+1; p!=input+k*k*k; ++p )
			rs >> *p, chksum += *p;
		for( T *p=input; p+k*k<=input+k*k*k; p+=k*k)
			std::sort(p, p+k*k);
		mc.start();
		w.start();
		for( int i=0; i<20000000/(k*k*k); ++i )
		{
			sorted_stream<T> output(k*k*k);
			for( T *p=input; p+k*k<=input+k*k*k; p+=k*k)
				merger.add_stream(stream(p,p+k*k));
			if( merger(output.begin(), output.end()) != output.end() )
				std::cout << "\t\t<error>Incomplete output!</error>" << std::endl;
			if( output.checksum() != chksum )
				std::cout << "\t\t<error>Checksum error: " << output.checksum() << " != " << chksum << "</error>" << std::endl;
			merger.reset();
		}
		w.stop();
		mc.stop();
		delete[] input;
		std::cout << w.xmltag(2) << mc.xmltag() << std::endl;
	}
	catch( std::exception& ex )
	{
		std::cout << "\t\t<error>Caught std::exception: " << ex.what() << "</error>" << std::endl;
	}
	std::cout << "\t</run>" << std::endl;
}

int main(int argc, char *argv[])
{
	local_time t;
	std::cout << "<?xml version=\"1.0\" encoding=\"UTF-8\"?>" << std::endl;
	std::cout << "<benchmark date=\"" << t << "\" os=\"" << sys_names::os() << "\" version=\"" << sys_names::release() << "\" host=\"" << sys_names::host() << "\">" << std::endl;
	using namespace funnel;
	for( funnel::order_t k=20; k<271; k+=5 )
	{
		alpha = 1; d = 3.0f;
		run_merger<merge_tree_types<element,2,buffer<element*>,simple_merger<2,buffer<element*>,element>,my_splitter<2> >::rec_heap_mveb>(k,"simple,1,3.0");
		run_merger<merge_tree_types<element,2,buffer<element*>,two_merger<2,buffer<element*>,element>,my_splitter<2> >::rec_heap_mveb>(k,"two,1,3.0");
		run_merger<merge_tree_types<element,2,buffer<element*>,hl_merger<2,buffer<element*>,element>,my_splitter<2> >::rec_heap_mveb>(k,"hl,1,3.0");
		run_merger<merge_tree_types<element,2,buffer<element*>,hyb0_merger<2,buffer<element*>,element>,my_splitter<2> >::rec_heap_mveb>(k,"hyb0,1,3.0");
		run_merger<merge_tree_types<element,2,buffer<element*>,hyb_merger<2,buffer<element*>,element>,my_splitter<2> >::rec_heap_mveb>(k,"hyb,1,3.0");
		run_merger<merge_tree_types<element,2,buffer<element*>,hyb3_merger<2,buffer<element*>,element>,my_splitter<2> >::rec_heap_mveb>(k,"hyb3,1,3.0");
		alpha = 4; d = 2.5f;
		run_merger<merge_tree_types<element,2,buffer<element*>,simple_merger<2,buffer<element*>,element>,my_splitter<2> >::rec_heap_mveb>(k,"simple,4,2.5");
		run_merger<merge_tree_types<element,2,buffer<element*>,two_merger<2,buffer<element*>,element>,my_splitter<2> >::rec_heap_mveb>(k,"two,4,2.5");
		run_merger<merge_tree_types<element,2,buffer<element*>,hl_merger<2,buffer<element*>,element>,my_splitter<2> >::rec_heap_mveb>(k,"hl,4,2.5");
		run_merger<merge_tree_types<element,2,buffer<element*>,hyb0_merger<2,buffer<element*>,element>,my_splitter<2> >::rec_heap_mveb>(k,"hyb0,4,2.5");
		run_merger<merge_tree_types<element,2,buffer<element*>,hyb_merger<2,buffer<element*>,element>,my_splitter<2> >::rec_heap_mveb>(k,"hyb,4,2.5");
		run_merger<merge_tree_types<element,2,buffer<element*>,hyb3_merger<2,buffer<element*>,element>,my_splitter<2> >::rec_heap_mveb>(k,"hyb3,4,2.5");
		alpha = 16; d = 1.5f;
		run_merger<merge_tree_types<element,2,buffer<element*>,simple_merger<2,buffer<element*>,element>,my_splitter<2> >::rec_heap_mveb>(k,"simple,16,1.5");
		run_merger<merge_tree_types<element,2,buffer<element*>,two_merger<2,buffer<element*>,element>,my_splitter<2> >::rec_heap_mveb>(k,"two,16,1.5");
		run_merger<merge_tree_types<element,2,buffer<element*>,hl_merger<2,buffer<element*>,element>,my_splitter<2> >::rec_heap_mveb>(k,"hl,16,1.5");
		run_merger<merge_tree_types<element,2,buffer<element*>,hyb0_merger<2,buffer<element*>,element>,my_splitter<2> >::rec_heap_mveb>(k,"hyb0,16,1.5");
		run_merger<merge_tree_types<element,2,buffer<element*>,hyb_merger<2,buffer<element*>,element>,my_splitter<2> >::rec_heap_mveb>(k,"hyb,16,1.5");
		run_merger<merge_tree_types<element,2,buffer<element*>,hyb3_merger<2,buffer<element*>,element>,my_splitter<2> >::rec_heap_mveb>(k,"hyb3,16,1.5");
	}
	std::cout << "</benchmark>" << std::endl;
}
