/*
 * Copyright (C) 2003 Kristoffer Vinther.
 * For contact information, please visit
 * http://www.brics.dk/~kv/thesis/.
 * If you use this code in commercial applications, please notify me.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * Free Software Foundation, Inc.
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "aarray.h"

const void* fetch_rq[1024];
const void* fetch_co[1024];
int rqc=0, coc=0;

#ifdef WIN32

__PrefetchThread::__PrefetchThread()
{
	m_Shared.m_hsEmpty = CreateSemaphoreA(NULL, 16, 16, NULL);
	m_Shared.m_hsFull = CreateSemaphoreA(NULL, 0, 16, NULL);
	m_hThread = CreateThread(NULL, 4096, __PrefetchThreadFunc, &m_Shared, 0, NULL);
	SetThreadPriority(m_hThread, THREAD_PRIORITY_ABOVE_NORMAL);
}

__PrefetchThread::~__PrefetchThread()
{
	Fetch(NULL);
	WaitForSingleObject(m_hThread, INFINITE);
	CloseHandle(m_hThread);
	CloseHandle(m_Shared.m_hsFull);
	CloseHandle(m_Shared.m_hsEmpty);
}

void __PrefetchThread::Fetch(const void* p)
{
	WaitForSingleObject(m_Shared.m_hsEmpty, INFINITE);
	m_Shared.m_pFetchAddr[m_Head] = p;
	m_Head = (m_Head+1) & 15;
	ReleaseSemaphore(m_Shared.m_hsFull, 1, NULL);
}

inline void LockMem(const void* start, const void* end)
{
	if( VirtualLock(const_cast<void*>(start), (SIZE_T)((char*)end-(char*)start)) == 0 )
	{
		SIZE_T MinWorkingSetSize, MaxWorkingSetSize;
		HANDLE hProc = GetCurrentProcess();
		GetProcessWorkingSetSize(hProc, &MinWorkingSetSize, &MaxWorkingSetSize);
		if( SetProcessWorkingSetSize(hProc, 2*MinWorkingSetSize, 2*MaxWorkingSetSize) == 0 )
		{
			std::cerr << "SetProcessWorkingSetSize unsuccesful" << std::endl;
			std::cerr << GetLastError() << std::endl;
		}
	}
	if( VirtualLock(const_cast<void*>(start), (SIZE_T)((char*)end-(char*)start)) == 0 )
	{
		std::cerr << "LockMem unsuccesful" << std::endl;
		std::cerr << GetLastError() << std::endl;
	}
}
inline void UnlockMem(const void* start, const void* end)
{
	if( VirtualUnlock(const_cast<void*>(start), (SIZE_T)((const char*)end-(const char*)start)) == 0 )
	{
		std::cerr << "UnlockMem unsuccesful" << std::endl;
		std::cerr << GetLastError() << std::endl;
	}
}

#else // WIN32

__PrefetchThread::__PrefetchThread()
{
	sem_init(&m_Shared.m_sEmpty, false, 16);
	sem_init(&m_Shared.m_sFull, false, 0);
	pthread_create(&m_Thread, NULL, __PrefetchThreadFunc, &m_Shared);
}

__PrefetchThread::~__PrefetchThread()
{
	void* ret;
	pthread_join(m_Thread, &ret);
	sem_destroy(&m_Shared.m_sFull);
	sem_destroy(&m_Shared.m_sEmpty);
/*	std::cout << "Pages requested: " << rqc << "; pages delivered: " << coc << std::endl;
	for( int i=0; i<rqc && i<coc; i++ )
		std::cout << fetch_rq[i] << '\t' << fetch_co[i] << std::endl;
*/}

void __PrefetchThread::Fetch(const void* p)
{
//	if( rqc < 1024 )
//		fetch_rq[rqc++] = p;
	sem_wait(&m_Shared.m_sEmpty);
	m_Shared.m_pFetchAddr[m_Head] = p;
	m_Head = (m_Head+1) & 15;
	sem_post(&m_Shared.m_sFull);
}

inline void LockMem(const void* start, const void* end)
{
	if( mlock(start, (char*)end-(char*)start) )
		std::clog << "LockMem unsuccesful" << std::endl;
}
inline void UnlockMem(const void* start, const void* end)
{
	if( munlock(start, (char*)end-(char*)start) )
		std::clog << "UnlockMem unsuccesful" << std::endl;
}

#endif // WIN32

#ifdef WIN32
DWORD WINAPI __PrefetchThreadFunc(LPVOID lpParameter)
#else // WIN32
void* __PrefetchThreadFunc(void* lpParameter)
#endif // WIN32
{
	__PrefetchThread::__PrefetchThreadData* pShared = (__PrefetchThread::__PrefetchThreadData*)lpParameter;
	unsigned char Tail = 0;
	char* pElmt;
	volatile char elmt;
	bool cont = true;
	while( cont )
	{
#ifdef WIN32
		WaitForSingleObject(pShared->m_hsFull, INFINITE);
#else // WIN32
		sem_wait(&pShared->m_sFull);
#endif // WIN32
		pElmt = (char*)pShared->m_pFetchAddr[Tail];
		Tail = (Tail+1) & 15;
#ifdef WIN32
		ReleaseSemaphore(pShared->m_hsEmpty, 1, NULL);
#else // WIN32
		sem_post(&pShared->m_sEmpty);
#endif // WIN32
		if( pElmt )
		{
			elmt = *pElmt;
//			if( coc < 1024 )
//				fetch_co[coc++] = pElmt;
		}
		else
		{
#ifdef WIN32
			ReleaseSemaphore(pShared->m_hsFull, 1, NULL);
#else // WIN32
			sem_post(&pShared->m_sFull);
#endif // WIN32
			cont = false;
		}
	}
	return 0;
}
