/*
 * Copyright (C) 2003 Kristoffer Vinther.
 * For contact information, please visit
 * http://www.brics.dk/~kv/thesis/.
 * If you use this code in commercial applications, please notify me.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * Free Software Foundation, Inc.
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "stopwatch.h"

#include <stdio.h>

const char* sys_names::compiler()
{
	static char cn[80];
#ifdef __IBMC__ // __IBMC__ = 430 for version 4.3 ...
	sprintf(cn, "IBM C %d.%d", __IBMC__/100, __IBMC__-(__IBMC__/100)*100);
#endif // __IBMCPP__

#ifdef __IBMCPP__ // __IBMCPP__ = 360 for version 3.6 ...
	sprintf(cn, "IBM C++ %d.%d", __IBMCPP__/100, __IBMCPP__-(__IBMCPP__/100)*100);
#endif // __IBMCPP__

#ifdef __INTEL_COMPILER // __INTEL_COMPILER = 500 for version 5.0 ...
	sprintf(cn, "Intel C/C++ %d.%d", __INTEL_COMPILER/100, __INTEL_COMPILER-(__INTEL_COMPILER/100)*100);
#endif // __INTEL_COMPILER

#ifdef _MSC_VER // _MSC_VER = 1300 for version 13 (.NET) ...
	sprintf(cn, "Microsoft C/C++ %d.%d", _MSC_VER/100, _MSC_VER-(_MSC_VER/100)*100);
#endif // _MSC_VER

#ifdef _MIPS_ISA
#ifdef _COMPILER_VERSION
	sprintf(cn, "MIPSpro %d.%d.%d", _COMPILER_VERSION/100, (_COMPILER_VERSION-(_COMPILER_VERSION/100)*100)/10, _COMPILER_VERSION-(_COMPILER_VERSION/10)*10);
#endif // _COMPILER_VERSION
#endif // _MIPS_ISA

#ifdef __GNUC__
	sprintf(cn, "GNU C %d.%d.%d", __GNUC__, __GNUC_MINOR__, __GNUC_PATCHLEVEL__);
#endif // __GNUC__

#ifdef __GNUG__
	sprintf(cn, "GNU C++ %d.%d.%d", __GNUC__, __GNUC_MINOR__, __GNUC_PATCHLEVEL__);
#endif // __GNUC__

	return cn;
}

std::ostream& operator<<(std::ostream& s, local_time t)
{
	static const char months[12][4] = { "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };
	return s << months[t.ltime->tm_mon] << ' ' << std::setw(2) << t.ltime->tm_mday << ' ' << 1900 + t.ltime->tm_year << ' ' << std::setfill('0') << std::setw(2) << t.ltime->tm_hour << ':' << std::setw(2) << t.ltime->tm_min << ':' << std::setw(2) << t.ltime->tm_sec;
}

#ifdef WIN32

#include <iostream>

sys_names::utsn::utsn()
{
	OSVERSIONINFO osvi;
	osvi.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
	if( GetVersionEx(&osvi) )
		sprintf(version, "%lu.%lu", osvi.dwMajorVersion, osvi.dwMinorVersion);
	else
		sprintf(version, "?.?");
}
const char* sys_names::host()
{
	return "royal";
}
const char* sys_names::os()
{
	return "Windows";
}
const char* sys_names::release()
{
	return name.version;
}

stopwatch::stopwatch() : overhead_tick(0), m_time(0.0)
{
	if( !QueryPerformanceFrequency(&perf_freq) )
	{
		std::clog << "Warning: no hardware support for high resolution timers." << std::endl;
		return;
	}
	// Messure overhead of 2xQueryPerformanceCounter()
	overhead_tick = 200;
	for( int i=0; i < 40; i++ )
	{
		LARGE_INTEGER b,e;
		int Ticks;
		QueryPerformanceCounter(&b);
		QueryPerformanceCounter(&e);
		Ticks = e.u.LowPart - b.u.LowPart;
		if( 0 <= Ticks && Ticks < overhead_tick )
			overhead_tick = Ticks;
	}
	// See if Freq fits in 32 bits; if not lose some precision. Should only apply to 4GHz+ processors?
	perf_freq_mag = 0;
	for( int High32 = perf_freq.u.HighPart; High32; High32>>=1 )
		perf_freq_mag++;
	begin.u.HighPart = 0;
	begin.u.LowPart = 0;
}

void stopwatch::do_stop(LARGE_INTEGER& end)
{
	int ReduceMag = 0;
	for( unsigned int High32 = end.u.HighPart-begin.u.HighPart; High32; High32>>=1 )
		ReduceMag++;
	if( perf_freq_mag || ReduceMag )
	{
		if( perf_freq_mag > ReduceMag )
			ReduceMag = perf_freq_mag;
		begin.QuadPart = Int64ShrlMod32(begin.QuadPart, ReduceMag);
		end.QuadPart = Int64ShrlMod32(end.QuadPart, ReduceMag);
		perf_freq.QuadPart = Int64ShrlMod32(perf_freq.QuadPart, ReduceMag);
		overhead_tick >>= ReduceMag;
	}
	if( perf_freq.u.LowPart != 0 )
		m_time += ((double)(end.u.LowPart-begin.u.LowPart-overhead_tick))/perf_freq.u.LowPart;
	begin.u.HighPart = 0;
	begin.u.LowPart = 0;
}


#if _WIN32_WINNT >= 0x0500

pf_counter::pf_counter() : begin(0), m_count(0)
{
	hJob = CreateJobObject(NULL, NULL);
	if( hJob == 0 )
	{
		std::clog << "Warning: jobs not supported." << std::endl;
		return;
	}
	HANDLE hProcess = GetCurrentProcess();
	if( !AssignProcessToJobObject(hJob, hProcess) )
	{
		std::clog << "Warning: unable to assign process to job." << std::endl;
		return;
	}
}

pf_counter::~pf_counter()
{
	CloseHandle(hJob);
}

void pf_counter::start()
{
	JOBOBJECT_BASIC_AND_IO_ACCOUNTING_INFORMATION info;
	DWORD length;
	if( QueryInformationJobObject(hJob, JobObjectBasicAndIoAccountingInformation, &info, sizeof(info), &length) )
		begin = info.BasicInfo.TotalPageFaultCount;
	else
	{
		std::clog << "Warning: unable to determine pagefaultcount." << std::endl;
		begin = 0;
		return;
	}
}

void pf_counter::stop()
{
	if( is_running() )
	{
		if( !hJob )
			return;
		JOBOBJECT_BASIC_AND_IO_ACCOUNTING_INFORMATION info;
		DWORD length;
		if( QueryInformationJobObject(hJob, JobObjectBasicAndIoAccountingInformation, &info, sizeof(info), &length) )
			m_count += info.BasicInfo.TotalPageFaultCount-begin;
		begin = 0;
	}
}

#else // _WIN32_WINNT

pf_counter::pf_counter() : begin(0), m_count(0)
{
}

pf_counter::~pf_counter()
{
}

void pf_counter::start()
{
}

void pf_counter::stop()
{
}

#endif // _WIN32_WINNT

#else // WIN32


const char* sys_names::host()
{
	return name.name.nodename;
}

const char* sys_names::os()
{
	return name.name.sysname;
}

const char* sys_names::release()
{
	return name.name.release;
}

stopwatch::stopwatch() : m_time(0.0)
{
	begin.tv_sec = 0;
	begin.tv_usec = 0;
	overhead_tick = 200;
	for( int i=0; i < 40; i++ )
	{
		long Ticks;
		struct timeval b, e;
		gettimeofday(&b, NULL);
		gettimeofday(&e, NULL);
		Ticks = e.tv_usec-b.tv_usec + 1000000*(e.tv_sec-b.tv_sec);
		if( 0 <= Ticks && Ticks < overhead_tick )
			overhead_tick = Ticks;
	}
}

void stopwatch::do_stop(struct timeval& end)
{
	m_time += (end.tv_sec+end.tv_usec/1000000.0)-(begin.tv_sec+(begin.tv_usec-overhead_tick)/1000000.0);
	begin.tv_sec = 0;
	begin.tv_usec = 0;
}


pf_counter::pf_counter() : begin(0), m_count(0) { }

void pf_counter::start()
{
	struct rusage usage_begin;
	getrusage(RUSAGE_SELF, &usage_begin);
	begin = usage_begin.ru_majflt;
}

void pf_counter::stop()
{
	if( is_running() )
	{
		struct rusage usage_end;
		getrusage(RUSAGE_SELF, &usage_end);
		m_count += usage_end.ru_majflt-begin;
		begin = 0;
	}
}

#endif // WIN32

std::ostream& operator<<(std::ostream& s, stopwatch::xt tag)
{
	for( int i=0; i<tag.indent; ++i )
		s << tag.ichar;
	return s << "<time unit=\"s\">" << tag.time << "</time>";
}

std::ostream& operator<<(std::ostream& s, pf_counter::xt tag)
{
	for( int i=0; i<tag.indent; ++i )
		s << tag.ichar;
	return s << "<pagefaults>" << tag.count << "</pagefaults>";
}

void pf_counter::reset()
{
	m_count = 0;
	if( is_running() )
		start();
}

sys_names::utsn sys_names::name;

#ifdef PAPI

#include <exception>

miss_counter::initializer::initializer()
{
	if( PAPI_library_init(PAPI_VER_CURRENT) != PAPI_VER_CURRENT )
		throw std::exception();//"PAPI_library_init");
}

miss_counter::initializer miss_counter::init;

miss_counter::miss_counter()
{
	if( PAPI_create_eventset(&EventSet) != PAPI_OK )
		throw std::exception();//"PAPI_create_eventset");
	if( PAPI_add_event(&EventSet, PAPI_L2_DCM) != PAPI_OK )
		throw std::exception();//"PAPI_add_event");
    if( PAPI_add_event(&EventSet, PAPI_TLB_TL) != PAPI_OK )
		throw std::exception();//"PAPI_add_event");
    if( PAPI_reset(EventSet) != PAPI_OK )
		throw std::exception();//"PAPI_reset");
}

miss_counter::~miss_counter()
{
	if( PAPI_rem_event(&EventSet, PAPI_TLB_TL) != PAPI_OK )
		throw std::exception();//"PAPI_rem_event");
    if( PAPI_rem_event(&EventSet, PAPI_L2_DCM) != PAPI_OK )
		throw std::exception();//"PAPI_rem_event");
	if( PAPI_destroy_eventset(&EventSet) != PAPI_OK )
		throw std::exception();//"PAPI_destroy_eventset");
}

void miss_counter::start()
{
	if( PAPI_start(EventSet) != PAPI_OK )
		throw std::exception();//"PAPI_start");
}

void miss_counter::stop()
{
	long_long counters[2];
	if( PAPI_stop(EventSet, counters) != PAPI_OK )
		throw std::exception();//"PAPI_stop");
}

void miss_counter::reset()
{
    if( PAPI_reset(EventSet) != PAPI_OK )
		throw std::exception();//"PAPI_reset");
}

std::ostream& operator<<(std::ostream& s, miss_counter::xt tag)
{
	for( int i=0; i<tag.indent; ++i )
		s << tag.ichar;
	return s << "<l2misses>" << tag.l2 << "</l2misses><tlbmisses>" << tag.tlb << "</tlbmisses>";
}

#else // PAPI

std::ostream& operator<<(std::ostream& s, miss_counter::xt tag)
{
	return s;
}

#endif // PAPI
