/*
 * Copyright (C) 2003 Kristoffer Vinther.
 * For contact information, please visit
 * http://www.brics.dk/~kv/thesis/.
 * If you use this code in commercial applications, please notify me.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * Free Software Foundation, Inc.
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __BENCHSUITE_H_INCLUDED_
#define __BENCHSUITE_H_INCLUDED_

#include <iostream>
#include <algorithm>
#include "Benchmark.h"

#include "Sort/PointerBMT.h"
#include "Sort/ImplicitBMT.h"
#include "Sort/LazyMerger.h"
#include "Sort/HeapMerger.h"

extern unsigned long TreeOps;
//typedef Counting<long int> sort_t;
typedef long int sort_t;

template<class SortAlg, class ArrayClass, class ArrayInitializer>
class Benchmarker
{
protected:
	typedef typename ArrayClass::size_type size_type;
public:
	static void Run(size_type size)
	{
		std::cout << "\t\t<array>" << ArrayClass::name() << "</array>" << std::endl;
		ArrayClass src(size);
		ArrayClass dest(size);
		ArrayInitializer::InitArray(src.begin(), src.end(size));
		start(size);
		SortAlg::Sort(src.begin(), src.end(size), dest.begin());
		stop();
	}
protected:
	inline static void start(size_type size)
	{
		std::cout << "\t\t<algorithm>" << SortAlg::name() << "</algorithm>" << std::endl << "\t\t<size>" << size << "</size>" << std::endl;
		MultiMapArray<sort_t>::PageFaults = 0;
		Counting<long int>::reset();
		TreeOps = 0;
		BeginPFCounter();
		StartTimer();
	}
	inline static void stop()
	{
		double time = StopTimer();
		unsigned long pf = EndPFCounter();
		std::cout	<< "\t\t<time unit=\"s\">" << time << "</time>" << std::endl 
					<< "\t\t<comparisons>" << Counting<long int>::getComps() << "</comparisons>" << std::endl 
					<< "\t\t<moves>" << Counting<long int>::getComps() << "</moves>" << std::endl
					<< "\t\t<pagefaults>" << pf << "</pagefaults>" << std::endl
					<< "\t\t<remaps>" << MultiMapArray<sort_t>::PageFaults << "</remaps>" << std::endl;
	}
};

template<class SortAlg, class ArrayClass, class ArrayInitializer>
class BenchmarkerI : public Benchmarker<SortAlg, ArrayClass, ArrayInitializer>
{
public:
	static void Run(typename Benchmarker<SortAlg, ArrayClass, ArrayInitializer>::size_type size)
	{
		std::cout << "\t\t<array>" << ArrayClass::name() << "</array>" << std::endl;
		ArrayClass src(size);
		ArrayInitializer::InitArray(src.begin(), src.end(size));
		start(size);
		SortAlg::Sort(src.begin(), src.end(size));
		stop();
	}
};
template<class SortAlg, class T, class ArrayInitializer>
class BenchmarkerC
{
public:
	static void Run(size_t size)
	{
		std::cout << "\t\t<array>c</array>" << std::endl;
		T* src = new T[size];
		T* dest = new T[size];
//		Sink<T> dest;
		ArrayInitializer::InitArray(src, src+size);
//		SortAlg::Sort(src, src+size, dest);
		start(size);
		SortAlg::Sort(src, src+size, dest);
		stop();
		delete[] dest;
		delete[] src;
	}
protected:
	inline static void start(size_t size)
	{
		std::cout << "\t\t<algorithm>" << SortAlg::name() << "</algorithm>" << std::endl << "\t\t<size>" << size << "</size>" << std::endl;
		MultiMapArray<sort_t>::PageFaults = 0;
		Counting<long int>::reset();
		TreeOps = 0;
		BeginPFCounter();
		StartTimer();
	}
	inline static void stop()
	{
		double time = StopTimer();
		unsigned long pf = EndPFCounter();
		std::cout	<< "\t\t<time unit=\"s\">" << time << "</time>" << std::endl 
					<< "\t\t<comparisons>" << Counting<long int>::getComps() << "</comparisons>" << std::endl 
					<< "\t\t<moves>" << Counting<long int>::getComps() << "</moves>" << std::endl
					<< "\t\t<pagefaults>" << pf << "</pagefaults>" << std::endl
					<< "\t\t<remaps>" << MultiMapArray<sort_t>::PageFaults << "</remaps>" << std::endl;
	}
};
template<class SortAlg, class T, class ArrayInitializer>
class BenchmarkerIC : public BenchmarkerC<SortAlg, T, ArrayInitializer>
{
public:
	static void Run(size_t size)
	{
		std::cout << "\t\t<array>c</array>" << std::endl;
		T* src = new T[size];
		ArrayInitializer::InitArray(src, src+size);
		start(size);
		SortAlg::Sort(src, src+size);
		stop();
		delete[] src;
	}
};
template<class SortAlg, class ArrayClass, class ArrayInitializer>
class BenchmarkerM : public Benchmarker<SortAlg, ArrayClass, ArrayInitializer>
{
public:
	static void Run(typename Benchmarker<SortAlg, ArrayClass, ArrayInitializer>::size_type size)
	{
		std::cout << "\t\t<array>" << ArrayClass::name() << "</array>" << std::endl;
		ArrayClass src(SRC_FILE, MAP_CREATE | MAP_READ | MAP_WRITE | MAP_TEMPORARY, size);
		ArrayClass dest(DST_FILE, MAP_CREATE | MAP_WRITE | MAP_TEMPORARY, size);
		ArrayInitializer::InitArray(src.begin(), src.end(size));
		start(size);
		SortAlg::Sort(src.begin(), src.end(size), dest.begin());
		stop();
	}
};
template<class SortAlg, class ArrayClass, class ArrayInitializer>
class BenchmarkerIM : public Benchmarker<SortAlg, ArrayClass, ArrayInitializer>
{
public:
	static void Run(typename Benchmarker<SortAlg, ArrayClass, ArrayInitializer>::size_type size)
	{
		std::cout << "\t\t<array>" << ArrayClass::name() << "</array>" << std::endl;
		ArrayClass src(SRC_FILE, MAP_CREATE | MAP_READ | MAP_WRITE | MAP_TEMPORARY, size);
		ArrayInitializer::InitArray(src.begin(), src.end(size));
		start(size);
		SortAlg::Sort(src.begin(), src.end(size));
		stop();
	}
};







class std_sorter
{
public:
	template<class RanIt>
	inline static void Sort(RanIt begin, RanIt end)
	{
		std::sort(begin, end);
	}
	static const char* name()
		{ return "std::sort"; }
};

template<class T>
int __qsort_comp(const void* a, const void* b)
{
	return *(T*)a - *(T*)b;
}

class qsorter
{
public:
	template<class RanIt>
	inline static void Sort(RanIt begin, RanIt end)
	{
		Sort(begin, end, *begin);
	}
	static const char* name()
		{ return "qsort"; }
private:
	template<class RanIt, class T>
	inline static void Sort(RanIt begin, RanIt end, T& dummy)
	{
		(void)dummy;
		int (*compare)(const void *, const void *) = __qsort_comp<T>;
		qsort(begin, end-begin, sizeof(*begin), compare);
	}
};

class std_heapsorter
{
public:
	template<class RanIt>
	inline static void Sort(RanIt begin, RanIt end)
	{
		std::make_heap(begin, end);
		std::sort_heap(begin, end);
	}
	static const char* name()
		{ return "std::sort_heap"; }
};

class imergesorter
{
public:
	template<class RanIt>
	inline static void Sort(RanIt begin, RanIt end)
	{
		inplace_mergesort(begin, end);
	}
	static const char* name()
		{ return "inplace_ms"; }
};

class QuickSorter
{
public:
	template<class RanIt>
	inline static void Sort(RanIt begin, RanIt end)
	{
		Sort(*begin, begin, end);
	}
	static const char* name()
		{ return "QuickSort"; }
private:
	template<class T, class RanIt>
	inline static void Sort(T t, RanIt begin, RanIt end)
	{
		(void)t;
		QuickSort<T>(begin, end-begin);
	}
};

class mergesorter
{
public:
	template<class RanIt, class FwdIt>
	inline static void Sort(RanIt begin, RanIt end, FwdIt dest)
	{
		mergesort(begin, end, dest);
	}
	static const char* name()
		{ return "mergesort"; }
};

class MemMergesorter
{
public:
	template<class RanIt, class FwdIt>
	inline static void Sort(RanIt begin, RanIt end, FwdIt dest)
	{
		MergeSort<HeapMerger<sort_t, InputStream<RanIt>, 200000000, 4096> >(begin, end, dest);
	}
	static const char* name()
		{ return "MemMergesort"; }
};

class CacheMergesorter
{
public:
	template<class RanIt, class FwdIt>
	inline static void Sort(RanIt begin, RanIt end, FwdIt dest)
	{
		MergeSort<HeapMerger<sort_t, InputStream<RanIt>, 240000, 128> >(begin, end, dest);
	}
	static const char* name()
		{ return "CacheMergesort"; }
};

class FunnelELSorter
{
public:
	template<class RanIt, class FwdIt>
	inline static void Sort(RanIt begin, RanIt end, FwdIt dest)
	{
//		MergeSort<LazyMerger<sort_t, typename std::vector<std::pair<RanIt,RanIt> >::iterator, PBBinMergeTree<sort_t> > >(begin, end, dest);
		MergeSort<LazyMerger<PBBinMergeTree<sort_t>, InputStream<RanIt> > >(begin, end, dest);
	}
	static const char* name()
		{ return "FunnelSort[EL]"; }
};

class FunnelILSorter
{
public:
	template<class RanIt, class FwdIt>
	inline static void Sort(RanIt begin, RanIt end, FwdIt dest)
	{
//		MergeSort<LazyMerger<sort_t, typename std::vector<std::pair<RanIt,RanIt> >::iterator, ImplBinMergeTree<sort_t> > >(begin, end, dest);
		MergeSort<LazyMerger<ImplBinMergeTree<sort_t>, InputStream<RanIt> > >(begin, end, dest);
	}
	static const char* name()
		{ return "FunnelSort[IL]"; }
};

class FunnelEILSorter
{
public:
	template<class RanIt>
	inline static void Sort(RanIt begin, RanIt end)
	{
		InplaceMergeSort<LazyMerger<PBBinMergeTree<sort_t>, InputStream<RanIt> > >(begin, end);
	}
	static const char* name()
		{ return "FunnelSort[EIL]"; }
};



#include <stdlib.h>
#include <time.h>

class RandArray
{
public:
	template<class FwdIt>
	static void InitArray(FwdIt begin, FwdIt end)
	{
#ifdef WIN32
#ifndef _DEBUG
		srand((unsigned)time(NULL));
#endif// _DEBUG
		while( begin != end )
			*begin = rand()^(rand()<<8)^(rand()<<16)^(rand()<<24), begin++;
#else // WIN32
#ifndef _DEBUG
		srand48((long int)time(NULL));
#endif// _DEBUG
		while( begin != end )
			*begin = lrand48(), begin++;
#endif // WIN32
	}
};



template<class RanIt, class T>
inline void _m9k_move(RanIt begin, RanIt end, T* min, int k)
{
	for( RanIt i=begin+1; i<end; i++ )
		*i = *(i+8-k);
	*end = min[k];
//	MedianKiller(begin+1, end-1);
}

template<class RanIt, class T, class Diff>
inline void _Med9Killer(RanIt begin, RanIt end, T e, Diff d)
{
	RanIt mid = begin+(end-begin)/2;
	Diff step = (end-begin)/8;
	T min[8];
	RanIt i=begin+1;
	for( int j=0; j<8; i++, j++ )
		min[j] = *i;
	_m9k_move(begin, begin+step, min, 0);
	_m9k_move(begin+step, begin+2*step, min, 1);
	_m9k_move(begin+2*step, mid-step, min, 2);
	_m9k_move(mid-step, mid, min, 3);
	_m9k_move(mid, mid+step, min, 4);
	_m9k_move(mid+step, end-2*step-1, min, 5);
	_m9k_move(end-2*step-1, end-step-1, min, 6);
	_m9k_move(end-step-1, end-1, min, 7);
}

template<class RanIt>
void MedianKiller(RanIt begin, RanIt end)
{
	if( 40 < end-begin )
	{ // Median-of-nine-killer
		_Med9Killer(begin, end, *begin, end-begin);
	}
	else if( 7 < end-begin )
	{ // Median-of-tree-killer
		RanIt mid = begin+(end-begin)/2;
		std::rotate(begin+1, begin+3, mid+1);
		std::rotate(mid, mid+1, end);
		std::iter_swap(mid-1, mid);
		MedianKiller(begin+1, mid);
		MedianKiller(mid+1, end);
	}
/*	else
		std::reverse(begin, end);
*/}

/*			for( size_t i=0; i<size; i++ )
				src[i] = rand();
			for( size_t i=0; i<size/2; i+=2 )
				src[i] = i+1;
			for( size_t i=1; i<size/2; i+=2 )
				src[i] = size/2+i;
			for( size_t i=size/2; i<size; i++ )
				src[i] = 2*(i-size/2+1);
*/
#endif // __BENCHSUITE_H_INCLUDED_
