/*
 * Copyright (C) 2003 Kristoffer Vinther.
 * For contact information, please visit
 * http://www.brics.dk/~kv/thesis/.
 * If you use this code in commercial applications, please notify me.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * Free Software Foundation, Inc.
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <iostream> // For stderr output
#include "Benchmark.h"

const char* compilername()
{
	static char cn[80];
#ifdef __IBMC__ // __IBMC__ = 430 for version 4.3 ...
	sprintf(cn, "IBM C %d.%d", __IBMC__/100, __IBMC__-(__IBMC__/100)*100);
#endif // __IBMCPP__

#ifdef __IBMCPP__ // __IBMCPP__ = 360 for version 3.6 ...
	sprintf(cn, "IBM C++ %d.%d", __IBMCPP__/100, __IBMCPP__-(__IBMCPP__/100)*100);
#endif // __IBMCPP__

#ifdef __INTEL_COMPILER // __INTEL_COMPILER = 500 for version 5.0 ...
	sprintf(cn, "Intel C/C++ %d.%d", __INTEL_COMPILER/100, __INTEL_COMPILER-(__INTEL_COMPILER/100)*100);
#endif // __INTEL_COMPILER

#ifdef _MSC_VER // _MSC_VER = 1300 for version 13 (.NET) ...
	sprintf(cn, "Microsoft C/C++ %d.%d", _MSC_VER/100, _MSC_VER-(_MSC_VER/100)*100);
#endif // _MSC_VER

#ifdef _MIPS_ISA
#ifdef _COMPILER_VERSION
	sprintf(cn, "MIPSpro %d.%d.%d", _COMPILER_VERSION/100, (_COMPILER_VERSION-(_COMPILER_VERSION/100)*100)/10, _COMPILER_VERSION-(_COMPILER_VERSION/10)*10);
#endif // _COMPILER_VERSION
#endif // _MIPS_ISA

#ifdef __GNUC__
	sprintf(cn, "GNU C %d.%d.%d", __GNUC__, __GNUC_MINOR__, __GNUC_PATCHLEVEL__);
#endif // __GNUC__

#ifdef __GNUG__
	sprintf(cn, "GNU C++ %d.%d.%d", __GNUC__, __GNUC_MINOR__, __GNUC_PATCHLEVEL__);
#endif // __GNUC__

	return cn;
}

const char* buildtime()
{
	return __DATE__ " " __TIME__;
}

#ifdef WIN32

/* HACK: Assume Win2000 or later
#undef _WIN32_WINNT
#define _WIN32_WINNT 0x0500
*/
#include "ansiwin.h"
#include <time.h>

void DummyFunc() { return; }
LARGE_INTEGER __PerfFreq, __begin_perf;
int __OverheadTick, __PerfFreqMag;
void StartTimer()
{
	if( !QueryPerformanceFrequency(&__PerfFreq) )
	{
		std::cerr << "Warning: no hardware support for high resolution timers." << std::endl;
		return;
	}
	// Messure overhead of 2xQueryPerformanceCounter() and a function call+return
	void (*pFunc)() = DummyFunc;
	__OverheadTick = 200;
	for( int i=0; i < 20; i++ )
	{
		LARGE_INTEGER b,e;
		int Ticks;
		QueryPerformanceCounter(&b);
		(*pFunc)();
		QueryPerformanceCounter(&e);
		Ticks = e.u.LowPart - b.u.LowPart;
		if( Ticks >= 0 && Ticks < __OverheadTick )
			__OverheadTick = Ticks;
	}
	// See if Freq fits in 32 bits; if not lose some precision. Should only apply to 4GHz+ processors?
	__PerfFreqMag = 0;
	int High32 = __PerfFreq.u.HighPart;
	while( High32 )
	{
		High32 >>= 1;
		__PerfFreqMag++;
	}
	QueryPerformanceCounter(&__begin_perf);
}

double StopTimer()
{
	LARGE_INTEGER end;
	QueryPerformanceCounter(&end);
	unsigned int High32 = end.u.HighPart - __begin_perf.u.HighPart;
	int ReduceMag = 0;
	while( High32 )
	{
		High32 >>= 1;
		ReduceMag++;
	}
	if( __PerfFreqMag || ReduceMag )
	{
		if( __PerfFreqMag > ReduceMag )
			ReduceMag = __PerfFreqMag;
		__begin_perf.QuadPart = Int64ShrlMod32(__begin_perf.QuadPart, ReduceMag);
		end.QuadPart = Int64ShrlMod32(end.QuadPart, ReduceMag);
		__PerfFreq.QuadPart = Int64ShrlMod32(__PerfFreq.QuadPart, ReduceMag);
		__OverheadTick >>= ReduceMag;
	}
	// Reduced numbers to 32 bits, now can do the math
	double t;
	if( __PerfFreq.u.LowPart == 0 )
		t = 0.0;
	else
		t = ((double)(end.u.LowPart - __begin_perf.u.LowPart - __OverheadTick))/__PerfFreq.u.LowPart;
	return t;
}

#if _WIN32_WINNT >= 0x0500

HANDLE __hJob;
unsigned long __pfCount;
void BeginPFCounter()
{
	HANDLE hProcess = GetCurrentProcess();
	__hJob = CreateJobObject(NULL, NULL);
	if( __hJob == 0 )
	{
		std::cerr << "Warning: jobs not supported." << std::endl;
		__pfCount = 0;
		return;
	}
	if( !AssignProcessToJobObject(__hJob, hProcess) )
	{
		std::cerr << "Warning: unable to assign process to job." << std::endl;
		__pfCount = 0;
		return;
	}
	JOBOBJECT_BASIC_AND_IO_ACCOUNTING_INFORMATION info;
	DWORD length;
	if( QueryInformationJobObject(__hJob, JobObjectBasicAndIoAccountingInformation, &info, sizeof(info), &length) )
		__pfCount = info.BasicInfo.TotalPageFaultCount;
	else
	{
		std::cerr << "Warning: unable to determine pagefaultcount." << std::endl;
		__pfCount = 0;
		return;
	}
}

unsigned long EndPFCounter()
{
	if( !__hJob )
		return 0;
	JOBOBJECT_BASIC_AND_IO_ACCOUNTING_INFORMATION info;
	DWORD length;
	unsigned long count;
	if( QueryInformationJobObject(__hJob, JobObjectBasicAndIoAccountingInformation, &info, sizeof(info), &length) )
		count = info.BasicInfo.TotalPageFaultCount;
	else
		count = 0;
	CloseHandle(__hJob);
	__hJob = NULL;
	return count-__pfCount;
}

#else

void BeginPFCounter()
{ }

unsigned long EndPFCounter()
{
	return 0;
}

#endif

#else // WIN32


#include <sys/resource.h>
#include <sys/time.h>
#include <unistd.h>

struct timeval __begin;
struct rusage __usage_begin;
void StartTimer()
{
	gettimeofday(&__begin, NULL);
}

double StopTimer()
{
	static struct timeval end;
	gettimeofday(&end, NULL);
	return (end.tv_sec+end.tv_usec/1000000.0)-(__begin.tv_sec+__begin.tv_usec/1000000.0);
}

void BeginPFCounter()
{
	getrusage(RUSAGE_SELF, &__usage_begin);
}

unsigned long EndPFCounter()
{
	static struct rusage usage_end;
	getrusage(RUSAGE_SELF, &usage_end);
	return usage_end.ru_majflt-__usage_begin.ru_majflt;
}

#endif // WIN32


#ifdef PAPI

#include <papiStdEventDefs.h>
#include <papi.h>
#include <exception>
#include "Sort/int64.h"

int EventSet;

void StartMisses()
{
	if( PAPI_library_init(PAPI_VER_CURRENT) != PAPI_VER_CURRENT )
		throw std::exception();//"PAPI_library_init");
	if( PAPI_create_eventset(&EventSet) != PAPI_OK )
		throw std::exception();//"PAPI_create_eventset");
	if( PAPI_add_event(&EventSet, PAPI_L2_DCM) != PAPI_OK )
		throw std::exception();//"PAPI_add_event");
    if( PAPI_add_event(&EventSet, PAPI_TLB_TL) != PAPI_OK )
		throw std::exception();//"PAPI_add_event");
    if( PAPI_reset(EventSet) != PAPI_OK )
		throw std::exception();//"PAPI_reset");
	if( PAPI_start(EventSet) != PAPI_OK )
		throw std::exception();//"PAPI_start");
}

std::pair<_int64_t,_int64_t> StopMisses()
{
	long_long counters[2];
	if( PAPI_stop(EventSet, counters) != PAPI_OK )
		throw std::exception();//"PAPI_stop");
	if( PAPI_rem_event(&EventSet, PAPI_TLB_TL) != PAPI_OK )
		throw std::exception();//"PAPI_rem_event");
    if( PAPI_rem_event(&EventSet, PAPI_L2_DCM) != PAPI_OK )
		throw std::exception();//"PAPI_rem_event");
	if( PAPI_destroy_eventset(&EventSet) != PAPI_OK )
		throw std::exception();//"PAPI_destroy_eventset");
	return std::pair<_int64_t,_int64_t>(counters[0],counters[1]);
}

#endif // PAPI
