/*
 * Copyright (C) 2003 Kristoffer Vinther.
 * For contact information, please visit
 * http://www.brics.dk/~kv/thesis/.
 * If you use this code in commercial applications, please notify me.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * Free Software Foundation, Inc.
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "../Bench/aarray.h"
#include "../Bench/elements.h"
#include <exception>
#include <iostream>
#include <cstdio>

void exit_usage()
{
	std::cerr << "chksort -e{i|p|r} file [source]" << std::endl;
	exit(1);
}

template<class T>
void check_sort(const char *file, const char *source)
{
	T chksum;
	{
		typename auto_array::map_array<T> A(file, auto_array::MAP_EXISTING | auto_array::MAP_READ | auto_array::MAP_WRITE);
		typename auto_array::map_array<T>::iterator p = A.begin();
		chksum = *p;
		T last = *p;
		for( ++p; p!=A.end(); ++p )
		{
			if( *p < last )
				std::cout << "Elements " << static_cast<int>(p-A.begin()-1) << " and " << static_cast<int>(p-A.begin()) << " are out of order: " << last << '>' << '=' << *p << '!' << std::endl;
			chksum += *p;
			last = *p;
		}
	}
	if( source )
	{
		typename auto_array::map_array<T> A(source, auto_array::MAP_EXISTING | auto_array::MAP_READ | auto_array::MAP_WRITE);
		typename auto_array::map_array<T>::iterator p = A.begin();
		T outsum = *p;
		for( ++p; p!=A.end(); ++p )
			outsum += *p;
		if( outsum != chksum )
			std::cout << "Checksum error: " << chksum << " != " << outsum << '!' << std::endl;
	}
}

int main(int argc, char *argv[])
{
	if( argc < 3 || argc > 4 )
		exit_usage();
	if( argv[1][0] != '-' || argv[1][1] != 'e' )
		exit_usage();
	try
	{
		switch( argv[1][2] )
		{
		case 'i':
			check_sort<int>(argv[2], argc==4? argv[3] : NULL);
			break;
		case 'p':
			check_sort<element>(argv[2], argc==4? argv[3] : NULL);
			break;
		case 'r':
			check_sort<record>(argv[2], argc==4? argv[3] : NULL);
			break;
		default:
			exit_usage();
		}
	}
	catch( std::exception& e )
	{
		std::cerr << "Exception in sortgen: " << e.what() << std::endl;
		return 1;
	}
	return 0;
}
