/*
 * Copyright (C) 2003 Kristoffer Vinther.
 * For contact information, please visit
 * http://www.brics.dk/~kv/thesis/.
 * If you use this code in commercial applications, please notify me.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * Free Software Foundation, Inc.
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

template<class T>
T _clog(T n)
{
	T logn = 0;
	for( T i=n-1; i; logn++, i>>=1 );
	return logn;
}

template<class T>
inline void ITERATORS_MUST_HAVE_SAME_DIFFERENCE_TYPE(T, T, T) { }
template<class T>
inline void ITERATORS_MUST_HAVE_SAME_VALUE_TYPE(T, T, T) { }
template<class T>
inline void ITERATORS_MUST_HAVE_SAME_VALUE_TYPE(T, T, T, T) { }
template<class T>
inline void ITERATORS_MUST_POINT_TO_PAIRS_OF_ITERATORS_OF_SAME_TYPE(std::pair<T, T>) { }

class LazyMergeNopOut
{
public:
	inline void operator()()
		{ }
};

class LazyMergeNop
{
public:
	template<class It>
	inline bool operator()(It&, It&)
		{ /*assert( begin == end );*/ return true; }
};


class LazyMergeStop
{
public:
	template<class It>
	inline bool operator()(It&, It&)
		{ return false; }
};

template<class Exception>
class LazyMergeThrow
{
public:
	LazyMergeThrow(const Exception& ex) : m_Exception(ex) { }
	template<class It>
	inline bool operator()(It&, It&) /*throw( Exception& )*/
		{ throw m_Exception; return false; }
private:
	Exception m_Exception;
};

class LazyMergeFail
{
public:
	template<class It>
	inline bool operator()(It& begin, It& end)
		{ (void)begin; (void)end; assert( begin == end ); assert( false ); return false; }
};


template<class InIt, class OutIt, class OnEmpty>
inline void _LazyCopy(InIt& inbegin, InIt inend, OutIt& outbegin, OutIt outend, OnEmpty oninempty)
{
	LazyMergeNopOut nop;
	_LazyCopy(inbegin, inend, outbegin, outend, oninempty, nop);
}

template<class InIt, class OutIt>
inline void _LazyCopy(InIt& inbegin, InIt inend, OutIt& outbegin, OutIt outend)
{
	LazyMergeNop nop;
	_LazyCopy(inbegin, inend, outbegin, outend, nop, nop);
}






template<class In1It, class In2It, class OutIt>
inline void LazyMerge(In1It& in1begin, In1It in1end, In2It& in2begin, In2It in2end, OutIt& outbegin, OutIt outend)
{
//	ITERATORS_MUST_HAVE_SAME_DIFFERENCE_TYPE
//		(in1end-in1begin, in2end-in2begin, outend-outbegin);
	if( in1begin!=in1end && in2begin!=in2end && outbegin!=outend )
	ITERATORS_MUST_HAVE_SAME_VALUE_TYPE
		(*in1begin, *in2begin, *outbegin);
	LazyMergeNop nop;
	LazyMergeNopOut nopout;
	_LazyMerge(in1begin, in1end, in2begin, in2end, outbegin, outend, 0, nop, nop, nopout, in1end-in1begin);
}
template<class In1It, class In2It, class OutIt>
inline void LazyMerge(In1It& in1begin, In1It in1end, In2It& in2begin, In2It in2end, OutIt& outbegin, OutIt outend, int limit)
{
//	ITERATORS_MUST_HAVE_SAME_DIFFERENCE_TYPE
//		(in1end-in1begin, in2end-in2begin, outend-outbegin);
	if( in1begin!=in1end && in2begin!=in2end && outbegin!=outend )
	ITERATORS_MUST_HAVE_SAME_VALUE_TYPE
		(*in1begin, *in2begin, *outbegin);
	LazyMergeNop nop;
	LazyMergeNopOut nopout;
	_LazyMerge(in1begin, in1end, in2begin, in2end, outbegin, outend, nop, nop, nopout, in1end-in1begin);
}
template<class In1It, class In2It, class OutIt, class OnEmpty1, class OnEmpty2>
inline void LazyMerge(In1It& in1begin, In1It in1end, In2It& in2begin, In2It in2end, OutIt& outbegin, OutIt outend, int limit, OnEmpty1 onin1empty, OnEmpty2 onin2empty)
{
//	ITERATORS_MUST_HAVE_SAME_DIFFERENCE_TYPE
//		(in1end-in1begin, in2end-in2begin, outend-outbegin);
	if( in1begin!=in1end && in2begin!=in2end && outbegin!=outend )
	ITERATORS_MUST_HAVE_SAME_VALUE_TYPE
		(*in1begin, *in2begin, *outbegin);
	LazyMergeNopOut nopout;
	_LazyMerge(in1begin, in1end, in2begin, in2end, outbegin, outend, limit, onin1empty, onin2empty, nopout, in1end-in1begin);
}
template<class In1It, class In2It, class OutIt, class OnEmpty1, class OnEmpty2, class OnFull>
inline void LazyMerge(In1It& in1begin, In1It in1end, In2It& in2begin, In2It in2end, OutIt& outbegin, OutIt outend, int limit, OnEmpty1 onin1empty, OnEmpty2 onin2empty, OnFull onoutfull)
{
//	ITERATORS_MUST_HAVE_SAME_DIFFERENCE_TYPE
//		(in1end-in1begin, in2end-in2begin, outend-outbegin);
	if( in1begin!=in1end && in2begin!=in2end && outbegin!=outend )
	ITERATORS_MUST_HAVE_SAME_VALUE_TYPE
		(*in1begin, *in2begin, *outbegin);
	_LazyMerge(in1begin, in1end, in2begin, in2end, outbegin, outend, limit, onin1empty, onin2empty, onoutfull, in1end-in1begin);
}
