/*
 * Copyright (C) 2003 Kristoffer Vinther.
 * For contact information, please visit
 * http://www.brics.dk/~kv/thesis/.
 * If you use this code in commercial applications, please notify me.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * Free Software Foundation, Inc.
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __BASICMERGER_H_INCLUDED_
#define __BASICMERGER_H_INCLUDED_

#include <stddef.h>
#include <cassert>
#include <utility>

#ifdef _MSC_VER
#define inline __forceinline
#endif // _MSC_VER

#include "BasicMerger.aux.h"

template<class InIt, class OutIt, class OnEmpty, class OnFull>
inline void _LazyCopy(InIt& inbegin, InIt inend, OutIt& outbegin, OutIt outend, OnEmpty oninempty, OnFull onoutfull)
{
	register InIt b = inbegin;
	register OutIt ob = outbegin;
	while( inend-b <= outend-ob )
	{
		while( b != inend )
			*ob = *b, ++b, ++ob;
		inbegin = b;
		outbegin = ob;
		if( !oninempty(inbegin, inend) || inbegin == inend )
			return;
		b = inbegin;
	}
	while( ob != outend )
		*ob = *b, ++b, ++ob;
	inbegin = b;
	outbegin = ob;
	onoutfull();
}

template<class In1It, class In2It, class OutIt, class OnEmpty1, class OnEmpty2, class OnFull, class Diff>
inline void _LazyMerge(In1It& in1begin, In1It in1end, In2It& in2begin, In2It in2end, OutIt& outbegin, OutIt outend, int limit, OnEmpty1 onin1empty, OnEmpty2 onin2empty, OnFull onoutfull, Diff min)
{
	assert( in1begin <= in1end );
	assert( in2begin <= in2end );
//	assert( outbegin <= outend );
/*	if( in1begin == in1end )
		if( !onin1empty(in1begin, in1end) )
			return outbegin;
	if( in2begin == in2end )
		if( !onin2empty(in2begin, in2end) )
			return outbegin;
*/	register In1It b1 = in1begin;
	register In2It b2 = in2begin;
	register OutIt ob = outbegin;

	do
	{
/*		if( in1end-b1 < limit )
		{
			in1begin = b1;
			onin1empty(in1begin, in1end);
			b1 = in1begin;
		}
		if( in2end-b2 < limit )
		{
			in2begin = b2;
			onin2empty(in2begin, in2end);
			b2 = in2begin;
		}
*/
		min = in1end - b1;
		if( in2end - b2 < min )
			min = in2end - b2;
		if( outend - ob < min )
			min = (Diff)(outend - ob);

		while( min )
		{
			min--;
			if( *b1 < *b2 )
				*ob = *b1, ++b1;
			else
				*ob = *b2, ++b2;
			++ob;
		}
		if( in1end-b1 <= (Diff)limit )
		{
			in1begin = b1;
			outbegin = ob;
			if( !onin1empty(in1begin, in1end) )
			{
				in2begin = b2;
				return;
			}
			if( in1begin == in1end )
			{
				in2begin = b2;
				outbegin = ob;
				_LazyCopy(in2begin, in2end, outbegin, outend, onin2empty, onoutfull);
				return;
			}
			b1 = in1begin;
		}
		else if( in2end-b2 <= (Diff)limit )
		{
			in2begin = b2;
			outbegin = ob;
			if( !onin2empty(in2begin, in2end) )
			{
				in1begin = b1;
				return;
			}
			if( in2begin == in2end )
			{
				in1begin = b1;
				outbegin = ob;
				_LazyCopy(in1begin, in1end, outbegin, outend, onin1empty, onoutfull);
				return;
			}
			b2 = in2begin;
		}
	} while( ob != outend );
	in1begin = b1;
	in2begin = b2;
	outbegin = ob;
	onoutfull();
}



template<class InIt, class OutIt, class OnEmpty1, class OnEmpty2, class OnFull, class Diff>
inline void _lazy_merge(InIt& in1begin, InIt in1end, InIt& in2begin, InIt in2end, OutIt& outbegin, OutIt outend, OnEmpty1 onin1empty, OnEmpty2 onin2empty, OnFull onoutfull, Diff min)
{
	assert( in1begin <= in1end );
	assert( in2begin <= in2end );
	assert( outbegin <= outend );
/*	if( in1begin == in1end )
		if( !onin1empty(in1begin, in1end) )
			return outbegin;
	if( in2begin == in2end )
		if( !onin2empty(in2begin, in2end) )
			return outbegin;
*/	InIt e[2] = { in1end, in2end };
	register OutIt ob = outbegin;
	register InIt b[2];
	b[0] = in1begin;
	b[1] = in2begin;
	int active = 2;

	do
	{
		min = (Diff)(outend - ob);
		for( int i=0; i<active; ++i )
			if( e[i] - b[i] < min )
				min = e[i] - b[i];

		for( ; min; --min )
		{
			if( *b[0] < *b[1] )
				*ob = *b[0], ++b[0];
			else
				*ob = *b[1], ++b[1];
			++ob;
/*			InIt *m;
			if( *b[0] < *b[1] )
				m = b;
			else
				m = b+1;
			*ob++ = **m, (*m)++;
*/		}
		if( e[0] == b[0] )
		{
			in1begin = b[0];
			outbegin = ob;
			onin1empty(in1begin, in1end);
			in2begin = b[1];
			return;
		}
		else if( e[1] == b[1] )
		{
			in2begin = b[1];
			outbegin = ob;
			onin2empty(in2begin, in2end);
			in1begin = b[0];
			return;
		}
	} while( ob != outend );
	in1begin = b[0];
	in2begin = b[1];
	outbegin = ob;
	onoutfull();
}

#ifdef _MSC_VER
#undef inline
#endif // _MSC_VER

#endif // __BASICMERGER_H_INCLUDED_
