/*
 * Copyright (C) 2003 Kristoffer Vinther.
 * For contact information, please visit
 * http://www.brics.dk/~kv/thesis/.
 * If you use this code in commercial applications, please notify me.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * Free Software Foundation, Inc.
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __IMPLICITBMT_H_INCLUDED_
#define __IMPLICITBMT_H_INCLUDED_

#include "MergeTree.h"

#ifdef _MSC_VER
#define inline __forceinline
#endif // _MSC_VER

#define MAX_HEIGHT 16

template<class T>
class ImplBinMergeTree
{
public:
//protected:
	class Reset;
	class Tables;
	class ImplMergerLoc;
	class ImplBuffer;

public:
	typedef T value_type;
	typedef ImplMergerLoc MergerLoc;
	typedef ImplBuffer Buffer;

	inline explicit ImplBinMergeTree(size_t height) : m_MergeTree(height), m_Height(height)
		{ assert( height < MAX_HEIGHT ); }

	inline void ResetMergers()
	{
		Reset r;
		m_MergeTree.EnumMergers(r);
	}
	template<class It>
	inline It empty(It b, It e)
	{
		EmptyBuffers<T,It> v(b,e);
		v = m_MergeTree.EnumMergers(v);
		return v.begin();
	}

	// Makes root locate the root
	inline void make_root(MergerLoc& root)
		{ MergerLoc::make_root(root, m_MergeTree.root(), m_Height); }

//protected:
	class ImplMerger
	{
		friend class ImplBuffer;
		friend class ImplMergerLoc;
		friend class Reset;
	public:
		inline ImplMerger(ImplMerger* pLeftChild, ImplMerger* pRightChild, T* out_buffer_begin, T* out_buffer_end)
#ifndef NDEBUG
			: m_pBegin(out_buffer_begin), m_pEnd(out_buffer_end), m_pLeft(pLeftChild), m_pRight(pRightChild)
#endif // NDEBUG
		{
			m_pHead = (T*)this;
			m_pTail = (T*)this;
			dry.bdry = false; dry.sdry.left = 0; dry.sdry.right = 0;
#ifndef NDEBUG
			if( m_pLeft )
			{
				m_pLeft->m_pParent = this;
				m_pRight->m_pParent = this;
			}
#else // NDEBUG
			(void)pLeftChild; (void)pRightChild; (void)out_buffer_begin; (void)out_buffer_end;
#endif // NDEBUG
		}
		inline void reset()
		{
			dry.bdry = false; dry.sdry.left = 0; dry.sdry.right = 0;
			m_pHead = m_pTail = (T*)this;
		}
		inline void set_root()
		{
//			m_pParent = NULL;
		}
		inline T* getHead()
			{ return m_pHead; }
		inline T* getTail()
			{ return m_pTail; }
	protected:
		T* m_pHead;
		T* m_pTail;
		union tag_union
		{
			bool bdry;
			struct s
			{
				short left;
				short right;
			} sdry;
		} dry;
#ifndef NDEBUG
		T* m_pBegin;
		T* m_pEnd;
		ImplMerger* m_pParent;
		ImplMerger* m_pLeft;
		ImplMerger* m_pRight;
#endif // NDEBUG
	};
	class ImplBuffer
	{
	public:
		inline ImplBuffer(ImplMerger* merger, size_t s) : m_pMerger(merger), m_pBegin((T*)((char*)merger-s)) { }
		inline T* begin()
			{ return m_pBegin; }
		inline const T* begin() const
			{ return m_pBegin; }
		inline T* end()
			{ return (T*)m_pMerger; }
		inline const T* end() const
			{ return (T*)m_pMerger; }
		inline T*& tail()
			{ return m_pMerger->m_pTail; }
		inline const T*& tail() const
			{ return m_pMerger->m_pTail; }
		inline T*& head()
			{ return m_pMerger->m_pHead; }
		inline const T*& head() const
			{ return m_pMerger->m_pHead; }
	private:
		ImplMerger* m_pMerger;
		T* m_pBegin; 
	};
	class ImplMergerLoc
	{
	public:
		inline ImplMergerLoc() { }
		static void make_root(ImplMergerLoc& root, ImplMerger* base, size_t height)
		{
			root.level = 0;
			root.m_bfs_index = 1;
			root.m_bfs_mask = (1<<height)-1;
			root.m_Tables.compute(height);
			root.Pos[0] = base;
		}
		inline void go_up()
		{
			m_bfs_index >>= 1;
			level--; //TreeOps++;
		}
		inline void go_left()
		{
			m_bfs_index <<= 1;
			level++; //TreeOps++;
			Pos[level] = (ImplMerger*)((char*)Pos[m_Tables.D(level)]+m_Tables.Ts(level)+(m_bfs_index & m_Tables.N(level))*m_Tables.B(level));
			assert( Pos[level] == Pos[level-1]->m_pLeft );
		}
		inline void go_right()
		{
			m_bfs_index <<= 1; m_bfs_index++;
			level++; //TreeOps++;
			Pos[level] = (ImplMerger*)((char*)Pos[m_Tables.D(level)]+m_Tables.Ts(level)+(m_bfs_index & m_Tables.N(level))*m_Tables.B(level));
			assert( Pos[level] == Pos[level-1]->m_pRight );
		}
		inline Buffer left_buffer()
			{ return Buffer((ImplMerger*)((char*)Pos[m_Tables.D(level+1)]+m_Tables.Ts(level+1)+((m_bfs_index<<1) & m_Tables.N(level+1))*m_Tables.B(level+1)), m_Tables.BS(level+1)); }
		inline Buffer right_buffer()
			{ return Buffer((ImplMerger*)((char*)Pos[m_Tables.D(level+1)]+m_Tables.Ts(level+1)+(((m_bfs_index<<1)+1) & m_Tables.N(level+1))*m_Tables.B(level+1)), m_Tables.BS(level+1)); }

		inline bool is_root() const
			{ return level == 0; }
		inline bool is_leaf() const
			{ return m_Tables.D(level+1) == (unsigned short)(-1); }
		inline unsigned int bfs_index() const
			{ return (2*m_bfs_index) & m_bfs_mask; }

		inline void set_left_exhausted()
			{ merger()->dry.sdry.left = 1; }
		inline void set_right_exhausted()
			{ merger()->dry.sdry.right = 1; }
		inline bool left_exhausted() const
			{ return merger()->dry.sdry.left != 0; }
		inline bool right_exhausted() const
			{ return merger()->dry.sdry.right != 0; }


		inline void goto_child(int i)
		{
			if( i == 0 )
				go_left();
			else if( i == 1 )
				go_right();
			else
				assert( false );
		}
		inline Buffer input_buffer(int i)
		{
			if( i == 0 )
				return left_buffer();
			else if( i == 1 )
				return right_buffer();
			else
				assert( false );
			return right_buffer();
		}


		inline bool single_stream()
			{ return false; }
		inline void single_input()
			{ }
		inline void set_dry()
			{ merger()->dry.bdry = true; }
		inline bool dry() const
			{ return merger()->dry.bdry; }
		inline void reset()
			{ merger()->reset(); }
		void assert_valid()
		{
			if( is_leaf() )
				assert( merger()->m_pLeft == NULL && merger()->m_pRight == NULL );
			else
			{
				assert( left_buffer().begin() == merger()->m_pLeft->m_pBegin );
				assert( left_buffer().end() == merger()->m_pLeft->m_pEnd );
				assert( right_buffer().begin() == merger()->m_pRight->m_pBegin );
				assert( right_buffer().end() == merger()->m_pRight->m_pEnd );
				go_left();
				go_up();
				assert( Pos[level+1] == merger()->m_pLeft );
				go_right();
				go_up();
				assert( Pos[level+1] == merger()->m_pRight );
			}
			if( level )
				assert( Pos[level-1] == merger()->m_pParent );
		}

		inline Buffer buffer()
			{ return Buffer(merger(), m_Tables.BS(level)); }
	protected:
		// Don't ever copy a merger locator
		ImplMergerLoc(const ImplMergerLoc& src) { }
		ImplMergerLoc& operator=(const ImplMergerLoc& rhs)
			{ return *this; }

		inline ImplMerger* merger() const
			{ return Pos[level]; }

		unsigned int m_bfs_index;
		unsigned int m_bfs_mask;
		int level;
		ImplMerger* Pos[MAX_HEIGHT];

		class Tables
		{
		public:
			inline Tables() { }
			void compute(size_t tree_height)
			{
				InitS();
				for( size_t i=0; i<MAX_HEIGHT; i++ )
					Tst[i] = (size_t)(-1), Nt[i] = (unsigned short)(-1), Bt[i] = (size_t)(-1), Dt[i] = (unsigned short)(-1);
				InitTBD(0, (unsigned short)tree_height);
				Bt[0] = St[tree_height];

				for( size_t i=0; i<tree_height; i++ )
				{
					Tst[i] += BSt[i];
					Bt[i] += BSt[i];
					if( i != 0 )
						assert( (1u<<(i-Dt[i]))-1 == Nt[i] );
				}
			}
			inline size_t BS(size_t level) const
				{ return BSt[level]; }
			inline size_t Ts(size_t level) const
				{ return Tst[level]; }
			inline unsigned short N(size_t level) const
				{ return Nt[level]; }
			inline size_t B(size_t level) const
				{ return Bt[level]; }
			inline unsigned short D(size_t level) const
				{ return Dt[level]; }
		protected:
			void InitS()
			{
				St[0] = (size_t)(-1);
				St[1] = sizeof(ImplMerger);
				for( size_t height=2; height<MAX_HEIGHT; height++ )
				{
					size_t top_height = BinMergeTree<T, ImplMerger>::split(height);
					size_t bottom_height = height-top_height;
					St[height] = St[top_height]+(1u<<top_height)*(BinMergeTree<T, ImplMerger>::out_buffer_size(bottom_height)*sizeof(T)+St[bottom_height]);
				}
			}
			void InitTBD(unsigned short level1, unsigned short level2)
			{
				unsigned short split = (unsigned short)BinMergeTree<T, ImplMerger>::split(level2-level1);
				if( level2-level1 > 2 )
				{
					InitTBD(level1, level1+split);
					InitTBD(level1+split, level2);
				}
				Tst[level1+split] = St[split];
				Nt[level1+split] = (1<<((unsigned short)split))-1;
				Dt[level1+split] = level1;
				Bt[level1+split] = St[level2-level1-split];
				BSt[level1+split] = BinMergeTree<T, ImplMerger>::out_buffer_size(level2-level1-split)*sizeof(T);
			}
			size_t St[MAX_HEIGHT]; // Size of a merger tree, excluding the root's output buffer
			size_t BSt[MAX_HEIGHT]; // Size of buffer
			size_t Tst[MAX_HEIGHT]; // Size of a merger toptree
			unsigned short Nt[MAX_HEIGHT]; // Number of mergers in toptree
			size_t Bt[MAX_HEIGHT]; // Size of bottomtree
			unsigned short Dt[MAX_HEIGHT]; // Depth of root of toptree
		} m_Tables;
	};

	class Reset { public: void operator()(ImplMerger* pm)
		{ pm->reset(); } };

	BinMergeTree<T, ImplMerger> m_MergeTree;
	size_t m_Height;
};

#ifdef _MSC_VER
#undef inline
#endif // _MSC_VER

#endif // __IMPLICITBMT_H_INCLUDED_
