/*
 * Copyright (C) 2003 Kristoffer Vinther.
 * For contact information, please visit
 * http://www.brics.dk/~kv/thesis/.
 * If you use this code in commercial applications, please notify me.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * Free Software Foundation, Inc.
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


template<class MergeTree, class InStream, class Allocator, class Refiller>
size_t LazyMerger<MergeTree, InStream, Allocator, Refiller>::prefered_order(_uint64_t N)
{
/*	size_t h = (size_t)floor(pow((double)N, 1.0/3)+.5);
	h = (size_t)floor(log((double)N)/(3*log(2.0))+0.5);
	return 1 << h;
	if( h > 4 )
		return (1 << h)/8;
	else
		return 1 << h;
	return (size_t)floor(pow((double)N, 1.0/3)+.5);
*/
#ifdef _CUBE_BUFFERS
	return (size_t)floor(pow((double)N, 1.0/3)+.5);
#else // _CUBE_BUFFERS
	return (size_t)floor(pow((double)N/24, 1.0/2.5)+.5);
#endif // _CUBE_BUFFERS
}

template<class MergeTree, class InStream, class Allocator, class Refiller>
_uint64_t LazyMerger<MergeTree, InStream, Allocator, Refiller>::switch_to_cache_aware()
{
#ifdef _CUBE_BUFFERS
	return 27;
#else // _CUBE_BUFFERS
	return 374;
#endif // _CUBE_BUFFERS
}

template<class MergeTree, class InStream, class Allocator, class Refiller>
void LazyMerger<MergeTree, InStream, Allocator, Refiller>::reset_input()
{
	for( InStream* p=input_streams; p<last_stream; ++p )
		p->head() = p->tail();
	last_stream = input_streams;
}

template<class MergeTree, class InStream, class Allocator, class Refiller>
void LazyMerger<MergeTree, InStream, Allocator, Refiller>::godownleft(MergerLoc& lm)
{
	while( !lm.is_leaf() )
	{
		lm.go_left();
		lm.buffer().tail() = lm.buffer().head();
		lm.buffer().head() = lm.buffer().begin();
	}
}

template<class MergeTree, class InStream, class Allocator, class Refiller>
_int64_t LazyMerger<MergeTree, InStream, Allocator, Refiller>::recwarmup(MergerLoc& lm)
{
	_int64_t min, l, r;
	if( lm.is_leaf() )
	{
		l = input_streams[lm.bfs_index()].tail()-input_streams[lm.bfs_index()].head();
		r = input_streams[lm.bfs_index()+1].tail()-input_streams[lm.bfs_index()+1].head();
		min = l<r? l : r;
		if( l == 0 )
			lm.set_left_exhausted();
		if( r == 0 )
			lm.set_right_exhausted();
		if( min )
			return min-(lm.buffer().end()-lm.buffer().begin());
		else
			return 0;
	}
	else
	{
		lm.go_right();
		r = recwarmup(lm);
		bool dry = lm.left_exhausted() && lm.right_exhausted();
		lm.go_up();
		if( dry )
			lm.set_right_exhausted();
		lm.go_left();
		l = recwarmup(lm);
		dry = lm.left_exhausted() && lm.right_exhausted();
		lm.go_up();
		if( dry )
			lm.set_left_exhausted();
		lm.buffer().head() = lm.buffer().tail() = lm.buffer().end();
		min = l<r? l : r;
		if( min > 0 )
			return min-(lm.buffer().end()-lm.buffer().begin());
		else
			return 0;
	}
}

template<class MergeTree, class InStream, class Allocator, class Refiller>
bool LazyMerger<MergeTree, InStream, Allocator, Refiller>::warmup()
{
	MergerLoc lm;
	m_MergeTree.make_root(lm);
	lm.go_right();
	_int64_t r = recwarmup(lm);
	lm.go_up();
	lm.go_left();
	_int64_t l = recwarmup(lm);
	lm.go_up();
	if( l > 0 && r > 0 )
		return true;
	else
		return false;
}
