/*
 * Copyright (C) 2003 Kristoffer Vinther.
 * For contact information, please visit
 * http://www.brics.dk/~kv/thesis/.
 * If you use this code in commercial applications, please notify me.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * Free Software Foundation, Inc.
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __POINTERBMT_H_INCLUDED_
#define __POINTERBMT_H_INCLUDED_

#include "MergeTree.h"

#ifdef _MSC_VER
#define inline __forceinline
#endif // _MSC_VER

template<class T>
class PBBinMergeTree
{
//protected:
public:
	class Reset;
	class PBMerger;
	class PBMergerLoc;
	class PBBuffer;
public:
	typedef T value_type;
	typedef PBMergerLoc MergerLoc;
	typedef PBBuffer Buffer;

	inline explicit PBBinMergeTree(size_t height) : m_MergeTree(height) { }

	inline void ResetMergers()
	{
		Reset r;
		m_MergeTree.EnumMergers(r);
	}

	template<class It>
	inline It empty(It b, It e)
	{
		EmptyBuffers<T,It> v(b,e);
		v = m_MergeTree.EnumMergers(v);
		return v.begin();
	}
	inline void make_root(MergerLoc& root)
		{ root = MergerLoc(m_MergeTree.root()); }

//protected:
	class Reset { public: inline void operator()(PBMerger* pm)
		{ pm->reset(); } };

	class PBMerger
	{
		friend class PBBuffer;
		friend class PBMergerLoc;
		friend class Reset;
	public:
		inline PBMerger(PBMerger* pLeftChild, PBMerger* pRightChild, T* out_buffer_begin, T* out_buffer_end) :
		  m_pHead(out_buffer_end), m_pTail(out_buffer_end), m_pBegin(out_buffer_begin), /*m_pEnd(out_buffer_end), */m_pLeft(pLeftChild), m_pRight(pRightChild)
		{
			dry.bdry = false; dry.sdry.left = 0; dry.sdry.right = 0;
			if( m_pLeft )
			{
				m_pLeft->m_pParent = this;
				m_pRight->m_pParent = this;
			}
			else
				assert( m_pRight == NULL );
		}
		void assert_valid()
		{
			if( m_pParent )
			{
//				assert( (void*)m_pEnd == (void*)this );
				assert( m_pParent->m_pLeft == this || m_pParent->m_pRight == this );
			}
			if( m_pLeft )
			{
				assert( m_pRight );
				assert( m_pLeft->m_pParent == this && m_pRight->m_pParent == this );
			}
			else
				assert( m_pRight == NULL );
		}
		inline void set_root()
			{ m_pParent = NULL; }
		inline void reset()
		{
			dry.bdry = false; dry.sdry.left = 0; dry.sdry.right = 0;
			if( m_pParent )
				m_pHead = m_pTail = (T*)this;
		}
		inline T* getHead()
			{ return m_pHead; }
		inline T* getTail()
			{ return m_pTail; }

		static size_t height()
			{ return 1; }
	protected:
		T* m_pHead;
		T* m_pTail;
		T* m_pBegin;
//		T* m_pEnd;
		union tag_union
		{
			bool bdry;
			struct s
			{
				short left;
				short right;
			} sdry;
		} dry;

		PBMerger* m_pParent;
		PBMerger* m_pLeft;
		PBMerger* m_pRight;
	};
	class PBBuffer
	{
	public:
		inline PBBuffer(PBMerger* merger) : m_pMerger(merger)
			{ }
		inline T* begin()
			{ return m_pMerger->m_pBegin; }
		inline const T* begin() const
			{ return m_pMerger->m_pBegin; }
		inline T* end()
			{ return (T*)m_pMerger/*->m_pEnd*/; }
		inline const T* end() const
			{ return (T*)m_pMerger/*->m_pEnd*/; }
		inline T*& tail()
			{ return m_pMerger->m_pTail; }
		inline const T*& tail() const
			{ return m_pMerger->m_pTail; }
		inline T*& head()
			{ return m_pMerger->m_pHead; }
		inline const T*& head() const
			{ return m_pMerger->m_pHead; }
	private:
		PBMerger* m_pMerger;
	};

	class PBMergerLoc
	{
	public:
		inline PBMergerLoc() : m_pMerger(NULL), m_bfs_index(0) { }
		inline explicit PBMergerLoc(PBMerger* pm) : m_pMerger(pm), m_bfs_index(0) { }
		inline void go_up()
			{ m_pMerger = m_pMerger->m_pParent; m_bfs_index/=2; /*TreeOps++;*/ }
		inline void go_left()
			{ m_pMerger = m_pMerger->m_pLeft; m_bfs_index*=2; /*TreeOps++;*/ }
		inline void go_right()
			{ m_pMerger = m_pMerger->m_pRight; m_bfs_index=2*m_bfs_index+1; /*TreeOps++;*/ }
		inline Buffer left_buffer()
			{ return Buffer(m_pMerger->m_pLeft); }
		inline Buffer right_buffer()
			{ return Buffer(m_pMerger->m_pRight); }
		inline bool is_root() const
			{ return m_pMerger->m_pParent == NULL; }
		inline bool is_leaf() const
			{ if( m_pMerger->m_pLeft == NULL ) assert( m_pMerger->m_pRight == NULL || m_pMerger->m_pRight == m_pMerger ); return m_pMerger->m_pLeft == NULL; }
		inline bool single_stream()
			{ return m_pMerger->m_pRight == m_pMerger; }
		inline void set_dry()
			{ m_pMerger->dry.bdry = true; }
		inline bool dry() const
			{ return m_pMerger->dry.bdry; }
		inline unsigned int bfs_index() const
			{ return 2*m_bfs_index; }

		inline void assert_valid()
			{ m_pMerger->assert_valid(); }
		inline void single_input()
			{ assert( is_leaf() ); m_pMerger->m_pRight = m_pMerger; }

		inline Buffer buffer()
			{ return Buffer(m_pMerger); }


		inline void goto_child(int i)
		{
			if( i == 0 )
				go_left();
			else if( i == 1 )
				go_right();
			else
				assert( false );
		}
		inline Buffer input_buffer(int i)
		{
			if( i == 0 )
				return left_buffer();
			else if( i == 1 )
				return right_buffer();
			else
				assert( false );
			return right_buffer();
		}

		inline void set_left_exhausted()
			{ m_pMerger->dry.sdry.left = 1; }
		inline void set_right_exhausted()
			{ m_pMerger->dry.sdry.right = 1; }
		inline bool left_exhausted() const
			{ return m_pMerger->dry.sdry.left != 0; }
		inline bool right_exhausted() const
			{ return m_pMerger->dry.sdry.right != 0; }

		inline void reset()
			{ m_pMerger->reset(); }
	protected:
		PBMerger* m_pMerger;
		unsigned int m_bfs_index;
	};

	BinMergeTree<T, PBMerger> m_MergeTree;
};

#ifdef _MSC_VER
#undef inline
#endif // _MSC_VER

#endif // __POINTERBMT_H_INCLUDED_
