/*
 * Copyright (C) 2003 Kristoffer Vinther.
 * For contact information, please visit
 * http://www.brics.dk/~kv/thesis/.
 * If you use this code in commercial applications, please notify me.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * Free Software Foundation, Inc.
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


/*
#include <iostream>
#include <vector>

class basestat
{
public:
	basestat() : sort_count(2048, 0)
		{ }
	~basestat()
	{
        for( size_t i=0; i<sort_count.size(); i++ )
			if( sort_count[i] )
				std::cout << (unsigned long)i << ';' << (unsigned long)sort_count[i] << std::endl;
	}
	void inc(size_t i)
	{
		if( sort_count.size() < i+1 )
			sort_count.resize(i+1);
		sort_count[i]++;
	}
protected:
	std::vector<size_t> sort_count;
};
*/

#include <iomanip>

template<class FwdIt>
void __dump(FwdIt b, FwdIt e)
{
	std::cerr << '[';
	if( b != e )
		std::cerr << std::setw(5) << *b++;
	for( FwdIt i=b; i!=e; i++ )
		std::cerr << ' ' << std::setw(5) << *i;
	std::cerr << ']' << std::endl;
}

#include <iterator>

template<class FwdIt>
void ___dump(FwdIt b, FwdIt e)
{
	std::cerr << '[';
	std::copy(card.begin(), card.end(), std::ostream_iterator<int>(std::cerr, ","));
	std::cerr << ']' << std::endl;
}


namespace iosort
{

// Generic comparison function for qsort()
template<class T>
int __qsort_comp(const void* a, const void* b)
{
	if( *(T*)b < *(T*)a )
		return 1;
	else if( *(T*)a < *(T*)b )
		return -1;
	else
		return 0;
}

// Generic wrapper for qsort()
template<class RanIt, class T>
inline void gen_qsort(RanIt begin, RanIt end, T& dummy)
{
	(void)dummy;
	int (*compare)(const void *, const void *) = __qsort_comp<T>;
	qsort(&(*begin), (size_t)(end-begin), sizeof(*begin), compare);
}


namespace base
{

enum tag_RECURSELIMIT { RECURSELIMIT = 32 };

template<int h, class It, class T>
inline void insert_step(It begin, It i, T e)
{
	while( begin <= i )
		if( e < *i )
		{
            *(i+h) = *i;
			i-=h;
		}
		else
			break;
	*(i+h) = e;
}

template<class It>
void shell_sort(It begin, It end)
{
	for( It j=begin+48; j<end; j++ )
		insert_step<48>(begin, j-48, *j);
	for( It j=begin+21; j<end; j++ )
		insert_step<21>(begin, j-21, *j);
	for( It j=begin+7; j<end; j++ )
		insert_step<7>(begin, j-7, *j);
	for( It j=begin+3; j<end; j++ )
		insert_step<3>(begin, j-5, *j);
	for( It j=begin+1; j<end; j++ )
		insert_step<1>(begin, j-1, *j);
}

template<class It>
void insertion_sort(It begin, It end)
{
	for( It j=begin+1; j<end; j++ )
		insert_step<1>(begin, j-1, *j);
}

template<class T, class It>
void __selection_sort(It begin, It end)
{
	for( ; begin!=end; begin++ )
	{
		It min = begin;
		for( It j=begin+1; j!=end; j++ )
			if( *j < *min )
			{
				min = j;
				for( ; j!=end; j++ )
					if( *j < *min )
						min = j;
				std::iter_swap(min, begin);
				break;
			}
	}
}
template<class T, class It>
inline void __selection_sort(It begin, It end, T)
{
	__selection_sort<T, It>(begin, end);
}
template<class It>
inline void selection_sort(It begin, It end)
{
	__selection_sort(begin, end, *begin);
}

template<class It, class ItDiff>
void __batcher_sort(It begin, It end, ItDiff size)
{
	size_t t = _clog(size)-1;
	ItDiff p = 1<<t;
	for( ; p != 1; p=p/2 )
	{
		It i, j;
		for( i=begin; i<end-p-p; i+=p )
		{
			j = i+p;
			for( ; i<j; i++ )
				if( *(i+p) < *i )
					std::iter_swap(i+p, i);
		}
		for( ; i<end-p; i++ )
			if( *(i+p) < *i )
				std::iter_swap(i+p, i);
		ItDiff q = 1<<t;
		while( q != p )
		{
			ItDiff d=q-p;
			q = q/2;
			for( i=begin+p; i<end-d-p; i+=p )
			{
				j = i+p;
				for( ; i<j; i++ )
					if( *(i+d) < *i )
						std::iter_swap(i+d, i);
			}
			for( ; i<end-d; i++ )
				if( *(i+d) < *i )
					std::iter_swap(i+d, i);
		}
	}
	if( p )
	{
		for( It i=begin; i<end-1; i+=2 )
			if( *(i+1) < *i )
				std::iter_swap(i+1, i);
		ItDiff q = 1<<t;
		while( q != 1 )
		{
			ItDiff d=q-1;
			q = q/2;
			for( It i=begin+1; i<end-d; i+=2 )
				if( *(i+d) < *i )
					std::iter_swap(i+d, i);
		}
	}
}
template<class It>
inline void batcher_sort(It begin, It end)
{
	__batcher_sort(begin, end, end-begin);
}


template<class _RanIt>
inline void __med3(_RanIt _First, _RanIt _Mid, _RanIt _Last)
{
	if( *_Mid < *_First )
		std::iter_swap(_Mid, _First);
	if( *_Last < *_Mid )
		std::iter_swap(_Last, _Mid);
	if( *_Mid < *_First )
		std::iter_swap(_Mid, _First);
}

template<class It, int SampleSize>
inline It __find_rank(It first, It last, int rank)
{
	assert( 0 <= rank && rank < SampleSize );
	if( last-first < SampleSize )
	{
		It mid = first+(last-first)/2;
		__med3(first, mid, last-1);
		return mid;
	}
	else if( last-first < SampleSize*(SampleSize/4) )
	{
		It mid = first+(last-first)/2;
		unsigned long step = (unsigned long)(last-first)/8;
		__med3(first, first+step, first+2*step);
		__med3(mid-step, mid, mid+step);
		__med3(last-2*step-1, last-step-1, last-1);
		__med3(first+step, mid, last-step-1);
		return mid;
	}
	else
	{
		It s1 = first;
		It s2 = s1+SampleSize/3;
		It e2 = last-1;
		It e1 = e2-SampleSize/3;
		It m1 = first+(last-first)/2;
		It m2 = m1+(SampleSize-(s2-s1)-(e2-e1))/2;
		m1 = m1-(SampleSize-(s2-s1)-(e2-e1)-(m2-m1));
		assert( SampleSize < last-first );
		assert( (s2-s1)+(e2-e1)+(m2-m1) == SampleSize );

/*		It i=s1, j=first;
		while( i < last )
		{
			std::iter_swap(i, j);
			if( i == m2 )
				i = e1;
			else if( i == s2 )
				i = m1;
			else
				i++;
		}
		__inplace_base_sort(first, first+SampleSize);
*/
		It p = s1;
		if( rank < SampleSize/2 )
		{
			for( ;; )
			{
				It min = p;
				It i=p+1;
				if( i == m2 )
					i = e1;
				else if( p == s2 )
					i = m1;
				while( i < last )
				{
					if( *i < *min )
						min = i;
					if( i == m2 )
						i = e1;
					else if( i == s2 )
						i = m1;
					else
						i++;
				}
				if( rank-- == 0 )
					return min;
				std::iter_swap(min, p);
				if( p == m2 )
					p = e1;
				else if( p == s2 )
					p = m1;
				else
					p++;
			}
		}
		else
		{
			for( ;; )
			{
				It max = p;
				It i=p+1;
				if( i == m2 )
					i = e1;
				else if( p == s2 )
					i = m1;
				while( i < last )
				{
					if( *i > *max )
						max = i;
					if( i == m2 )
						i = e1;
					else if( i == s2 )
						i = m1;
					else
						i++;
				}
				if( ++rank == SampleSize )
					return max;
				std::iter_swap(max, p);
				if( p == m2 )
					p = e1;
				else if( p == s2 )
					p = m1;
				else
					p++;
			}
		}
	}
}

template<class It, class T, class Pred>
inline It __partition(It first, It last, T pivot, Pred comp)
{
	for( ;; )
	{
		while( comp(*first,pivot) )
			first++;
		last--;
		while( comp(pivot,*last) )
			last--;
		if( !(first<last) )
			return first;
		std::iter_swap(first, last);
		first++;
	}
}

template<class It>
inline	It __partition_good(It first, It last, It goodbegin, It goodend)
{
	return __partition_good(first, last, goodbegin, goodend, *first);
}

template<class It, class T>
inline	It __partition_good(It first, It last, It goodbegin, It goodend, T)
{
	assert( first <= goodbegin && goodbegin <= goodend && goodend <= last );
	It ideal = goodbegin+(goodend-goodbegin)/2;
	for( ;; )
	{
		const int SampleSize = 32;
		int rank = SampleSize-1-(int)(((_int64_t)SampleSize*(ideal-first))/(last-first)); // TODO: Find a better rank, using binomials.
		It r = __find_rank<It, SampleSize>(first, last, rank);
		It p = __partition(first, last, *r, std::greater<T>());
#ifndef NDEBUG
		It i;
		for( i=first; i!=p; i++ )
			assert( !(*i < *p) );
		for( ; i!=last; i++ )
			assert( !(*(p-1) < *i) );
#endif // NDEBUG
		if( p < goodbegin )
			first = p;
		else if( p > goodend )
			last = p;
		else
			return p;
#ifdef _DEBUG
		std::cerr << '!';
#endif // _DEBUG
	}
}

template<class _RanIt, class _Diff, class T>
inline void __intro_sort(_RanIt _First, _RanIt _Last, _Diff _Ideal, T)
{
	__intro_sort<_RanIt, _Diff, T>(_First, _Last, _Ideal);
}

template<class _RanIt, class _Diff, class T>
void __intro_sort(_RanIt _First, _RanIt _Last, _Diff _Ideal)
{
	_Diff _Count;
	while( (_Count = _Last-_First) > RECURSELIMIT && _Ideal > 0 )
	{
		_RanIt r = __find_rank<_RanIt, 8>(_First, _Last, 4);
		_RanIt _Mid = __partition(_First, _Last, *r, std::less<T>());
		_Ideal /= 2, _Ideal += _Ideal / 2; // allow 1.5 log2(N) divisions

		if( _Mid-_First < _Last-_Mid ) // loop on larger half
			__intro_sort<_RanIt, _Diff, T>(_First, _Mid, _Ideal), _First = _Mid;
		else
			__intro_sort<_RanIt, _Diff, T>(_Mid, _Last, _Ideal), _Last = _Mid;
	}

	if( _Count > RECURSELIMIT )
	{
		std::make_heap(_First, _Last);
		std::sort_heap(_First, _Last);
	}
	else if( _Count > 1 )
		__inplace_base_sort(_First, _Last);
}


// Sorts the range [begin, end) inplace efficiently when the range is small
template<class It>
inline void __inplace_base_sort(It begin, It end)
{
/*	if( end-begin < RECURSELIMIT )
		selection_sort(begin, end);
	else
		shell_sort(begin, end);
//		batcher_sort(begin, end);
//		__intro_sort(begin, end, end-begin, *begin);
*/
#ifdef __ia64__
	batcher_sort(begin, end);
#else // __ia64__
	std::sort(begin, end);
#endif // __ia64__
}

// Copies the range [begin, end) to dest in sorted order. Efficient only when the range is small
template<class InIt, class OutIt>
inline void __base_sort(InIt begin, InIt end, OutIt dest)
{
/*	if( end-begin < RECURSELIMIT )
	{
		for( InIt i=begin; i!=end; i++, dest++ )
		{
			InIt min = i;
			for( InIt j=i+1; j!=end; j++ )
				if( *j < *min )
					min = j;
			*dest = *min;
			*min = *i;
			*i = *dest;
		}
	}
	else
*/	{
//		shell_sort(begin, end);
//		batcher_sort(begin, end);
//		__intro_sort(begin, end, end-begin, *begin);
#ifdef __ia64__
		batcher_sort(begin, end);
#else // __ia64__
		std::sort(begin, end);
#endif // __ia64__
		std::copy(begin, end, dest);
	}
}

} // namespace base

} // namespace iosort
