/*
 * Copyright (C) 2003 Kristoffer Vinther.
 * For contact information, please visit
 * http://www.brics.dk/~kv/thesis/.
 * If you use this code in commercial applications, please notify me.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * Free Software Foundation, Inc.
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __BASIC_MERGER_H_INCLUDED_
#define __BASIC_MERGER_H_INCLUDED_

#include "../../FArray.h"
#include "funnel_base.h"
#include <functional>
#include <iterator>

#ifdef __GNUC__
#define inline inline __attribute__ ((always_inline))
#endif // __GNUC__
#ifdef _MSC_VER
#define inline __forceinline
#endif // _MSC_VER


namespace funnel
{

template<basic_order_t order>
class basic_merger_stat
{
public:
	basic_merger_stat()
	{
		for( int i=0; i<order+1; ++i )
			act[i] = 0;
		for( int i=0; i<101; ++i )
			shift[i] = 0;
	}
	~basic_merger_stat()
	{
/*		std::cerr << '{';
		for( int i=0; i<order+1; ++i )
			std::cerr << act[i] << ',';
		std::cerr << '}';
*/		std::cerr << std::endl;
		for( int i=0; i<101; ++i )
			std::cerr << shift[i] << std::endl;
		std::cerr << std::endl;
	}
	void active(int a)
		{ ++act[a]; }
	template<class Diff>
	void inps(Diff a, Diff b)
	{ 
		Diff max, min = a;
		if( b < min )
			min = b, max = a;
		else
			max = b;
		++shift[100*min/max];
	}
protected:
	int act[order+1];
	int shift[101];
};


template<class T1, class T2>
class CCHECK
{
public:
	static inline T1 ITERATOR_DIFFERENCES_MUST_BE_COMPATIBLE(T1 x, T1) { return x; }
	static inline T2 ITERATOR_DIFFERENCES_MUST_BE_COMPATIBLE(T2 x, T2) { return x; }
};
template<class T>
class CCHECK<T,T>
{
public:
	static inline T ITERATOR_DIFFERENCES_MUST_BE_COMPATIBLE(T x, T) { return x; }
};

template<basic_order_t order, class Stream, class Token>
class basic_merger_base
{
public:
	inline basic_merger_base() : active(0) { }

    inline void add_stream(Stream *s, Token t)
	{
		if( s->begin() != s->end() )
		{
			stream[active] = s;
			token[active] = t;
			++active;
		}
	}

protected:
	template<class It>
	inline Token copy(It& b, It e, Token outtoken)
	{
		if( stream[0]->end()-stream[0]->begin() > e-b )
		{
			typename Stream::pointer p;
			for( p = stream[0]->begin(); b!=e; ++b, ++p )
				*b = *p;
			*stream[0] = Stream(p, stream[0]->end());
			return outtoken;
		}
		else
		{
			for( typename Stream::pointer p = stream[0]->begin(); p!=stream[0]->end(); ++b, ++p )
				*b = *p;
			*stream[0] = Stream(stream[0]->end(), stream[0]->end());
			return token[0];
		}
	}

	basic_order_t active;
	Stream *stream[order];
	Token token[order];
};



template<basic_order_t order, class Stream, class Token>
class simple_merger : public basic_merger_base<order,Stream,Token>
{
	using basic_merger_base<order,Stream,Token>::active;
	using basic_merger_base<order,Stream,Token>::stream;
	using basic_merger_base<order,Stream,Token>::token;
	using basic_merger_base<order,Stream,Token>::copy;
public:
	template<basic_order_t O, class S=Stream, class T=Token>
	struct rebind { typedef simple_merger<O,S,T> other; };
	///<summary>Invokes the merger</summary>
	template<class It>
	inline Token operator()(It& b, It e, Token outtoken)
	{
		if( active == 0 )
			return outtoken;
		if( active == 1 )
			return copy(b,e,outtoken);
		return invoke_invoke(b,e,outtoken,e-b,*b,*stream[0]->begin());
	}
	///<summary>Invokes the merger</summary>
	template<class It>
	inline Token operator()(It& b, Token outtoken)
	{
		if( active == 0 )
			return outtoken;
		if( active == 1 )
		{
			b = std::copy(stream[0]->begin(),stream[0]->end(),b);
			*stream[0] = Stream(stream[0]->end(), stream[0]->end());
			return token[0];
		}
		return invoke_invoke(b,*b,*stream[0]->begin());
	}
	///<summary>Invokes the merger</summary>
	template<class It, class Comp>
	inline Token operator()(It& b, It e, Token outtoken, Comp comp)
	{
		if( active == 0 )
			return outtoken;
		if( active == 1 )
			return copy(b,e,outtoken);
		return invoke_invoke(b,e,outtoken,comp,*b,*stream[0]->begin());
	}

private:
	template<class It, class T, class Diff>
	inline Token invoke_invoke(It& b, It e, Token outtoken, Diff min, T, T)
		{ std::less<T> comp; return invoke<It,T>(b,e,outtoken,min,comp); }
	template<class It, class Diff, class Comp, class T>
	inline Token invoke_invoke(It& b, It e, Token outtoken, Diff min, Comp& comp, T, T)
		{ return invoke<It,T>(b,e,outtoken,min,comp); }
	template<class It, class T>
	inline Token invoke_invoke(It& b, T, T)
		{ std::less<T> comp; return invoke<It>(b,comp); }
	template<class It, class Comp>
	inline Token invoke(It& b, Comp& comp)
	{
		typename Stream::pointer head[2] = { stream[0]->begin(), stream[1]->begin() };
		typename Stream::pointer tail[2] = { stream[0]->end(), stream[1]->end() };
		for( ;; )
		{
			if( comp(*head[0],*head[1]) )
			{
				*b = *head[0], ++head[0], ++b;
				if( head[0] == tail[0] )
				{
					*stream[0] = Stream(head[0],stream[0]->end());
					*stream[1] = Stream(head[1],stream[1]->end());
					return token[0];
				}
			}
			else
			{
				*b = *head[1], ++head[1], ++b;
				if( head[1] == tail[1] )
				{
					*stream[0] = Stream(head[0],stream[0]->end());
					*stream[1] = Stream(head[1],stream[1]->end());
					return token[1];
				}
			}
		}
	}
	template<class It, class T, class Diff, class Comp>
	inline Token invoke(It& b, It e, Token outtoken, Diff, Comp& comp)
	{
		typename Stream::pointer head[2] = { stream[0]->begin(), stream[1]->begin() };
		typename Stream::pointer tail[2] = { stream[0]->end(), stream[1]->end() };
		It p = b;
		while( p != e )
		{
			if( comp(*head[0],*head[1]) )
			{
				*p = *head[0], ++head[0], ++p;
				if( head[0] == tail[0] )
				{
					outtoken = token[0];
					break;
				}
			}
			else
			{
				*p = *head[1], ++head[1], ++p;
				if( head[1] == tail[1] )
				{
					outtoken = token[1];
					break;
				}
			}
		}
		*stream[0] = Stream(head[0],stream[0]->end());
		*stream[1] = Stream(head[1],stream[1]->end());
		b = p;
		return outtoken;
/*		Diff i = b-e;
		while( i )
		{
			if( comp(*head[0],*head[1]) )
			{
				*(e+i) = *head[0], ++head[0], ++i;
				if( head[0] == tail[0] )
				{
					outtoken = token[0];
					break;
				}
			}
			else
			{
				*(e+i) = *head[1], ++head[1], ++i;
				if( head[1] == tail[1] )
				{
					outtoken = token[1];
					break;
				}
			}
		}
		*stream[0] = Stream(head[0],stream[0]->end());
		*stream[1] = Stream(head[1],stream[1]->end());
		b = e+i;
		return outtoken;
*/	}
};

template<basic_order_t order, class Stream, class Token>
class two_merger : public basic_merger_base<order,Stream,Token>
{
	using basic_merger_base<order,Stream,Token>::active;
	using basic_merger_base<order,Stream,Token>::stream;
	using basic_merger_base<order,Stream,Token>::token;
	using basic_merger_base<order,Stream,Token>::copy;
public:
	template<basic_order_t O, class S=Stream, class T=Token>
	struct rebind { typedef two_merger<O,S,T> other; };
	///<summary>Invokes the merger</summary>
	template<class It>
	inline Token operator()(It& b, It e, Token outtoken)
	{
		if( active == 0 )
			return outtoken;
		if( active == 1 )
			return copy(b,e,outtoken);
		return invoke_invoke(b,e,outtoken,CCHECK<typename std::iterator_traits<It>::difference_type,typename std::iterator_traits<typename Stream::pointer>::difference_type>::ITERATOR_DIFFERENCES_MUST_BE_COMPATIBLE(e-b,stream[0]->end()-stream[0]->end()),*b,*stream[0]->begin());
	}
	///<summary>Invokes the merger</summary>
	template<class It, class Comp>
	inline Token operator()(It& b, It e, Token outtoken, Comp comp)
	{
		if( active == 0 )
			return outtoken;
		if( active == 1 )
			return copy(b,e,outtoken);
		return invoke_invoke(b,e,outtoken,comp,*b,*stream[0]->begin());
	}

private:
	class assign
	{
	public:
		assign(Token& lhs, Token rhs) : lhs(lhs), rhs(rhs) { }
		template<class It>
		bool operator()(It,It)
			{ lhs = rhs; return false; }
		void operator()()
			{ lhs = rhs; }
	protected:
		void operator=(const assign&)
			{ }
		Token& lhs;
		Token rhs;
	};

	template<class It, class T, class Diff>
	inline Token invoke_invoke(It& b, It e, Token outtoken, Diff min, T, T)
		{ std::less<T> comp; return invoke<It,T>(b,e,outtoken,min,comp); }
	template<class It, class Diff, class Comp, class T>
	inline Token invoke_invoke(It& b, It e, Token outtoken, Diff min, Comp& comp, T, T)
		{ return invoke<It,T>(b,e,outtoken,min,comp); }
	template<class It, class T, class Diff, class Comp>
	inline Token invoke(It& b, It e, Token outtoken, Diff min, Comp& comp)
	{
		It p = b;
		typename Stream::pointer head[2] = { stream[0]->begin(), stream[1]->begin() };
		if( stream[0]->end()-head[0] < min )
			min = stream[0]->end()-head[0];
		if( stream[1]->end()-head[1] < min )
			min = stream[1]->end()-head[1];
		do
		{
			for( ; min; --min )
			{
				if( comp(*head[0],*head[1]) )
					*p = *head[0], ++head[0];
				else
					*p = *head[1], ++head[1];
				++p;
			}
			min = stream[0]->end()-head[0];
			if( stream[1]->end()-head[1] < min )
				min = stream[1]->end()-head[1];
			if( e-p < min )
				min = e-p;
		}
		while( min );
		*stream[0] = Stream(head[0],stream[0]->end());
		*stream[1] = Stream(head[1],stream[1]->end());
		b = p;
		if( stream[0]->end() == head[0] )
			return token[0];
		if( stream[1]->end() == head[1] )
			return token[1];
		return outtoken;
/*		Token t;
		assign a1(t, token[0]), a2(t, token[1]), a0(t, token[order]);
		typename Stream::pointer h1 = stream[0]->begin();
		typename Stream::pointer h2 = stream[1]->begin();
//		LazyMerge(h1,stream[0]->end(),h2,stream[1]->end(),b,e,0,a1,a2,a0);
		_lazy_merge(h1,stream[0]->end(),h2,stream[1]->end(),b,e,a1,a2,a0,e-b);
		*stream[0] = Stream(h1,stream[0]->end());
		*stream[1] = Stream(h2,stream[1]->end());
		return t;
*/	}
};

template<basic_order_t order, class Stream, class Token>
class hl_merger : public basic_merger_base<order,Stream,Token>
{
	using basic_merger_base<order,Stream,Token>::active;
	using basic_merger_base<order,Stream,Token>::stream;
	using basic_merger_base<order,Stream,Token>::token;
	using basic_merger_base<order,Stream,Token>::copy;
public:
	template<basic_order_t O, class S=Stream, class T=Token>
	struct rebind { typedef hl_merger<O,S,T> other; };
	///<summary>Invokes the merger</summary>
	template<class It>
	inline Token operator()(It& b, It e, Token outtoken)
	{
		if( active == 0 )
			return outtoken;
		if( active == 1 )
			return copy(b,e,outtoken);
		return invoke_invoke(b,e,outtoken,e-b,*b,*stream[0]->begin());
	}
	///<summary>Invokes the merger</summary>
	template<class It, class Comp>
	inline Token operator()(It& b, It e, Token outtoken, Comp comp)
	{
		if( active == 0 )
			return outtoken;
		if( active == 1 )
			return copy(b,e,outtoken);
		return invoke_invoke(b,e,outtoken,comp,*b,*stream[0]->begin());
	}

private:
	/// returns It i, st. either i == b or for all s in [b;i): comp(s,t)
	template<class It, class T, class Comp>
	inline It bin_search(It b, It e, T t, Comp& comp)
	{
		assert(!( b > e ));
		if( comp(t,*b) )
			return b;
		if( comp(*(e-1),t) )
			return e;
		while( e-b != 1 )
		{
			It m = b+(e-b)/2;
			if( comp(*m,t) )
				b = m;
			else
				e = m;
		}
		return e;
	}
	template<class It, class OutIt>
	inline OutIt lcopy(It b, It e, OutIt p)
	{
		for( ; b!=e; ++p, ++b )
			*p = *b;
		return p;
	}
	template<class It, class T, class Diff>
	inline Token invoke_invoke(It& b, It e, Token outtoken, Diff min, T, T)
		{ std::less<T> comp; return invoke<It,T>(b,e,outtoken,min,comp); }
	template<class It, class Diff, class Comp, class T>
	inline Token invoke_invoke(It& b, It e, Token outtoken, Diff min, Comp& comp, T, T)
		{ return invoke<It,T>(b,e,outtoken,min,comp); }
	template<class It, class T, class Diff, class Comp>
	inline Token invoke(It& b, It e, Token outtoken, Diff, Comp& comp)
	{
		typename Stream::pointer h[2] = { stream[0]->begin(), stream[1]->begin() };
		typename Stream::pointer const t[2] = { stream[0]->end(), stream[1]->end() };
		It p = b;

		for( ;; )
		{
			if( t[0]-h[0] < t[1]-h[1] )
			{
				typename Stream::pointer s = h[1]+(t[1]-h[1])/(t[0]-h[0]);
				s = bin_search(h[1],s,*h[0],comp);
				if(!( s-h[1] < e-p ))
				{
					s = h[1]+(e-p);
					if( s == t[1] )
						outtoken = token[1];
					p = lcopy(h[1],s,p);
					h[1] = s;
					break;
				}
				p = lcopy(h[1],s,p);
				if( s == h[1] )
				{
					*p = *h[0], ++p, ++h[0];
					if( h[0] == t[0] )
					{
						outtoken = token[0];
						break;
					}
				}
				else
				{
					h[1] = s;
					if( h[1] == t[1] )
					{
						outtoken = token[1];
						break;
					}
				}
			}
			else
			{
				typename Stream::pointer s = h[0]+(t[0]-h[0])/(t[1]-h[1]);
				s = bin_search(h[0],s,*h[1],comp);
				if(!( s-h[0] < e-p ))
				{
					s = h[0]+(e-p);
					if( s == t[0] )
						outtoken = token[0];
					p = lcopy(h[0],s,p);
					h[0] = s;
					break;
				}
				p = lcopy(h[0],s,p);
				if( s == h[0] )
				{
					*p = *h[1], ++p, ++h[1];
					if( h[1] == t[1] )
					{
						outtoken = token[1];
						break;
					}
				}
				else
				{
					h[0] = s;
					if( h[0] == t[0] )
					{
						outtoken = token[0];
						break;
					}
				}
			}
			if( p == e )
				break;
		}
		*stream[0] = Stream(h[0],stream[0]->end());
		*stream[1] = Stream(h[1],stream[1]->end());
		b = p;
		return outtoken;
	}
};



template<basic_order_t order, class Stream, class Token>
class hyb_merger : public basic_merger_base<order,Stream,Token>
{
	using basic_merger_base<order,Stream,Token>::active;
	using basic_merger_base<order,Stream,Token>::stream;
	using basic_merger_base<order,Stream,Token>::token;
	using basic_merger_base<order,Stream,Token>::copy;
public:
	template<basic_order_t O, class S=Stream, class T=Token>
	struct rebind { typedef hyb_merger<O,S,T> other; };
	///<summary>Invokes the merger</summary>
	template<class It>
	inline Token operator()(It& b, It e, Token outtoken)
	{
		if( active == 0 )
			return outtoken;
		if( active == 1 )
			return copy(b,e,outtoken);
		return invoke_invoke(b,e,outtoken,CCHECK<typename std::iterator_traits<It>::difference_type,typename std::iterator_traits<typename Stream::pointer>::difference_type>::ITERATOR_DIFFERENCES_MUST_BE_COMPATIBLE(e-b,stream[0]->end()-stream[0]->end()),*b,*stream[0]->begin());
	}
	///<summary>Invokes the merger</summary>
	template<class It, class Comp>
	inline Token operator()(It& b, It e, Token outtoken, Comp comp)
	{
		if( active == 0 )
			return outtoken;
		if( active == 1 )
			return copy(b,e,outtoken);
		return invoke_invoke(b,e,outtoken,comp,*b,*stream[0]->begin());
	}

private:
	template<class It, class T, class Diff>
	inline Token invoke_invoke(It& b, It e, Token outtoken, Diff min, T, T)
		{ std::less<T> comp; return invoke<It,T>(b,e,outtoken,min,comp); }
	template<class It, class Diff, class Comp, class T>
	inline Token invoke_invoke(It& b, It e, Token outtoken, Diff min, Comp& comp, T, T)
		{ return invoke<It,T>(b,e,outtoken,min,comp); }
	template<class It, class T, class Diff, class Comp>
	inline Token invoke(It& b, It e, Token outtoken, Diff, Comp& comp)
	{
		typename Stream::pointer head[2] = { stream[0]->begin(), stream[1]->begin() };
		typename Stream::pointer const tail[2] = { stream[0]->end(), stream[1]->end() };
		Diff min = tail[0]-head[0];
		It p = b;
		if( tail[1]-head[1] < min )
			min = tail[1]-head[1];
		if( e-p < min )
		{
			min = e-p;
			for( ; min; --min )
			{
				if( comp(*head[0],*head[1]) )
					*p = *head[0], ++head[0];
				else
					*p = *head[1], ++head[1];
				++p;
			}
			*stream[0] = Stream(head[0],stream[0]->end());
			*stream[1] = Stream(head[1],stream[1]->end());
			b = p;
			return outtoken;
		}
		do
		{
			for( ; min; --min )
			{
				if( comp(*head[0],*head[1]) )
					*p = *head[0], ++head[0];
				else
					*p = *head[1], ++head[1];
				++p;
			}
			min = tail[0]-head[0];
			if( tail[1]-head[1] < min )
				min = tail[1]-head[1];
			if( e-p < min )
				min = e-p;
		}
		while( min > 7 );
		if( min )
		{
			for( ; p!=e; ++p )
			{
				if( comp(*head[0],*head[1]) )
				{
					*p = *head[0], ++head[0];
					if( head[0] == tail[0] )
					{
						++p, outtoken = token[0];
						break;
					}
				}
				else
				{
					*p = *head[1], ++head[1];
					if( head[1] == tail[1] )
					{
						++p, outtoken = token[1];
						break;
					}
				}
			}
		}
		else
		{
			if( tail[0] == head[0] )
				outtoken = token[0];
			if( tail[1] == head[1] )
				outtoken = token[1];
		}
		*stream[0] = Stream(head[0],stream[0]->end());
		*stream[1] = Stream(head[1],stream[1]->end());
		b = p;
		return outtoken;
	}
};

template<basic_order_t order, class Stream, class Token>
class hyb0_merger : public basic_merger_base<order,Stream,Token>
{
	using basic_merger_base<order,Stream,Token>::active;
	using basic_merger_base<order,Stream,Token>::stream;
	using basic_merger_base<order,Stream,Token>::token;
	using basic_merger_base<order,Stream,Token>::copy;
public:
	template<basic_order_t O, class S=Stream, class T=Token>
	struct rebind { typedef hyb0_merger<O,S,T> other; };
	///<summary>Invokes the merger</summary>
	template<class It>
	inline Token operator()(It& b, It e, Token outtoken)
	{
		if( active == 0 )
			return outtoken;
		if( active == 1 )
			return copy(b,e,outtoken);
		return invoke_invoke(b,e,outtoken,CCHECK<typename std::iterator_traits<It>::difference_type,typename std::iterator_traits<typename Stream::pointer>::difference_type>::ITERATOR_DIFFERENCES_MUST_BE_COMPATIBLE(e-b,stream[0]->end()-stream[0]->end()),*b,*stream[0]->begin());
	}
	///<summary>Invokes the merger</summary>
	template<class It, class Comp>
	inline Token operator()(It& b, It e, Token outtoken, Comp comp)
	{
		if( active == 0 )
			return outtoken;
		if( active == 1 )
			return copy(b,e,outtoken);
		return invoke_invoke(b,e,outtoken,comp,*b,*stream[0]->begin());
	}

private:
	/// returns It i, st. either i == b or for all s in [b;i): comp(s,t)
	template<class It, class T, class Diff>
	inline Token invoke_invoke(It& b, It e, Token outtoken, Diff min, T, T)
		{ std::less<T> comp; return invoke<It,T>(b,e,outtoken,min,comp); }
	template<class It, class Diff, class Comp, class T>
	inline Token invoke_invoke(It& b, It e, Token outtoken, Diff min, Comp& comp, T, T)
		{ return invoke<It,T>(b,e,outtoken,min,comp); }
	template<class It, class T, class Diff, class Comp>
	inline Token invoke(It& b, It e, Token outtoken, Diff, Comp& comp)
	{
		typename Stream::pointer head[2] = { stream[0]->begin(), stream[1]->begin() };
		typename Stream::pointer const tail[2] = { stream[0]->end(), stream[1]->end() };
		Diff min = tail[0]-head[0];
		It p = b;
		if( tail[1]-head[1] < min )
			min = tail[1]-head[1];
		if( e-p < min )
		{
			min = e-p;
			for( ; min; --min, ++p )
				if( comp(*head[0],*head[1]) )
					*p = *head[0], ++head[0];
				else
					*p = *head[1], ++head[1];
			*stream[0] = Stream(head[0],stream[0]->end());
			*stream[1] = Stream(head[1],stream[1]->end());
			b = e;
			return outtoken;
		}
		while( p != e )
		{
			if( comp(*head[0],*head[1]) )
			{
				*p = *head[0], ++head[0], ++p;
				if( head[0] == tail[0] )
				{
					outtoken = token[0];
					break;
				}
			}
			else
			{
				*p = *head[1], ++head[1], ++p;
				if( head[1] == tail[1] )
				{
					outtoken = token[1];
					break;
				}
			}
		}
		*stream[0] = Stream(head[0],stream[0]->end());
		*stream[1] = Stream(head[1],stream[1]->end());
		b = p;
		return outtoken;
	}
};

template<basic_order_t order, class Stream, class Token>
class hyb3_merger : public basic_merger_base<order,Stream,Token>
{
	using basic_merger_base<order,Stream,Token>::active;
	using basic_merger_base<order,Stream,Token>::stream;
	using basic_merger_base<order,Stream,Token>::token;
	using basic_merger_base<order,Stream,Token>::copy;
public:
	template<basic_order_t O, class S=Stream, class T=Token>
	struct rebind { typedef hyb_merger<O,S,T> other; };
	///<summary>Invokes the merger</summary>
	template<class It>
	inline Token operator()(It& b, It e, Token outtoken)
	{
		if( active == 0 )
			return outtoken;
		if( active == 1 )
			return copy(b,e,outtoken);
		return invoke_invoke(b,e,outtoken,CCHECK<typename std::iterator_traits<It>::difference_type,typename std::iterator_traits<typename Stream::pointer>::difference_type>::ITERATOR_DIFFERENCES_MUST_BE_COMPATIBLE(e-b,stream[0]->end()-stream[0]->end()),*b,*stream[0]->begin());
	}
	///<summary>Invokes the merger</summary>
	template<class It, class Comp>
	inline Token operator()(It& b, It e, Token outtoken, Comp comp)
	{
		if( active == 0 )
			return outtoken;
		if( active == 1 )
			return copy(b,e,outtoken);
		return invoke_invoke(b,e,outtoken,comp,*b,*stream[0]->begin());
	}

private:
	/// returns It i, st. either i == b or for all s in [b;i): comp(s,t)
	template<class It, class T, class Comp>
	inline It bin_search(It b, It e, T t, Comp& comp)
	{
		assert(!( b > e ));
		if( comp(t,*b) )
			return b;
		if( comp(*(e-1),t) )
			return e;
		while( e-b != 1 )
		{
			It m = b+(e-b)/2;
			if( comp(*m,t) )
				b = m;
			else
				e = m;
		}
		return e;
	}
	template<class It, class OutIt>
	inline OutIt lcopy(It b, It e, OutIt p)
	{
		for( ; b!=e; ++p, ++b )
			*p = *b;
		return p;
	}
	template<class It, class T, class Diff>
	inline Token invoke_invoke(It& b, It e, Token outtoken, Diff min, T, T)
		{ std::less<T> comp; return invoke<It,T>(b,e,outtoken,min,comp); }
	template<class It, class Diff, class Comp, class T>
	inline Token invoke_invoke(It& b, It e, Token outtoken, Diff min, Comp& comp, T, T)
		{ return invoke<It,T>(b,e,outtoken,min,comp); }
	template<class It, class T, class Diff, class Comp>
	inline Token invoke(It& b, It e, Token outtoken, Diff, Comp& comp)
	{
		typename Stream::pointer head[2] = { stream[0]->begin(), stream[1]->begin() };
		typename Stream::pointer const tail[2] = { stream[0]->end(), stream[1]->end() };
		Diff min = tail[0]-head[0], max = min;
		It p = b;
		if( tail[1]-head[1] < min )
			min = tail[1]-head[1];
		if( max < tail[1]-head[1] )
			max = tail[1]-head[1];
		if( max > 4*min )
		{
			for( ;; )
			{
				if( tail[0]-head[0] < tail[1]-head[1] )
				{
					typename Stream::pointer s = head[1]+(tail[1]-head[1])/(tail[0]-head[0]);
					s = bin_search(head[1],s,*head[0],comp);
					if(!( s-head[1] < e-p ))
					{
						s = head[1]+(e-p);
						if( s == tail[1] )
							outtoken = token[1];
						p = lcopy(head[1],s,p);
						head[1] = s;
						break;
					}
					p = lcopy(head[1],s,p);
					if( s == head[1] )
					{
						*p = *head[0], ++p, ++head[0];
						if( head[0] == tail[0] )
						{
							outtoken = token[0];
							break;
						}
					}
					else
					{
						head[1] = s;
						if( head[1] == tail[1] )
						{
							outtoken = token[1];
							break;
						}
					}
				}
				else
				{
					typename Stream::pointer s = head[0]+(tail[0]-head[0])/(tail[1]-head[1]);
					s = bin_search(head[0],s,*head[1],comp);
					if(!( s-head[0] < e-p ))
					{
						s = head[0]+(e-p);
						if( s == tail[0] )
							outtoken = token[0];
						p = lcopy(head[0],s,p);
						head[0] = s;
						break;
					}
					p = lcopy(head[0],s,p);
					if( s == head[0] )
					{
						*p = *head[1], ++p, ++head[1];
						if( head[1] == tail[1] )
						{
							outtoken = token[1];
							break;
						}
					}
					else
					{
						head[0] = s;
						if( head[0] == tail[0] )
						{
							outtoken = token[0];
							break;
						}
					}
				}
				if( p == e )
					break;
			}
			*stream[0] = Stream(head[0],stream[0]->end());
			*stream[1] = Stream(head[1],stream[1]->end());
			b = p;
			return outtoken;
		}
		if( e-p < min )
		{
			min = e-p;
			for( ; min; --min )
			{
				if( comp(*head[0],*head[1]) )
					*p = *head[0], ++head[0];
				else
					*p = *head[1], ++head[1];
				++p;
			}
			*stream[0] = Stream(head[0],stream[0]->end());
			*stream[1] = Stream(head[1],stream[1]->end());
			b = p;
			return outtoken;
		}
		do
		{
			for( ; min; --min )
			{
				if( comp(*head[0],*head[1]) )
					*p = *head[0], ++head[0];
				else
					*p = *head[1], ++head[1];
				++p;
			}
			min = tail[0]-head[0];
			if( tail[1]-head[1] < min )
				min = tail[1]-head[1];
			if( e-p < min )
				min = e-p;
		}
		while( min > 7 );
		if( min )
		{
			for( ; p!=e; ++p )
			{
				if( comp(*head[0],*head[1]) )
				{
					*p = *head[0], ++head[0];
					if( head[0] == tail[0] )
					{
						++p, outtoken = token[0];
						break;
					}
				}
				else
				{
					*p = *head[1], ++head[1];
					if( head[1] == tail[1])
					{
						++p, outtoken = token[1];
						break;
					}
				}
			}
		}
		else
		{
			if( tail[0] == head[0] )
				outtoken = token[0];
			if( tail[1] == head[1] )
				outtoken = token[1];
		}
		*stream[0] = Stream(head[0],stream[0]->end());
		*stream[1] = Stream(head[1],stream[1]->end());
		b = p;
		return outtoken;
	}
};

/** <remark>Unfinished</remark>*/
template<basic_order_t order, class Stream, class Token>
class adv_merger : public basic_merger_base<order,Stream,Token>
{
	using basic_merger_base<order,Stream,Token>::active;
	using basic_merger_base<order,Stream,Token>::stream;
	using basic_merger_base<order,Stream,Token>::token;
	using basic_merger_base<order,Stream,Token>::copy;
public:
	template<basic_order_t O, class S=Stream, class T=Token>
	struct rebind { typedef adv_merger<O,S,T> other; };
	///<summary>Invokes the merger</summary>
	template<class It>
	inline Token operator()(It& b, It e, Token outtoken)
	{
		if( active == 0 || b == e )
			return outtoken;
		if( active == 1 )
			return copy(b,e,outtoken);
		return invoke_invoke(b,e,outtoken,e-b,*b,*stream[0]->begin());
	}
	///<summary>Invokes the merger</summary>
	template<class It, class Comp>
	inline Token operator()(It& b, It e, Token outtoken, Comp comp)
	{
		if( active == 0 )
			return outtoken;
		if( active == 1 )
			return copy(b,e,outtoken);
		return invoke_invoke(b,e,outtoken,comp,*b,*stream[0]->begin());
	}

private:
	template<class It, class T, class Diff>
	inline Token invoke_invoke(It& b, It e, Token outtoken, Diff min, T, T)
		{ std::less<T> comp; return invoke<It,T>(b,e,outtoken,min,comp); }
	template<class It, class Diff, class Comp, class T>
	inline Token invoke_invoke(It& b, It e, Token outtoken, Diff min, Comp& comp, T, T)
		{ return invoke<It,T>(b,e,outtoken,min,comp); }

	/// returns It i, st. either i == b or for all s in [b;i): comp(s,t)
	template<class It, class T, class Comp>
	inline It bin_search(It b, It e, T t, Comp& comp)
	{
		assert(!( b > e ));
		while( e-b != 1 )
		{
			It m = b+(e-b)/2;
			if( comp(*m,t) )
				b = m;
			else
				e = m;
//			assert( comp(*b,t) );
//			assert( !comp(*e,t) );
		}
		return e;
	}
	template<class T, class Diff, class Comp>
	inline std::pair<Token,Diff> done(Diff n, Token outtoken, Comp& comp)
	{
		typename Stream::pointer B[2] = { stream[0]->begin(), stream[1]->begin() };
		typename Stream::pointer E[2] = { stream[0]->end(), stream[1]->end() };
		typename Stream::pointer b[2] = { stream[0]->begin(), stream[1]->begin() };
		typename Stream::pointer e[2] = { stream[0]->end(), stream[1]->end() };
		if( e[0]-b[0] >= n && e[1]-b[1] >= n )
			return std::pair<Token,Diff>(outtoken,n);
		if( (e[0]-b[0])+(e[1]-b[1]) < n )
			if( comp(*(e[0]-1),*(e[1]-1)) )
				return std::pair<Token,Diff>(token[0],(e[0]-b[0])+(bin_search(b[1],e[1],*(e[0]-1),comp))-b[1]);
			else
				return std::pair<Token,Diff>(token[1],(e[1]-b[1])+(bin_search(b[0],e[0],*(e[1]-1),comp))-b[0]);
		if( n < e[0]-b[0] )
			e[0] = b[0]+n;
		if( n < e[1]-b[1] )
			e[1] = b[1]+n;

		for( typename Stream::pointer p=b[0]; p!=e[0]; ++p )
			std::cout << *p << ' ';
		std::cout << std::endl;
		for( typename Stream::pointer p=b[1]; p!=e[1]; ++p )
			std::cout << *p << ' ';
		std::cout << std::endl;

		while( !comp(*b[0],*b[1]) )
		{
			++b[1];
			if( b[1] == e[1] )
				if( n < e[1]-B[1] )
					return std::pair<Token,Diff>(outtoken,n);
				else
					return std::pair<Token,Diff>(token[1],e[1]-B[1]);
		}
		while( !comp(*(e[1]-1),*(e[0]-1)) )
		{
			--e[1];
			if( b[1] == e[1] )
				if( n < (b[1]-B[1])+(e[0]-b[0]) )
					return std::pair<Token,Diff>(outtoken,n);
				else
					return std::pair<Token,Diff>(token[0],(b[1]-B[1])+(e[0]-b[0]));
		}
		while( (b[0]-B[0])+(b[1]-B[1]) != n || (b[0]-B[0])+(b[1]-B[1]) != n )
		{
			assert( n < (e[0]-B[0])+(e[1]-B[1]) && (b[0]-B[0])+(b[1]-B[1]) < n );
			assert( comp(*b[0],*b[1]) && comp(*(e[1]-1),*(e[0]-1)) );
			typename Stream::pointer m[2] = { b[0]+(e[0]-b[0])/2, b[1]+(e[1]-b[1])/2 };
			bool here = false;
			if( comp(*m[0],*b[1]) && (m[0]-B[0])+(b[1]-B[1]) < n )
				b[0] = m[0], here = true;
			if( comp(*b[0],*m[1]) && (b[0]-B[0])+(m[1]-B[1]) < n )
				b[1] = m[1], here = true;;
			if( comp(*(e[1]-1),*(m[0]-1)) && n < (m[0]-B[0])+(e[1]-B[1]) )
				e[0] = m[0], here = true;;
			if( comp(*(m[1]-1),*(e[0]-1)) && n < (e[0]-B[0])+(m[1]-B[1]) )
				e[1] = m[1], here = true;
			assert( here );
			if( b[0] == e[0] )
				return std::pair<Token,Diff>(token[0],E[0]-B[0]);
			if( b[1] == e[1] )
				return std::pair<Token,Diff>(token[1],E[1]-B[1]);
		}
		return std::pair<Token,Diff>(outtoken,n);
	}
	template<class It, class T, class Diff, class Comp>
	inline Token invoke(It& b, It e, Token outtoken, Diff, Comp& comp)
	{
		std::pair<Token,Diff> P = done<T>(e-b,outtoken,comp);
		register typename Stream::pointer head[2] = { stream[0]->begin(), stream[1]->begin() };
		register It p = b;
		for( ; p!=e; ++p )
		{
			if( comp(*head[0],*head[1]) )
			{
				*p = *head[0], ++head[0];
				if( head[0] == stream[0]->end() )
				{
					++p;
					outtoken = token[0];
					break;
				}
			}
			else
			{
				*p = *head[1], ++head[1];
				if( head[1] == stream[1]->end() )
				{
					++p;
					outtoken = token[1];
					break;
				}
			}
		}
		assert( P.first == outtoken );
		assert( p-b == P.second );
		*stream[0] = Stream(head[0],stream[0]->end());
		*stream[1] = Stream(head[1],stream[1]->end());
		b = p;
		return outtoken;

/*		typename Stream::pointer head[2] = { stream[0]->begin(), stream[1]->begin() };
		std::pair<Token,Diff> p = done<T>(e-b,outtoken,comp);
		It q = b;
		do
		{
			if( comp(*head[0],*head[1]) )
				*q = *head[0], ++head[0], ++q;
			else
				*q = *head[1], ++head[1], ++q;
		}
		while( --p.second );
		*stream[0] = Stream(head[0],stream[0]->end());
		*stream[1] = Stream(head[1],stream[1]->end());
		return p.first;

		b = b+p.second;
		It q = b;
		p.second = -p.second;
		do
		{
			if( comp(*head[0],*head[1]) )
				*(q+p.second) = *head[0], ++head[0];
			else
				*(q+p.second) = *head[1], ++head[1];
		}
		while( ++p.second);
*/	}
};

template<basic_order_t order, class Stream, class Token>
class comp_merger : public basic_merger_base<order,Stream,Token>
{
	using basic_merger_base<order,Stream,Token>::active;
	using basic_merger_base<order,Stream,Token>::stream;
	using basic_merger_base<order,Stream,Token>::token;
	using basic_merger_base<order,Stream,Token>::copy;
public:
	template<basic_order_t O, class S=Stream, class T=Token>
	struct rebind { typedef comp_merger<O,S,T> other; };
	///<summary>Invokes the merger</summary>
	template<class It>
	inline Token operator()(It& b, It e, Token outtoken)
	{
		if( active == 0 || b == e )
			return outtoken;
		if( active == 1 )
			return copy(b,e,outtoken);
		return invoke_invoke(b,e,outtoken,CCHECK<typename std::iterator_traits<It>::difference_type,typename std::iterator_traits<typename Stream::pointer>::difference_type>::ITERATOR_DIFFERENCES_MUST_BE_COMPATIBLE(e-b,stream[0]->end()-stream[0]->end()),*b,*stream[0]->begin());
	}
	///<summary>Invokes the merger</summary>
	template<class It, class Comp>
	inline Token operator()(It& b, It e, Token outtoken, Comp comp)
	{
		if( active == 0 || b == e  )
			return outtoken;
		if( active == 1 )
			return copy(b,e,outtoken);
		return invoke_invoke(b,e,outtoken,comp,*b,*stream[0]->begin());
	}

public: /*ICC*/
	template<basic_order_t O, class SIt, class Ptr, class Comp>
	class move_one
	{
	public:
		inline void operator()(SIt *head, Ptr dest, Comp& comp)
		{
			move_one<O-1,SIt,Ptr,Comp> mm;
			if( comp(*head[0],*head[O-1]) )
				mm(head,dest,comp);
			else
				mm(head+1,dest,comp);
		}
	};
	template<class SIt, class Ptr, class Comp>
	class move_one<2,SIt,Ptr,Comp>
	{
	public:
		inline void operator()(SIt *head, Ptr dest, Comp& comp)
		{
			if( comp(*head[0],*head[1]) )
				*dest = *head[0], ++head[0];
			else
				*dest = *head[1], ++head[1];
		}
	};
	template<basic_order_t Order, class Ptr, class SIt, class Diff, class Comp>
	class move_min
	{
	public:
		inline Ptr operator()(Ptr p, Diff min, SIt *head, Comp& comp)
		{
			move_one<Order,SIt,Ptr,Comp> mo;
			do
				mo(head,p,comp), ++p;
			while( --min );
			return p;
		}
	};
	template<basic_order_t Order, class T, class SIt, class Diff, class Comp>
	class move_min<Order, T*, SIt, Diff, Comp>
	{
	public:
		inline T* operator()(T *p, Diff min, SIt *head, Comp& comp)
		{
			p += min;
			min = -min;
			move_one<Order,SIt,T*,Comp> mo;
			do
				mo(head,p+min,comp);
			while( ++min );
			return p;
		}
	};
private:
	template<class It, class T, class Diff>
	inline Token invoke_invoke(It& b, It e, Token outtoken, Diff min, T, T)
		{ std::less<T> comp; return invoke<It,T>(b,e,outtoken,min,comp); }
	template<class It, class Diff, class Comp, class T>
	inline Token invoke_invoke(It& b, It e, Token outtoken, Diff min, Comp& comp, T, T)
		{ return invoke<It,T>(b,e,outtoken,min,comp); }
	template<class It, class T, class Diff, class Comp>
	inline Token invoke(It& b, It e, Token outtoken, Diff min, Comp& comp)
	{
		register typename Stream::pointer head[order];
		for( basic_order_t i=0; i<active; ++i )
		{
			head[i] = stream[i]->begin();
//			auto_array::CachePrefetch(head[i]);
			if( stream[i]->end()-head[i] < min )
				min = stream[i]->end()-head[i];
		}
		assert( min > 0 );
		register It p = b;
		for( ;; )
		{
			assert( min > 0 );
			switch( active )
			{
				case 4:
				{
					move_min<4,It,typename Stream::pointer,Diff,Comp> mm;
					p = mm(p, min, head, comp);
					break;
				}
				case 3:
				{
					move_min<3,It,typename Stream::pointer,Diff,Comp> mm;
					p = mm(p, min, head, comp);
					break;
				}
				case 2:
				{
					move_min<2,It,typename Stream::pointer,Diff,Comp> mm;
					p = mm(p, min, head, comp);
					break;
				}
				case 5:
				{
					move_min<5,It,typename Stream::pointer,Diff,Comp> mm;
					p = mm(p, min, head, comp);
					break;
				}
				case 6:
				{
					move_min<6,It,typename Stream::pointer,Diff,Comp> mm;
					p = mm(p, min, head, comp);
					break;
				}
				default:
					do
					{
						typename Stream::pointer *m = head;
						for( typename Stream::pointer *q=m+1; q<head+active; ++q )
							if( comp(**q,**m) )
								m = q;
						*p = **m, ++*m, ++p;
					}
					while( --min );
			}
			min = stream[0]->end()-head[0];
			if( min == 0 )
			{
				for( basic_order_t j=0; j<active; ++j )
					*stream[j] = Stream(head[j],stream[j]->end());
				b = p;
				return token[0];
			}
			for( basic_order_t i=1; i<active; ++i )
				if( stream[i]->end()-head[i] < min )
				{
					min = stream[i]->end()-head[i];
					if( min == 0 )
					{
						for( basic_order_t j=0; j<active; ++j )
							*stream[j] = Stream(head[j],stream[j]->end());
						b = p;
						return token[i];
					}
				}
			if( e-p < min )
			{
				min = e-p;
				if( min == 0 )
				{
					for( basic_order_t i=0; i<active; ++i )
						*stream[i] = Stream(head[i],stream[i]->end());
					b = e;
					return outtoken;
				}
			}
		}
	}
};

template<basic_order_t order, class Stream, class Token>
class simple_comp_merger : public basic_merger_base<order,Stream,Token>
{
	using basic_merger_base<order,Stream,Token>::active;
	using basic_merger_base<order,Stream,Token>::stream;
	using basic_merger_base<order,Stream,Token>::token;
	using basic_merger_base<order,Stream,Token>::copy;
public:
	template<basic_order_t O, class S=Stream, class T=Token>
	struct rebind { typedef simple_comp_merger<O,S,T> other; };
	///<summary>Invokes the merger</summary>
	template<class It>
	inline Token operator()(It& b, It e, Token outtoken)
	{
		if( active == 0 || b == e )
			return outtoken;
		if( active == 1 )
			return copy(b,e,outtoken);
		return invoke_invoke(b,e,outtoken,CCHECK<typename std::iterator_traits<It>::difference_type,typename std::iterator_traits<typename Stream::pointer>::difference_type>::ITERATOR_DIFFERENCES_MUST_BE_COMPATIBLE(e-b,stream[0]->end()-stream[0]->end()),*b,*stream[0]->begin());
	}
	///<summary>Invokes the merger</summary>
	template<class It, class Comp>
	inline Token operator()(It& b, It e, Token outtoken, Comp comp)
	{
		if( active == 0 || b == e  )
			return outtoken;
		if( active == 1 )
			return copy(b,e,outtoken);
		return invoke_invoke(b,e,outtoken,comp,*b,*stream[0]->begin());
	}

public: /*ICC*/
	template<basic_order_t O, class SIt, class Ptr, class Comp>
	class move_one
	{
	public:
		inline bool operator()(SIt *head, SIt *tail, Token *t, Ptr& dest, Token &r, Comp& comp)
		{
			move_one<O-1,SIt,Ptr,Comp> mm;
			if( comp(*head[0],*head[O-1]) )
				return mm(head,tail,t,dest,r,comp);
			else
				return mm(head+1,tail+1,t+1,dest,r,comp);
		}
	};
	template<class SIt, class Ptr, class Comp>
	class move_one<2,SIt,Ptr,Comp>
	{
	public:
		inline bool operator()(SIt *head, SIt *tail, Token *t, Ptr& dest, Token &r, Comp& comp)
		{
			if( comp(*head[0],*head[1]) )
			{
				*dest = *head[0], ++head[0], ++dest;
				if( head[0] == tail[0] )
				{
					r = t[0];
					return true;
				}
				else
					return false;
			}
			else
			{
				*dest = *head[1], ++head[1], ++dest;
				if( head[1] == tail[1] )
				{
					r = t[1];
					return true;
				}
				else
					return false;
			}
		}
	};
	template<basic_order_t O, class SIt, class Ptr, class Comp>
	class switch_active
	{
	public:
		inline void operator()(basic_order_t active, SIt *head, SIt *tail, Token *t, Ptr& b, Ptr e, Token &r, Comp& comp)
		{
			if( active == O )
			{
				Ptr p = b;
				move_one<O,SIt,Ptr,Comp> mo;
				while( p != e )
					if( mo(head,tail,t,p,r,comp) )
						break;
				b = p;
			}
			else
			{
				switch_active<O-1,typename Stream::pointer,Ptr,Comp> sa;
				sa(active,head,tail,t,b,e,r,comp);
			}
		}
	};
	template<class SIt, class Ptr, class Comp>
	class switch_active<2,SIt,Ptr,Comp>
	{
	public:
		inline void operator()(basic_order_t active, SIt *head, SIt *tail, Token *t, Ptr& b, Ptr e, Token &r, Comp& comp)
		{
			assert( active == 2 );
			Ptr p = b;
			move_one<2,SIt,Ptr,Comp> mo;
			while( p != e )
				if( mo(head,tail,t,p,r,comp) )
					break;
			b = p;
		}
	};
private:
	template<class It, class T, class Diff>
	inline Token invoke_invoke(It& b, It e, Token outtoken, Diff min, T, T)
		{ std::less<T> comp; return invoke<It,T>(b,e,outtoken,min,comp); }
	template<class It, class Diff, class Comp, class T>
	inline Token invoke_invoke(It& b, It e, Token outtoken, Diff min, Comp& comp, T, T)
		{ return invoke<It,T>(b,e,outtoken,min,comp); }
	template<class It, class T, class Diff, class Comp>
	inline Token invoke(It& b, It e, Token outtoken, Diff min, Comp& comp)
	{
		typename Stream::pointer head[order], tail[order];
		for( basic_order_t i=0; i<active; ++i )
			head[i] = stream[i]->begin(), tail[i] = stream[i]->end();
		switch_active<order,typename Stream::pointer,It,Comp> sa;
		sa(active,head,tail,token,b,e,outtoken,comp);
		for( basic_order_t i=0; i<active; ++i )
			*stream[i] = Stream(head[i],stream[i]->end());
		return outtoken;
	}
};


template<basic_order_t order, class Stream, class Token>
class four_merger : public basic_merger_base<order,Stream,Token>
{
	using basic_merger_base<order,Stream,Token>::active;
	using basic_merger_base<order,Stream,Token>::stream;
	using basic_merger_base<order,Stream,Token>::token;
	using basic_merger_base<order,Stream,Token>::copy;
public:
	template<basic_order_t O, class S=Stream, class T=Token>
	struct rebind { typedef four_merger<O,S,T> other; };
	///<summary>Invokes the merger</summary>
	template<class It>
	inline Token operator()(It& b, It e, Token outtoken)
	{
		if( active == 0 || b == e )
			return outtoken;
		if( active == 1 )
			return copy(b,e,outtoken);
		typename std::iterator_traits<It>::iterator_category cat;
		return invoke_invoke(b,e,outtoken,CCHECK<typename std::iterator_traits<It>::difference_type,typename std::iterator_traits<typename Stream::pointer>::difference_type>::ITERATOR_DIFFERENCES_MUST_BE_COMPATIBLE(e-b,stream[0]->end()-stream[0]->end()),*b,*stream[0]->begin(),cat);
	}
	///<summary>Invokes the merger</summary>
	template<class It>
	inline Token operator()(It& b, Token outtoken)
	{
		if( active == 0 )
			return outtoken;
		if( active == 1 )
		{
			b = std::copy(stream[0]->begin(),stream[0]->end(),b);
			*stream[0] = Stream(stream[0]->end(), stream[0]->end());
			return token[0];
		}
		typename std::iterator_traits<It>::iterator_category cat;
		return invoke_invoke(b,*b,*stream[0]->begin(),cat);
	}
	///<summary>Invokes the merger</summary>
	template<class It, class Comp>
	inline Token operator()(It& b, It e, Token outtoken, Comp comp)
	{
		if( active == 0 || b == e  )
			return outtoken;
		if( active == 1 )
			return copy(b,e,outtoken);
		typename std::iterator_traits<It>::iterator_category cat;
		return invoke_invoke(b,e,outtoken,comp,*b,*stream[0]->begin(),cat);
	}
private:
	template<class It, class T, class Diff, class ItCat>
	inline Token invoke_invoke(It& b, It e, Token outtoken, Diff min, T, T, ItCat)
		{ std::less<T> comp; return invoke<It,T,ItCat>(b,e,outtoken,min,comp); }
	template<class It, class Diff, class Comp, class T, class ItCat>
	inline Token invoke_invoke(It& b, It e, Token outtoken, Diff min, Comp& comp, T, T, ItCat)
		{ return invoke<It,T,ItCat>(b,e,outtoken,min,comp); }
	template<class It, class T, class ItCat>
	inline Token invoke_invoke(It& b, T, T, ItCat)
		{ std::less<T> comp; return invoke<It,ItCat>(b,comp); }
	template<class It, class T, class ItCat, class Diff, class Comp>
	inline Token invoke(It& b, It e, Token outtoken, Diff min, Comp& comp)
	{
		typename Stream::pointer head[order], tail[order];
		for( basic_order_t i=0; i<active; ++i )
			head[i] = stream[i]->begin(), tail[i] = stream[i]->end();
		It p = b;
		if( active == 4 )
		{
			while( p != e )
				if( comp(*head[0],*head[3]) )
				{
					if( comp(*head[0],*head[2]) )
					{
						if( comp(*head[0],*head[1]) )
						{
							*p = *head[0], ++head[0], ++p;
							if( head[0] == tail[0] )
							{
								outtoken = token[0];
								break;
							}
						}
						else
						{
							*p = *head[1], ++head[1], ++p;
							if( head[1] == tail[1] )
							{
								outtoken = token[1];
								break;
							}
						}
					}
					else
					{
						if( comp(*head[1],*head[2]) )
						{
							*p = *head[1], ++head[1], ++p;
							if( head[1] == tail[1] )
							{
								outtoken = token[1];
								break;
							}
						}
						else
						{
							*p = *head[2], ++head[2], ++p;
							if( head[2] == tail[2] )
							{
								outtoken = token[2];
								break;
							}
						}
					}
				}
				else
				{
					if( comp(*head[1],*head[3]) )
					{
						if( comp(*head[1],*head[2]) )
						{
							*p = *head[1], ++head[1], ++p;
							if( head[1] == tail[1] )
							{
								outtoken = token[1];
								break;
							}
						}
						else
						{
							*p = *head[2], ++head[2], ++p;
							if( head[2] == tail[2] )
							{
								outtoken = token[2];
								break;
							}
						}
					}
					else
					{
						if( comp(*head[2],*head[3]) )
						{
							*p = *head[2], ++head[2], ++p;
							if( head[2] == tail[2] )
							{
								outtoken = token[2];
								break;
							}
						}
						else
						{
							*p = *head[3], ++head[3], ++p;
							if( head[3] == tail[3] )
							{
								outtoken = token[3];
								break;
							}
						}
					}
				}
		}
		else if( active == 3 )
		{
			while( p != e )
				if( comp(*head[0],*head[2]) )
				{
					if( comp(*head[0],*head[1]) )
					{
						*p = *head[0], ++head[0], ++p;
						if( head[0] == tail[0] )
						{
							outtoken = token[0];
							break;
						}
					}
					else
					{
						*p = *head[1], ++head[1], ++p;
						if( head[1] == tail[1] )
						{
							outtoken = token[1];
							break;
						}
					}
				}
				else
				{
					if( comp(*head[1],*head[2]) )
					{
						*p = *head[1], ++head[1], ++p;
						if( head[1] == tail[1] )
						{
							outtoken = token[1];
							break;
						}
					}
					else
					{
						*p = *head[2], ++head[2], ++p;
						if( head[2] == tail[2] )
						{
							outtoken = token[2];
							break;
						}
					}
				}
		}
		else if( active == 2 )
			while( p != e )
				if( comp(*head[0],*head[1]) )
				{
					*p = *head[0], ++head[0], ++p;
					if( head[0] == tail[0] )
					{
						outtoken = token[0];
						break;
					}
				}
				else
				{
					*p = *head[1], ++head[1], ++p;
					if( head[1] == tail[1] )
					{
						outtoken = token[1];
						break;
					}
				}
		b = p;
		for( basic_order_t i=0; i<active; ++i )
			*stream[i] = Stream(head[i],stream[i]->end());
		return outtoken;
	}
	template<class It, class ItCat, class Comp>
	inline Token invoke(It& b, Comp& comp)
	{
		typename Stream::pointer head[order], tail[order];
		for( basic_order_t i=0; i<active; ++i )
			head[i] = stream[i]->begin(), tail[i] = stream[i]->end();
		if( active == 4 )
		{
			for( ;; )
				if( comp(*head[0],*head[3]) )
				{
					if( comp(*head[0],*head[2]) )
					{
						if( comp(*head[0],*head[1]) )
						{
							*b = *head[0], ++head[0], ++b;
							if( head[0] == tail[0] )
							{
								for( basic_order_t i=0; i<active; ++i )
									*stream[i] = Stream(head[i],stream[i]->end());
								return token[0];
							}
						}
						else
						{
							*b = *head[1], ++head[1], ++b;
							if( head[1] == tail[1] )
							{
								for( basic_order_t i=0; i<active; ++i )
									*stream[i] = Stream(head[i],stream[i]->end());
								return token[1];
							}
						}
					}
					else
					{
						if( comp(*head[1],*head[2]) )
						{
							*b = *head[1], ++head[1], ++b;
							if( head[1] == tail[1] )
							{
								for( basic_order_t i=0; i<active; ++i )
									*stream[i] = Stream(head[i],stream[i]->end());
								return token[1];
							}
						}
						else
						{
							*b = *head[2], ++head[2], ++b;
							if( head[2] == tail[2] )
							{
								for( basic_order_t i=0; i<active; ++i )
									*stream[i] = Stream(head[i],stream[i]->end());
								return token[2];
							}
						}
					}
				}
				else
				{
					if( comp(*head[1],*head[3]) )
					{
						if( comp(*head[1],*head[2]) )
						{
							*b = *head[1], ++head[1], ++b;
							if( head[1] == tail[1] )
							{
								for( basic_order_t i=0; i<active; ++i )
									*stream[i] = Stream(head[i],stream[i]->end());
								return token[1];
							}
						}
						else
						{
							*b = *head[2], ++head[2], ++b;
							if( head[2] == tail[2] )
							{
								for( basic_order_t i=0; i<active; ++i )
									*stream[i] = Stream(head[i],stream[i]->end());
								return token[2];
							}
						}
					}
					else
					{
						if( comp(*head[2],*head[3]) )
						{
							*b = *head[2], ++head[2], ++b;
							if( head[2] == tail[2] )
							{
								for( basic_order_t i=0; i<active; ++i )
									*stream[i] = Stream(head[i],stream[i]->end());
								return token[2];
							}
						}
						else
						{
							*b = *head[3], ++head[3], ++b;
							if( head[3] == tail[3] )
							{
								for( basic_order_t i=0; i<active; ++i )
									*stream[i] = Stream(head[i],stream[i]->end());
								return token[3];
							}
						}
					}
				}
		}
		else if( active == 3 )
		{
			for( ;; )
				if( comp(*head[0],*head[2]) )
				{
					if( comp(*head[0],*head[1]) )
					{
						*b = *head[0], ++head[0], ++b;
						if( head[0] == tail[0] )
						{
							for( basic_order_t i=0; i<active; ++i )
								*stream[i] = Stream(head[i],stream[i]->end());
							return token[0];
						}
					}
					else
					{
						*b = *head[1], ++head[1], ++b;
						if( head[1] == tail[1] )
						{
							for( basic_order_t i=0; i<active; ++i )
								*stream[i] = Stream(head[i],stream[i]->end());
							return token[1];
						}
					}
				}
				else
				{
					if( comp(*head[1],*head[2]) )
					{
						*b = *head[1], ++head[1], ++b;
						if( head[1] == tail[1] )
						{
							for( basic_order_t i=0; i<active; ++i )
								*stream[i] = Stream(head[i],stream[i]->end());
							return token[1];
						}
					}
					else
					{
						*b = *head[2], ++head[2], ++b;
						if( head[2] == tail[2] )
						{
							for( basic_order_t i=0; i<active; ++i )
								*stream[i] = Stream(head[i],stream[i]->end());
							return token[2];
						}
					}
				}
		}
		else if( active == 2 )
			for( ;; )
				if( comp(*head[0],*head[1]) )
				{
					*b = *head[0], ++head[0], ++b;
					if( head[0] == tail[0] )
					{
						for( basic_order_t i=0; i<active; ++i )
							*stream[i] = Stream(head[i],stream[i]->end());
						return token[0];
					}
				}
				else
				{
					*b = *head[1], ++head[1], ++b;
					if( head[1] == tail[1] )
					{
						for( basic_order_t i=0; i<active; ++i )
							*stream[i] = Stream(head[i],stream[i]->end());
						return token[1];
					}
				}
	}
};

template<basic_order_t order, class Stream, class Token>
class for_merger : public basic_merger_base<order,Stream,Token>
{
//	static basic_merger_stat<order> bms;
	using basic_merger_base<order,Stream,Token>::active;
	using basic_merger_base<order,Stream,Token>::stream;
	using basic_merger_base<order,Stream,Token>::token;
	using basic_merger_base<order,Stream,Token>::copy;
public:
	template<basic_order_t O, class S=Stream, class T=Token>
	struct rebind { typedef for_merger<O,S,T> other; };
	///<summary>Invokes the merger</summary>
	template<class It>
	inline Token operator()(It& b, It e, Token outtoken)
	{
//		bms.active(active);
		if( active == 0 )
			return outtoken;
		if( active == 1 )
			return copy(b,e,outtoken);
		return invoke_invoke(b,e,outtoken,CCHECK<typename std::iterator_traits<It>::difference_type,typename std::iterator_traits<typename Stream::pointer>::difference_type>::ITERATOR_DIFFERENCES_MUST_BE_COMPATIBLE(e-b,stream[0]->end()-stream[0]->end()),*b,*stream[0]->begin());
	}
	///<summary>Invokes the merger</summary>
	template<class It, class Comp>
	inline Token operator()(It& b, It e, Token outtoken, Comp comp)
	{
//		bms.active(active);
		if( active == 0 )
			return outtoken;
		if( active == 1 )
			return copy(b,e,outtoken);
		return invoke_invoke(b,e,outtoken,comp,*b,*stream[0]->begin());
	}

private:
	template<class It, class T, class Diff>
	inline Token invoke_invoke(It& b, It e, Token outtoken, Diff min, T, T)
		{ std::less<T> comp; return invoke<It,T>(b,e,outtoken,min,comp); }
	template<class It, class Diff, class Comp, class T>
	inline Token invoke_invoke(It& b, It e, Token outtoken, Diff min, Comp& comp, T, T)
		{ return invoke<It,T>(b,e,outtoken,min,comp); }
	template<class It, class T, class Diff, class Comp>
	inline Token invoke(It& b, It e, Token outtoken, Diff min, Comp& comp)
	{
		typename Stream::pointer head[order];
		for( basic_order_t i=0; i<active; ++i )
		{
			head[i] = stream[i]->begin();
			if( stream[i]->end()-head[i] < min )
				min = stream[i]->end()-head[i];
		}
		It p = b;
		for( ;; )
		{
//			bms.inps(stream[0]->end()-stream[0]->begin(),stream[1]->end()-stream[1]->begin());
			for( ; min; --min )
			{
				typename Stream::pointer *m = head;
				for( typename Stream::pointer *q=m+1; q<head+active; ++q )
					if( comp(**q,**m) )
						m = q;
				*p = **m, ++*m, ++p;
			}
			min = static_cast<Diff>(stream[0]->end()-head[0]);
			if( min == 0 )
			{
				for( basic_order_t j=0; j<active; ++j )
					*stream[j] = Stream(head[j],stream[j]->end());
				b = p;
				return token[0];
			}
			for( basic_order_t i=1; i<active; ++i )
				if( stream[i]->end()-head[i] < min )
				{
					min = static_cast<Diff>(stream[i]->end()-head[i]);
					if( min == 0 )
					{
						for( basic_order_t j=0; j<active; ++j )
							*stream[j] = Stream(head[j],stream[j]->end());
						b = p;
						return token[i];
					}
				}
			if( e-p < min )
			{
				min = e-p;
				if( min == 0 )
				{
					for( basic_order_t i=0; i<active; ++i )
						*stream[i] = Stream(head[i],stream[i]->end());
					b = e;
					return outtoken;
				}
			}
		}
	}
};
//template<basic_order_t order, class Stream, class Token>
//basic_merger_stat<order> for_merger<order,Stream,Token>::bms;

template<basic_order_t order, class Stream, class Token>
class simple_for_merger : public basic_merger_base<order,Stream,Token>
{
	using basic_merger_base<order,Stream,Token>::active;
	using basic_merger_base<order,Stream,Token>::stream;
	using basic_merger_base<order,Stream,Token>::token;
	using basic_merger_base<order,Stream,Token>::copy;
public:
	template<basic_order_t O, class S=Stream, class T=Token>
	struct rebind { typedef simple_for_merger<O,S,T> other; };
	///<summary>Invokes the merger</summary>
	template<class It>
	inline Token operator()(It& b, It e, Token outtoken)
	{
		if( active == 0 )
			return outtoken;
		if( active == 1 )
			return copy(b,e,outtoken);
		return invoke_invoke(b,e,outtoken,CCHECK<typename std::iterator_traits<It>::difference_type,typename std::iterator_traits<typename Stream::pointer>::difference_type>::ITERATOR_DIFFERENCES_MUST_BE_COMPATIBLE(e-b,stream[0]->end()-stream[0]->end()),*b,*stream[0]->begin());
	}
	///<summary>Invokes the merger</summary>
	template<class It, class Comp>
	inline Token operator()(It& b, It e, Token outtoken, Comp comp)
	{
		if( active == 0 )
			return outtoken;
		if( active == 1 )
			return copy(b,e,outtoken);
		return invoke_invoke(b,e,outtoken,comp,*b,*stream[0]->begin());
	}

private:
	template<class It, class T, class Diff>
	inline Token invoke_invoke(It& b, It e, Token outtoken, Diff min, T, T)
		{ std::less<T> comp; return invoke<It,T>(b,e,outtoken,min,comp); }
	template<class It, class Diff, class Comp, class T>
	inline Token invoke_invoke(It& b, It e, Token outtoken, Diff min, Comp& comp, T, T)
		{ return invoke<It,T>(b,e,outtoken,min,comp); }
	template<class It, class T, class Diff, class Comp>
	inline Token invoke(It& b, It e, Token outtoken, Diff min, Comp& comp)
	{
		struct ht
		{
			typename Stream::pointer h, t;
		} s[order];
		for( basic_order_t i=0; i<active; ++i )
			s[i].h = stream[i]->begin(), s[i].t = stream[i]->end();
		It p = b;
		while( p != e )
		{
			ht *m = s;
			for( ht *q=m+1; q<s+active; ++q )
				if( comp(*q->h,*m->h) )
					m = q;
			*p = *m->h, ++(m->h), ++p;
			if( m->h == m->t )
			{
				outtoken = token[m-s];
				break;
			}
		}
		for( basic_order_t i=0; i<active; ++i )
			*stream[i] = Stream(s[i].h,stream[i]->end());
		b = p;
		return outtoken;
	}
};

template<basic_order_t order, class Stream, class Token>
class for_val_merger : public basic_merger_base<order,Stream,Token>
{
	using basic_merger_base<order,Stream,Token>::active;
	using basic_merger_base<order,Stream,Token>::stream;
	using basic_merger_base<order,Stream,Token>::token;
	using basic_merger_base<order,Stream,Token>::copy;
public:
	template<basic_order_t O, class S=Stream, class T=Token>
	struct rebind { typedef for_val_merger<O,S,T> other; };
	///<summary>Invokes the merger</summary>
	template<class It>
	inline Token operator()(It& b, It e, Token outtoken)
	{
		if( active == 0 || b == e )
			return outtoken;
		if( active == 1 )
			return copy(b,e,outtoken);
		return invoke_invoke(b,e,outtoken,CCHECK<typename std::iterator_traits<It>::difference_type,typename std::iterator_traits<typename Stream::pointer>::difference_type>::ITERATOR_DIFFERENCES_MUST_BE_COMPATIBLE(e-b,stream[0]->end()-stream[0]->end()),*b,*stream[0]->begin());
	}
	///<summary>Invokes the merger</summary>
	template<class It, class Comp>
	inline Token operator()(It& b, It e, Token outtoken, Comp comp)
	{
		if( active == 0 )
			return outtoken;
		if( active == 1 )
			return copy(b,e,outtoken);
		return invoke_invoke(b,e,outtoken,comp,*b,*stream[0]->begin());
	}

private:
	template<class It, class T, class Diff>
	inline Token invoke_invoke(It& b, It e, Token outtoken, Diff min, T, T)
		{ std::less<T> comp; return invoke<It,T>(b,e,outtoken,min,comp); }
	template<class It, class Diff, class Comp, class T>
	inline Token invoke_invoke(It& b, It e, Token outtoken, Diff min, Comp& comp, T, T)
		{ return invoke<It,T>(b,e,outtoken,min,comp); }
	template<class It, class T, class Diff, class Comp>
	inline Token invoke(It& b, It e, Token outtoken, Diff min, Comp& comp)
	{
		struct heads { T t; typename Stream::pointer p; } head[order];
		for( basic_order_t i=0; i!=active; ++i )
		{
			head[i].p = stream[i]->begin();
			if( stream[i]->end()-head[i].p < min )
				min = stream[i]->end()-head[i].p;
			head[i].t = *head[i].p;
		}
		It p = b;
		heads *m = head;
		for( ;; )
		{
//			p += min;
//			min = -min;
			do
			{
				assert( min > 0 );
				m->t = *m->p;
				m = head;
				for( heads *q=head+1; q!=head+active; ++q )
					if( comp(q->t,m->t) )
						m = q;
//				*(p+min) = m->t;
				*p = m->t, ++p;
				m->p = m->p+1;
			}
			while( --min );
			min = stream[0]->end()-head[0].p;
			if( min == 0 )
			{
				for( basic_order_t j=0; j!=active; ++j )
					*stream[j] = Stream(head[j].p,stream[j]->end());
				b = p;
				return token[0];
			}
			for( basic_order_t i=1; i!=active; ++i )
				if( stream[i]->end()-head[i].p < min )
				{
					min = stream[i]->end()-head[i].p;
					if( min == 0 )
					{
						for( basic_order_t j=0; j<active; ++j )
							*stream[j] = Stream(head[j].p,stream[j]->end());
						b = p;
						return token[i];
					}
				}
			if( e-p < min )
			{
				min = e-p;
				if( min == 0 )
				{
					for( basic_order_t i=0; i<active; ++i )
						*stream[i] = Stream(head[i].p,stream[i]->end());
					b = e;
					return outtoken;
				}
			}
		}
	}
};


template<basic_order_t order, class Stream, class Token>
class simple_for_val_merger : public basic_merger_base<order,Stream,Token>
{
	using basic_merger_base<order,Stream,Token>::active;
	using basic_merger_base<order,Stream,Token>::stream;
	using basic_merger_base<order,Stream,Token>::token;
	using basic_merger_base<order,Stream,Token>::copy;
public:
	template<basic_order_t O, class S=Stream, class T=Token>
	struct rebind { typedef simple_for_val_merger<O,S,T> other; };
	///<summary>Invokes the merger</summary>
	template<class It>
	inline Token operator()(It& b, It e, Token outtoken)
	{
		if( active == 0 )
			return outtoken;
		if( active == 1 )
			return copy(b,e,outtoken);
		return invoke_invoke(b,e,outtoken,CCHECK<typename std::iterator_traits<It>::difference_type,typename std::iterator_traits<typename Stream::pointer>::difference_type>::ITERATOR_DIFFERENCES_MUST_BE_COMPATIBLE(e-b,stream[0]->end()-stream[0]->end()),*b,*stream[0]->begin());
	}
	///<summary>Invokes the merger</summary>
	template<class It, class Comp>
	inline Token operator()(It& b, It e, Token outtoken, Comp comp)
	{
		if( active == 0 )
			return outtoken;
		if( active == 1 )
			return copy(b,e,outtoken);
		return invoke_invoke(b,e,outtoken,comp,*b,*stream[0]->begin());
	}

private:
	template<class It, class T, class Diff>
	inline Token invoke_invoke(It& b, It e, Token outtoken, Diff min, T, T)
		{ std::less<T> comp; return invoke<It,T>(b,e,outtoken,min,comp); }
	template<class It, class Diff, class Comp, class T>
	inline Token invoke_invoke(It& b, It e, Token outtoken, Diff min, Comp& comp, T, T)
		{ return invoke<It,T>(b,e,outtoken,min,comp); }
	template<class It, class T, class Diff, class Comp>
	inline Token invoke(It& b, It e, Token outtoken, Diff min, Comp& comp)
	{
		struct ht
		{
			typename Stream::pointer h, t; T e;
		} s[order];
		for( basic_order_t i=0; i<active; ++i )
			s[i].t = stream[i]->end(), s[i].h = stream[i]->begin(), s[i].e = *s[i].h;
		It p = b;
		while( p != e )
		{
			ht *m = s;
			for( ht *q=s+1; q<s+active; ++q )
				if( comp(q->e,m->e) )
					m = q;
			*p = m->e, ++(m->h), ++p;
			if( m->h == m->t )
			{
				outtoken = token[m-s];
				break;
			}
			m->e = *m->h;
		}
		for( basic_order_t i=0; i<active; ++i )
			*stream[i] = Stream(s[i].h,stream[i]->end());
		b = p;
		return outtoken;
	}
};

template<basic_order_t order, class Stream, class Token>
class loser_merger : public basic_merger_base<order,Stream,Token>
{
	using basic_merger_base<order,Stream,Token>::active;
	using basic_merger_base<order,Stream,Token>::stream;
	using basic_merger_base<order,Stream,Token>::token;
	using basic_merger_base<order,Stream,Token>::copy;
public:
	template<basic_order_t O, class S=Stream, class T=Token>
	struct rebind { typedef loser_merger<O,S,T> other; };
	///<summary>Invokes the merger</summary>
	template<class It>
	inline Token operator()(It& b, It e, Token outtoken)
	{
		if( active == 0 )
			return outtoken;
		if( active == 1 )
			return copy(b,e,outtoken);
		return invoke_invoke(b,e,outtoken,CCHECK<typename std::iterator_traits<It>::difference_type,typename std::iterator_traits<typename Stream::pointer>::difference_type>::ITERATOR_DIFFERENCES_MUST_BE_COMPATIBLE(e-b,stream[0]->end()-stream[0]->end()),*b,*stream[0]->begin());
	}
	///<summary>Invokes the merger</summary>
	template<class It, class Comp>
	inline Token operator()(It& b, It e, Token outtoken, Comp comp)
	{
		if( active == 0 )
			return outtoken;
		if( active == 1 )
			return copy(b,e,outtoken);
		return invoke_invoke(b,e,outtoken,comp,*b,*stream[0]->begin());
	}

private:
	template<class T>
	class match
	{
	public:
		T t;
		basic_order_t i;
	};
	template<class T, class Comp>
	static match<T> build(typename Stream::pointer *head, match<T> *tree, Comp& comp, basic_order_t root_index)
	{
		if( 2*root_index+1 == order )
		{
			match<T> l = build(head,tree,comp,2*root_index);
			if( comp(l.t,*head[2*root_index+1-order]) )
			{
				tree[root_index].t = *head[2*root_index+1-order], tree[root_index].i=static_cast<basic_order_t>(2*root_index+1-order);
				return l;
			}
			else
			{
				match<T> ret;
				tree[root_index] = l;
				ret.t = *head[2*root_index+1-order], ret.i=static_cast<basic_order_t>(2*root_index+1-order);
				return ret;
			}
		}
		else if( 2*root_index < order )
		{
			match<T> l1 = build(head,tree,comp,2*root_index);
			match<T> l2 = build(head,tree,comp,2*root_index+1);
			if( comp(l1.t,l2.t) )
			{
				tree[root_index] = l2;
				return l1;
			}
			else
			{
				tree[root_index] = l1;
				return l2;
			}
		}
		else
		{
			match<T> ret;
			if( comp(*head[2*root_index-order],*head[2*root_index+1-order]) )
			{
				tree[root_index].t = *head[2*root_index+1-order], tree[root_index].i=static_cast<basic_order_t>(2*root_index+1-order);
				ret.t = *head[2*root_index-order], ret.i=static_cast<basic_order_t>(2*root_index-order);
			}
			else
			{
				tree[root_index].t = *head[2*root_index-order], tree[root_index].i=static_cast<basic_order_t>(2*root_index-order);
				ret.t = *head[2*root_index+1-order], ret.i=static_cast<basic_order_t>(2*root_index+1-order);
			}
			return ret;
		}
	}
	template<class T, class Comp, basic_order_t i>
	class find_min
	{
	public:
		inline void operator()(match<T> *tree, Comp& comp)
		{
			if( comp(tree[i].t,tree[0].t) )
			{
				match<T> tmp = tree[i];
				tree[i] = tree[0], tree[0] = tmp;
			}
			find_min<T,Comp,i/2> fm;
			fm(tree,comp);
		}
	};
	template<class T, class Comp>
	class find_min<T,Comp,1>
	{
	public:
		inline void operator()(match<T> *tree, Comp& comp)
		{
			if( comp(tree[1].t,tree[0].t) )
			{
				match<T> tmp = tree[1];
				tree[1] = tree[0], tree[0] = tmp;
			}
		}
	};
	template<class It, class T, class Diff>
	inline Token invoke_invoke(It& b, It e, Token outtoken, Diff min, T, T)
		{ std::less<T> comp; return invoke<It,T>(b,e,outtoken,min,comp); }
	template<class It, class Diff, class Comp, class T>
	inline Token invoke_invoke(It& b, It e, Token outtoken, Diff min, Comp& comp, T, T)
		{ return invoke<It,T>(b,e,outtoken,min,comp); }
	template<class It, class T, class Diff, class Comp>
	inline Token invoke(It& b, It e, Token outtoken, Diff min, Comp& comp)
	{
		if( active != order )
			return invoke_active<It,T,Diff,Comp>(b, e, outtoken, min, comp);
		register typename Stream::pointer head[order];
		match<T> tree[order];
		for( basic_order_t i=0; i<order; ++i )
		{
			head[i] = stream[i]->begin();
			if( stream[i]->end()-head[i] < min )
				min = stream[i]->end()-head[i];
		}

		register It p = b;
		tree[0] = build(head,tree,comp,1);
		for( ;; )
		{
			for( ; min; --min )
			{
				*p = tree[0].t, ++p;
				++head[tree[0].i];
				tree[0].t = *head[tree[0].i];
				switch( tree[0].i )
				{
				case 0:
					{
						find_min<T,Comp,(order+0)/2> fm;
						fm(tree,comp);
					}
					break;
				case 1:
					{
						find_min<T,Comp,(order+1)/2> fm;
						fm(tree,comp);
					}
					break;
				case 2:
					{
						find_min<T,Comp,(order+2)/2> fm;
						fm(tree,comp);
					}
					break;
				case 3:
					{
						find_min<T,Comp,(order+3)/2> fm;
						fm(tree,comp);
					}
					break;
				case 4:
					{
						find_min<T,Comp,(order+4)/2> fm;
						fm(tree,comp);
					}
					break;
				case 5:
					{
						find_min<T,Comp,(order+5)/2> fm;
						fm(tree,comp);
					}
					break;
				case 6:
					{
						find_min<T,Comp,(order+6)/2> fm;
						fm(tree,comp);
					}
					break;
				case 7:
					{
						find_min<T,Comp,(order+7)/2> fm;
						fm(tree,comp);
					}
					break;
				case 8:
					{
						find_min<T,Comp,(order+8)/2> fm;
						fm(tree,comp);
					}
					break;
				case 9:
					{
						find_min<T,Comp,(order+9)/2> fm;
						fm(tree,comp);
					}
					break;
				case 10:
					{
						find_min<T,Comp,(order+10)/2> fm;
						fm(tree,comp);
					}
					break;
				case 11:
					{
						find_min<T,Comp,(order+11)/2> fm;
						fm(tree,comp);
					}
					break;
				case 12:
					{
						find_min<T,Comp,(order+12)/2> fm;
						fm(tree,comp);
					}
					break;
				default:
					break;
				}
			}

			min = stream[0]->end()-head[0];
			if( min == 0 )
			{
				for( basic_order_t j=0; j<order; ++j )
					*stream[j] = Stream(head[j],stream[j]->end());
				b = p;
				return token[0];
			}
			for( basic_order_t i=1; i<order; ++i )
				if( stream[i]->end()-head[i] < min )
				{
					min = stream[i]->end()-head[i];
					if( min == 0 )
					{
						for( basic_order_t j=0; j<order; ++j )
							*stream[j] = Stream(head[j],stream[j]->end());
						b = p;
						return token[i];
					}
				}
			if( e-p < min )
			{
				min = e-p;
				if( min == 0 )
				{
					for( basic_order_t i=0; i<order; ++i )
						*stream[i] = Stream(head[i],stream[i]->end());
					b = e;
					return outtoken;
				}
			}
		}
	}
	template<class It, class T, class Diff, class Comp>
	Token invoke_active(It& b, It e, Token outtoken, Diff min, Comp& comp)
	{
		typename Stream::pointer head[order];
		for( basic_order_t i=0; i<active; ++i )
		{
			head[i] = stream[i]->begin();
			if( stream[i]->end()-head[i] < min )
				min = stream[i]->end()-head[i];
		}
		It p = b;
		for( ;; )
		{
			for( ; min; --min )
			{
				typename Stream::pointer *m = head;
				for( typename Stream::pointer *q=m+1; q<head+active; ++q )
					if( comp(**q,**m) )
						m = q;
				*p = **m, ++*m, ++p;
			}
			min = stream[0]->end()-head[0];
			if( min == 0 )
			{
				for( basic_order_t j=0; j<active; ++j )
					*stream[j] = Stream(head[j],stream[j]->end());
				b = p;
				return token[0];
			}
			for( basic_order_t i=1; i<active; ++i )
				if( stream[i]->end()-head[i] < min )
				{
					min = stream[i]->end()-head[i];
					if( min == 0 )
					{
						for( basic_order_t j=0; j<active; ++j )
							*stream[j] = Stream(head[j],stream[j]->end());
						b = p;
						return token[i];
					}
				}
			if( e-p < min )
			{
				min = e-p;
				if( min == 0 )
				{
					for( basic_order_t i=0; i<active; ++i )
						*stream[i] = Stream(head[i],stream[i]->end());
					b = e;
					return outtoken;
				}
			}
		}
	}
};

} // namespace funnel


#ifdef __GNUC__
#undef inline
#endif // __GNUC__
#ifdef _MSC_VER
#undef inline
#endif // _MSC_VER

#endif // __STACK_ALLOCATOR_H_INCLUDED_
