/*
 * Copyright (C) 2003 Kristoffer Vinther.
 * For contact information, please visit
 * http://www.brics.dk/~kv/thesis/.
 * If you use this code in commercial applications, please notify me.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * Free Software Foundation, Inc.
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __FUNNEL_BASE_H_INCLUDED_
#define __FUNNEL_BASE_H_INCLUDED_

#include <stddef.h>
#include <cmath>

#ifdef __GNUC__
#define inline inline __attribute__ ((always_inline))
#endif // __GNUC__
#ifdef _MSC_VER
#define inline __forceinline
#endif // _MSC_VER


namespace funnel
{

enum _tag_mmth { MAX_MERGE_TREE_HEIGHT = 16 };
typedef short height_t;
typedef int order_t;
typedef short basic_order_t;


template<basic_order_t order>
inline height_t logc(order_t k)
{
	height_t h = 0;
	for( order_t i=k-1; i; h++, i/=order );
	return h;
}
/* Computes floor(log()), not ceil(log()) :(  - Also beware of little and bigendianness(?)
template<>
inline height_t logc<2>(order_t k)
{
	float f = static_cast<float>(k);
	return static_cast<height_t>(((*reinterpret_cast<int*>(&f)) >> 23) - 127);
}*/
template<basic_order_t order>
inline order_t pow_of_order(height_t h)
{
	order_t k = 1;
	for( ; h; h--, k*=order );
	return k;
}
template<>
inline order_t pow_of_order<2>(height_t h)
	{ return 1<<h; }
template<>
inline order_t pow_of_order<4>(height_t h)
	{ return 1<<(2*h); }
template<>
inline order_t pow_of_order<8>(height_t h)
	{ return 1<<(3*h); }
template<>
inline order_t pow_of_order<16>(height_t h)
	{ return 1<<(4*h); }

template<basic_order_t order>
class bfs_index
{
public:
	inline bfs_index() : i(1) { }
	inline operator order_t() const
		{ return i; }

	inline bfs_index<order> operator=(const bfs_index<order>& rhs)
		{ i = rhs.i; return *this; }
	inline bool operator==(const bfs_index<order>& rhs) const
		{ return i == rhs.i; }
	inline bool operator!=(const bfs_index<order>& rhs) const
		{ return i != rhs.i; }

    inline void child(basic_order_t ch)
		{ i = static_cast<order_t>(order*(i-1)+ch+2); }
    inline void parent()
		{ /*assert( i>1 );*/ i = static_cast<order_t>((i-2)/order+1); }

	inline basic_order_t child_index() const
		{ /*assert( i>1 );*/ return static_cast<basic_order_t>((i+order-2) % order); }
private:
	order_t i;
};

} // namespace funnel


#ifdef __GNUC__
#undef inline
#endif // __GNUC__
#ifdef _MSC_VER
#undef inline
#endif // _MSC_VER

#endif // __FUNNEL_BASE_H_INCLUDED_
