/*
 * Copyright (C) 2003 Kristoffer Vinther.
 * For contact information, please visit
 * http://www.brics.dk/~kv/thesis/.
 * If you use this code in commercial applications, please notify me.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * Free Software Foundation, Inc.
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __FUNNEL_STRUCTS_H_INCLUDED_
#define __FUNNEL_STRUCTS_H_INCLUDED_

#include "funnel_base.h"

#ifdef __GNUC__
#define inline inline __attribute__ ((always_inline))
#endif // __GNUC__
#ifdef _MSC_VER
#define inline __forceinline
#endif // _MSC_VER


namespace funnel
{

template<class Pointer>
class buffer
{
public:
	typedef Pointer pointer;

	buffer() : head(), tail() { }
	buffer(Pointer begin, Pointer end) : head(begin), tail(end) { }

	pointer begin()
		{ return head; }
	pointer end()
		{ return tail; }
protected:
	Pointer head, tail;
};


template<basic_order_t order, class pointer, class nodepointer>
struct node_info
{
public:
	buffer<pointer> buffers[order];
	nodepointer children[order];
	basic_order_t n;
};

template<basic_order_t Order, class Pointer>
class node
{
public:
	typedef Pointer pointer;
	enum { order = Order };

	explicit inline node(basic_order_t) { } // Construct a leaf
	template<class node_pointer>
	explicit inline node(node_info<order,pointer,node_pointer>& ni) // Construct internal node
	{
		basic_order_t i;
		for( i=0; i!=ni.n; ++i )
			stream[i] = buffer<pointer>(ni.buffers[i].end(),ni.buffers[i].end());
		for( ; i!=order; ++i )
			stream[i] = buffer<pointer>();
	}

	inline buffer<pointer> *input(basic_order_t i)
		{ return stream+i; }

protected:
	buffer<pointer> stream[order];
};

template<basic_order_t order, class pointer>
class flip_node : public node<order,pointer>
{
	using node<order,pointer>::stream;
public:
	explicit inline flip_node(basic_order_t n) : node<order,pointer>(n) { }
	template<class nodepointer>
	inline flip_node(node_info<order,pointer,nodepointer>& ni) : node<order,pointer>(ni)
	{
		basic_order_t i;
		for( i=0; i!=ni.n; ++i )
		{
			begin[i] = ni.buffers[i].begin();
			end[i] = ni.buffers[i].end();
		}
		for( ; i!=order; ++i )
		{
			begin[i] = pointer();
			end[i] = pointer();
		}
	}

	inline void flip_buffer(basic_order_t i)
		{ stream[i] = buffer<pointer>(begin[i], stream[i].begin()); }
	inline void reset_buffer(basic_order_t i)
		{ stream[i] = buffer<pointer>(end[i],end[i]); }

	inline pointer input_begin(basic_order_t i)
		{ return begin[i]; }
	inline pointer input_end(basic_order_t i)
		{ return end[i]; }
protected:
	pointer begin[order], end[order];
};

template<basic_order_t order, class pointer>
class pointer_node : public node<order,pointer>
{
public:
	explicit inline pointer_node(basic_order_t k) : node<order,pointer>(k) // Construct a leaf
	{
		for( basic_order_t i=0; i!=order; i++ )
			children[i] = NULL;
	}
	explicit inline pointer_node(node_info<order,pointer,pointer_node<order,pointer>*>& ni) : node<order,pointer>(ni) // Construct internal node
	{
		basic_order_t i;
		for( i=0; i!=ni.n; i++ )
		{
			children[i] = ni.children[i];
			assert( children[i]+1 == reinterpret_cast<pointer_node*>(ni.buffers[i].begin()) );
		}
		for( ; i!=order; i++ )
			children[i] = NULL;
	}

	inline pointer_node *child(basic_order_t i)
		{ return children[i]; }

protected:
	pointer_node *children[order];
};

template<basic_order_t order, class pointer>
class pointer_flip_node : public flip_node<order,pointer>
{
public:
	explicit inline pointer_flip_node(basic_order_t k) : flip_node<order,pointer>(k) // Construct a leaf
	{
		for( basic_order_t i=0; i!=order; i++ )
			children[i] = NULL;
	}
	explicit inline pointer_flip_node(node_info<order,pointer,pointer_flip_node<order,pointer>*>& ni) : flip_node<order,pointer>(ni) // Construct internal node
	{
		basic_order_t i;
		for( i=0; i!=ni.n; i++ )
			children[i] = ni.children[i];
		for( ; i!=order; i++ )
			children[i] = NULL;
	}

	inline pointer_flip_node *child(basic_order_t i)
		{ return children[i]; }
protected:
	pointer_flip_node *children[order];
};

} // namespace funnel


#ifdef __GNUC__
#undef inline
#endif // __GNUC__
#ifdef _MSC_VER
#undef inline
#endif // _MSC_VER

#endif // __FUNNEL_STRUCTS_H_INCLUDED_
