/*
 * Copyright (C) 2003 Kristoffer Vinther.
 * For contact information, please visit
 * http://www.brics.dk/~kv/thesis/.
 * If you use this code in commercial applications, please notify me.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * Free Software Foundation, Inc.
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef LAYOUT_MBF_H_INCLUDED__
#define LAYOUT_MBF_H_INCLUDED__

#include "funnel_base.h"
#include "funnel_structs.h"
#include "layout_base.h"

#include "pointer_navigator.h"

#ifdef __GNUC__
#define inline inline __attribute__ ((always_inline))
#endif // __GNUC__
#ifdef _MSC_VER
#define inline __forceinline
#endif // _MSC_VER


namespace funnel
{

template<class Navigator, class Splitter, class T, class Allocator=std::allocator<T> >
class mixed_bf_layout : public mixed_layout_base<Navigator,Splitter,T,Allocator>
{
	using mixed_layout_base<Navigator,Splitter,T,Allocator>::compute_buffer_sizes;
	using mixed_layout_base<Navigator,Splitter,T,Allocator>::new_buffer;
	using mixed_layout_base<Navigator,Splitter,T,Allocator>::delete_buffer;
public:
	typedef typename Navigator::node node;

	static node *do_layout(order_t k, Allocator& alloc)
	{
		Navigator::REQUIRE_ALLOCATOR(alloc);
		MIXED_BF_LAYOUT_REQUIRES_NAVIGATOR(Navigator(NULL,k));
		height_t height = logc<Navigator::node::order>(k);
		size_t buffer_sizes[MAX_MERGE_TREE_HEIGHT];
		compute_buffer_sizes(buffer_sizes,buffer_sizes+height);
		typename Allocator::template rebind<T>::other buffer_alloc(alloc);
		typename Allocator::template rebind<node>::other node_alloc(alloc);
		node *root = do_layout(k, height, buffer_sizes, node_alloc, buffer_alloc);
		Navigator nav(root, k);
		typename Navigator::level_iterator leaf = nav.begin_level(height-1);
		while( k > 0 )
		{
			basic_order_t d;
			for( d=0; k>0 && d!=node::order; ++d, --k );
			node n(d);
			node_alloc.construct(*leaf, n);
			++leaf;
		}
		return root;
	}

	static void destroy(node *root, order_t k, Allocator& alloc)
	{
		height_t height = logc<node::order>(k);
		size_t buffer_sizes[MAX_MERGE_TREE_HEIGHT];
		compute_buffer_sizes(buffer_sizes,buffer_sizes+height);
		typename Allocator::template rebind<T>::other buffer_alloc(alloc);
		typename Allocator::template rebind<node>::other node_alloc(alloc);
		Navigator nav(root, k);
		typename Navigator::level_iterator leaf = nav.begin_level(height-1);
		for( order_t i=0; i<k/node::order; ++i )
		{
			--leaf;
			node_alloc.destroy(*leaf);
		}
		destroy(root, k, height, buffer_sizes, node_alloc, buffer_alloc);
	}

	static typename Allocator::size_type size_of(order_t k)
		{ return layout_base<Navigator,Splitter,T,Allocator>::size_of(k, logc<Navigator::node::order>(k)); }

protected:
	template<class NAlloc, class BAlloc>
	static node *do_layout(order_t k, height_t height, size_t *buffer_sizes, NAlloc& nalloc, BAlloc& balloc)
	{
		if( height > 1 )
		{
			node *root = do_layout(static_cast<order_t>((k+node::order-1)/node::order), height-1, buffer_sizes, nalloc, balloc);
			Navigator nav(root, static_cast<order_t>((k+node::order-1)/node::order));
			typename Navigator::level_iterator leaf = nav.begin_level(height-2);
			while( k > 0 )
			{
				node_info<node::order,typename BAlloc::pointer,typename NAlloc::pointer> info;
				for( info.n=0; k>0 && info.n!=node::order; ++info.n, k-=node::order )
				{
					typename BAlloc::pointer b = balloc.allocate(buffer_sizes[height-1]);
                    info.buffers[info.n] = buffer<typename BAlloc::pointer>(b,b+buffer_sizes[height-1]);
					info.children[info.n] = nalloc.allocate(1);
				}
				node n(info);
				nalloc.construct(*leaf, n);
				++leaf;
			}
			return root;
		}
		else
			return nalloc.allocate(1);
	}
	template<class NAlloc, class BAlloc>
	static void destroy(node *root, order_t k, height_t height, size_t *buffer_sizes, NAlloc& nalloc, BAlloc& balloc)
	{
		if( height > 1 )
		{
			k = static_cast<order_t>((k+node::order-1)/node::order);
			Navigator nav(root, k);
			typename Navigator::level_iterator leaf = nav.end_level(height-2);
			basic_order_t n = static_cast<basic_order_t>((k+node::order-1)%node::order);
			for( order_t i=0; i<(k+node::order-1)/node::order; ++i )
			{
				--leaf;
				for( ; n>=0; --n )
				{
					nalloc.deallocate(leaf.child(n),1);
					delete_buffer(leaf.input_begin(n), leaf.input_begin(n)+buffer_sizes[height-1], balloc);
				}
				n = node::order-1;
				nalloc.destroy(*leaf);
			}
			destroy(root, k, height-1, buffer_sizes, nalloc, balloc);
		}
		else
			nalloc.deallocate(root,1);
	}

	template<class Node>
	static inline void MIXED_BF_LAYOUT_REQUIRES_NAVIGATOR(const pointer_navigator<Node,Splitter>&) {}
	template<class Node>
	static inline void MIXED_BF_LAYOUT_REQUIRES_NAVIGATOR(const pointer_flip_navigator<Node,Splitter>&) {}
};

} // namespace funnel


#ifdef __GNUC__
#undef inline
#endif // __GNUC__
#ifdef _MSC_VER
#undef inline
#endif // _MSC_VER

#endif // LAYOUT_MBF_H_INCLUDED__
