/*
 * Copyright (C) 2003 Kristoffer Vinther.
 * For contact information, please visit
 * http://www.brics.dk/~kv/thesis/.
 * If you use this code in commercial applications, please notify me.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * Free Software Foundation, Inc.
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __MERGE_TREE_TYPES_H_INCLUDED_
#define __MERGE_TREE_TYPES_H_INCLUDED_

#include "merge_tree.h"
#include "layout_mveb.h"
#include "layout_mbf.h"
#include "layout_mdf.h"
#include "layout_pveb.h"
#include "layout_pbf.h"
#include "layout_pdf.h"
#include "pointer_navigator.h"

namespace funnel
{

template<class T, basic_order_t Order, class InStream, class BasicMerger = comp_merger<Order, buffer<int*>, int>, class Splitter=splitter<Order> >
struct merge_tree_types
{
private:
#ifdef NDEBUG
	typedef pointer_flip_node<Order,T*> pfn;
	typedef pointer_node<Order,T*> pn;
	typedef flip_node<Order,T*> fn;
	typedef node<Order,T*> n;
#else // NDEBUG
	typedef pointer_flip_node<Order,T*> pfn;
	typedef pointer_flip_node<Order,T*> pn;
	typedef pointer_flip_node<Order,T*> fn;
	typedef pointer_flip_node<Order,T*> n;
#endif // NDEBUG
	typedef Splitter sp;
	typedef std::allocator<T> heap;
	typedef stack_allocator<T> stack;
public:
	typedef merge_tree<T, InStream, mixed_veb_layout	<pointer_flip_navigator	<pfn,sp>, sp, T,	heap>, BasicMerger,Splitter> pb_heap_mveb;
	typedef merge_tree<T, InStream, mixed_veb_layout	<pointer_navigator		<pn,sp>, sp, T,		stack>, BasicMerger,Splitter> pb_stack_mveb;
	typedef merge_tree<T, InStream, mixed_veb_layout	<mixed_veb_navigator	<n,sp>, sp, T,		stack>, BasicMerger,Splitter> impl_mveb;
	typedef merge_tree<T, InStream, pooled_veb_layout	<pointer_flip_navigator	<pfn,sp>, sp, T,	heap>, BasicMerger,Splitter> pb_heap_veb;
	typedef merge_tree<T, InStream, pooled_veb_layout	<pointer_flip_navigator	<pfn,sp>, sp, T,	stack>, BasicMerger,Splitter> pb_stack_veb;
	typedef merge_tree<T, InStream, pooled_veb_layout	<pooled_veb_navigator	<fn,sp>, sp, T,		stack>, BasicMerger,Splitter> impl_veb;
	typedef merge_tree<T, InStream, mixed_df_layout		<pointer_flip_navigator	<pfn,sp>, sp, T,	heap>, BasicMerger,Splitter> pb_heap_mdf;
	typedef merge_tree<T, InStream, mixed_df_layout		<pointer_navigator		<pn,sp>, sp, T,		stack>, BasicMerger,Splitter> pb_stack_mdf;
	typedef merge_tree<T, InStream, pooled_df_layout	<pointer_flip_navigator	<pfn,sp>, sp, T,	heap>, BasicMerger,Splitter> pb_heap_df;
	typedef merge_tree<T, InStream, pooled_df_layout	<pointer_flip_navigator	<pfn,sp>, sp, T,	stack>, BasicMerger,Splitter> pb_stack_df;
	typedef merge_tree<T, InStream, pooled_df_layout	<pooled_df_navigator	<fn,sp>, sp, T,		stack>, BasicMerger,Splitter> impl_df;
	typedef merge_tree<T, InStream, mixed_bf_layout		<pointer_flip_navigator	<pfn,sp>, sp, T,	heap>, BasicMerger,Splitter> pb_heap_mbf;
	typedef merge_tree<T, InStream, mixed_bf_layout		<pointer_navigator		<pn,sp>, sp, T,		stack>, BasicMerger,Splitter> pb_stack_mbf;
	typedef merge_tree<T, InStream, pooled_bf_layout	<pointer_flip_navigator	<pfn,sp>, sp, T,	heap>, BasicMerger,Splitter> pb_heap_bf;
	typedef merge_tree<T, InStream, pooled_bf_layout	<pointer_flip_navigator	<pfn,sp>, sp, T,	stack>, BasicMerger,Splitter> pb_stack_bf;
	typedef merge_tree<T, InStream, pooled_bf_layout	<pooled_bf_navigator	<fn,sp>, sp, T,		stack>, BasicMerger,Splitter> impl_bf;

	typedef rec_merge_tree<T, InStream, mixed_veb_layout	<pointer_flip_navigator	<pfn,sp>, sp, T, heap>, BasicMerger,Splitter> rec_heap_mveb;
	typedef rec_merge_tree<T, InStream, mixed_veb_layout	<pointer_flip_navigator	<pfn,sp>, sp, T, stack>, BasicMerger,Splitter> rec_stack_mveb;
	typedef rec_merge_tree<T, InStream, pooled_veb_layout	<pointer_flip_navigator	<pfn,sp>, sp, T, heap>, BasicMerger,Splitter> rec_heap_veb;
	typedef rec_merge_tree<T, InStream, pooled_veb_layout	<pointer_flip_navigator	<pfn,sp>, sp, T, stack>, BasicMerger,Splitter> rec_stack_veb;
	typedef rec_merge_tree<T, InStream, mixed_df_layout		<pointer_flip_navigator	<pfn,sp>, sp, T, heap>, BasicMerger,Splitter> rec_heap_mdf;
	typedef rec_merge_tree<T, InStream, mixed_df_layout		<pointer_flip_navigator	<pfn,sp>, sp, T, stack>, BasicMerger,Splitter> rec_stack_mdf;
	typedef rec_merge_tree<T, InStream, pooled_df_layout	<pointer_flip_navigator	<pfn,sp>, sp, T, heap>, BasicMerger,Splitter> rec_heap_df;
	typedef rec_merge_tree<T, InStream, pooled_df_layout	<pointer_flip_navigator	<pfn,sp>, sp, T, stack>, BasicMerger,Splitter> rec_stack_df;
	typedef rec_merge_tree<T, InStream, mixed_bf_layout		<pointer_flip_navigator	<pfn,sp>, sp, T, heap>, BasicMerger,Splitter> rec_heap_mbf;
	typedef rec_merge_tree<T, InStream, mixed_bf_layout		<pointer_flip_navigator	<pfn,sp>, sp, T, stack>, BasicMerger,Splitter> rec_stack_mbf;
	typedef rec_merge_tree<T, InStream, pooled_bf_layout	<pointer_flip_navigator	<pfn,sp>, sp, T, heap>, BasicMerger,Splitter> rec_heap_bf;
	typedef rec_merge_tree<T, InStream, pooled_bf_layout	<pointer_flip_navigator	<pfn,sp>, sp, T, stack>, BasicMerger,Splitter> rec_stack_bf;

	typedef rec_merge_tree<T, InStream, mixed_veb_layout	<pointer_flip_navigator	<pfn,sp>, sp, T, stack>, BasicMerger,Splitter> fast;
};

} // namespace funnel

#endif // __MERGE_TREE_TYPES_H_INCLUDED_
