/*
 * Copyright (C) 2003 Kristoffer Vinther.
 * For contact information, please visit
 * http://www.brics.dk/~kv/thesis/.
 * If you use this code in commercial applications, please notify me.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * Free Software Foundation, Inc.
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __RANDOM_H_INCLUDED_
#define __RANDOM_H_INCLUDED_

#include <ctime>
#include <cstdlib>

class random_source
{
public:
#ifndef _DEBUG
	inline random_source()
	{
#ifdef WIN32
		srand((unsigned)time(NULL));
#else // WIN32
		srand48((long int)time(NULL));
#endif // WIN32
	}
#endif// _DEBUG
};
template<class T>
inline random_source& operator>>(random_source& s, T& t)
{
	double d;
	s >> d;
	t = factory<T>::from_scalar(d);
	return s;
}
#ifdef WIN32
template<>
inline random_source& operator>>(random_source& s, int& t)
{
	t = rand()^(rand()<<8)^(rand()<<16)^(rand()<<24);
	return s;
}
template<>
inline random_source& operator>>(random_source& s, double& t)
{
	int i;
	s >> i;
	t = static_cast<double>(i)/(static_cast<double>(std::numeric_limits<int>::max())-std::numeric_limits<int>::min())+0.5;
	return s;
}
#else // WIN32
template<>
inline random_source& operator>>(random_source& s, double& t)
{
	t = drand48();
	return s;
}
template<>
inline random_source& operator>>(random_source& s, int& t)
{
	t = mrand48();
	return s;
}
template<>
inline random_source& operator>>(random_source& s, unsigned int& t)
{
	t = static_cast<unsigned int>(lrand48());
	return s;
}
#endif // WIN32

#endif // __RANDOM_H_INCLUDED_
