/*
 * Copyright (C) 2003 Kristoffer Vinther.
 * For contact information, please visit
 * http://www.brics.dk/~kv/thesis/.
 * If you use this code in commercial applications, please notify me.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * Free Software Foundation, Inc.
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "sortgen.h"
#include <exception>
#include <iostream>
#include <cstdio>

void exit_usage()
{
	std::cerr << "sortgen -e{i|p|r} -d{u|m|M|d|D|e|t|k} size name" << std::endl
		<< std::endl
		<< "  u: uniform distribution" << std::endl
		<< "  m: almost sorted; missing merge-phase; log(size) runs" << std::endl
		<< "  M: almost sorted; missing merge-phase; sqrt(size) runs" << std::endl
		<< "  d: almost sorted; distributed; log(size) sized bucket" << std::endl
		<< "  D: almost sorted; distributed; sqrt(size) sized  bucket" << std::endl
		<< "  e: few distinct; log(size) distinct elements" << std::endl
		<< "  E: few distinct; sqrt(size) distinct elements" << std::endl
		<< "  t: unbalanced" << std::endl
		<< "  k: partition killer" << std::endl;
	exit(1);
}

int main(int argc, char *argv[])
{
	if( argc != 5 )
		exit_usage();
	if( argv[1][0] != '-' || argv[2][0] != '-' || argv[1][1] != 'e' || argv[2][1] != 'd' )
		exit_usage();
	try
	{
		switch( argv[1][2] )
		{
		case 'i':
			sortgen<int>(argv[4], atoi(argv[3]), argv[2][2]);
			break;
		case 'p':
			sortgen<element>(argv[4], atoi(argv[3]), argv[2][2]);
			break;
		case 'r':
			sortgen<record>(argv[4], atoi(argv[3]), argv[2][2]);
			break;
		default:
			exit_usage();
		}
	}
	catch( std::exception& e )
	{
		std::cerr << "Exception in sortgen: " << e.what() << std::endl;
		return 1;
	}
	return 0;
}
