
#include "../../Bench/aarray.h"
#include "../../Bench/elements.h"
#include "../../Bench/stopwatch.h"
#include <exception>
#include <iostream>
#include "../../Sort/Sort.h"
#include "../../Sort/funnel/layout_mveb.h"
#include "../../Sort/funnel/pointer_navigator.h"
#include "../../Sort/funnel/merge_tree.h"
#include "../../Sort/funnel/stack_allocator.h"

void exit_usage()
{
	std::cerr << "ipfunnelsort -e{i|p|r} input [output]" << std::endl;
	exit(1);
}

stopwatch w;
pf_counter pfc;
miss_counter mc;

template<class T>
void do_sort(const char *in, const char *out)
{
	using namespace funnel;
	typedef rec_merge_tree<T, buffer<T*>, mixed_veb_layout<pointer_flip_navigator<pointer_flip_node<4,T*>,splitter<4> >, splitter<4>, T, stack_allocator<T> >, comp_merger<4, buffer<T*>, T> > merger;
	
	typename auto_array::map_array<T> A(in, auto_array::MAP_EXISTING | auto_array::MAP_READ | auto_array::MAP_WRITE);
	pfc.start(); mc.start(); w.start();
	iosort::inplace_merge_sort<merger,splitter<4> >(A.begin(), A.end());
	w.stop(); mc.stop(); pfc.stop();
	if( out )
	{
		typename auto_array::map_array<T> B(out, auto_array::MAP_CREATE | auto_array::MAP_READ | auto_array::MAP_WRITE, A.end()-A.begin());
		pfc.start(); mc.start(); w.start();
		std::copy(A.begin(), A.end(), B.begin());
		w.stop(); mc.stop(); pfc.stop();
	}
}

int main(int argc, char *argv[])
{
	if( argc < 3 || argc > 4 )
		exit_usage();
	if( argv[1][0] != '-' || argv[1][1] != 'e' )
		exit_usage();
	try
	{
		switch( argv[1][2] )
		{
			case 'i':
				do_sort<int>(argv[2], argc==4? argv[3] : NULL);
				break;
			case 'p':
				do_sort<element>(argv[2], argc==4? argv[3] : NULL);
				break;
			case 'r':
				do_sort<record>(argv[2], argc==4? argv[3] : NULL);
				break;
			break;
		default:
			exit_usage();
		}
		std::cout << sys_names::compiler() << std::endl;
		std::cout << w.xmltag() << std::endl;
		std::cout << pfc.xmltag() << std::endl;
		std::cout << mc.xmltag() << std::endl;
	}
	catch( std::exception& e )
	{
		std::cerr << "Exception in funnelsort: " << e.what() << std::endl;
		return 1;
	}
	return 0;
}
