/*
 * Copyright (C) 2003 Kristoffer Vinther
 * This code is based largely on code developed by
 *
 *   Li Xiao, Xiaodong Zhang, and Stefan A. Kubricht
 *
 * available at
 * http://www.jea.acm.org/2000/XiaoMemory/
 */

#include "../../Bench/aarray.h"
#include "../../Bench/elements.h"
#include "../../Bench/stopwatch.h"
#include <exception>
#include <iostream>

#ifdef WIN32
#define OUT_FILE "C:\\Documents and Settings\\kv\\Local Settings\\Temp\\sorted.img"
#else // WIN32

#ifndef IMAGEDIR
#define IMAGEDIR "/tmp/"
#endif // IMAGEDIR
#define OUT_FILE IMAGEDIR "sorted.img"

#endif // WIN32

void exit_usage()
{
	std::cerr << "msort-c -e{i|p|r} input [output]" << std::endl;
	exit(1);
}

stopwatch watch;
pf_counter pfc;
miss_counter mc;

template<class T>
void do_sort(const char *in, const char *out)
{
	typename auto_array::map_array<T> source,target,xchg;
	size_t n,cachesize,pagesize;
	size_t i,j1,j2,jj2,jj3,k,kk,m,w,it=0,min;

	typename auto_array::map_array<T> a(in, auto_array::MAP_EXISTING | auto_array::MAP_READ | auto_array::MAP_WRITE);
	n = a.end()-a.begin();
	std::cout << "\t\t<size>" << n << "</size>" << std::endl;
#ifdef _MIPS_ISA
	cachesize = 1024*1024/sizeof(T)/2;
	pagesize = 8192;
#else // _MIPS_ISA
	cachesize = 256*1024/sizeof(T)/2;
	pagesize = 4096;
#endif // _MIPS_ISA
	typename auto_array::map_array<T> b(out, auto_array::MAP_CREATE | auto_array::MAP_READ | auto_array::MAP_WRITE, n);

	pfc.start(); mc.start(); watch.start();
	min = cachesize;
	if( n < min )
		min = n;
  kk=0;
  do {
    jj2=kk;
    jj3=jj2+min;
    for (i=jj2; i<jj3; i=i+4)
      for (j1=i; j1<i+3;j1++)
	for (j2=j1+1; j2<i+4; j2++) {
	  if (a[j1]>a[j2]) {
	    T xt=a[j1];
	    a[j1]=a[j2];
	    a[j2]=xt;
	  }
	}

    source=a;
    target=b;
    w=2;
    m=min/(2*w);
    while (2*w<min) {
      k=kk;
      w=w*2;
      m=m/2;
      j2=kk;
      for (i=0;i<m;i++){
	j1=j2;
	j2=j2+w;
	jj2=j2;
	jj3=j2+w;
	while ((j1<jj2)&&(j2<jj3)) {
	  if (source[j1]<source[j2])
	    target[k++]=source[j1++];
	  else
	    target[k++]=source[j2++];
	}
	if (j1==jj2) {
	  while (j2<jj3)
	    target[k++]=source[j2++];
	}
	if (j2==jj3) {
	  while (j1<jj2)
	    target[k++]=source[j1++];
	}
      }
      xchg=source;
      source=target;
      target=xchg;      
    }

    kk=kk+min;
    it++;
  }while(it<n/cachesize);

  // second phase: base mergesort

  m=n/(2*w);
  while (2*w<n) {
    k=0;
    w=w*2;
    m=m/2;
    j2=0;
    for (i=0;i<m;i++) {
      j1=j2;
      j2=j2+w;
      jj2=j2;
      jj3=j2+w;
      while ((j1<jj2)&&(j2<jj3)) {
	if (source[j1]<source[j2])
	  target[k++]=source[j1++];
	else
	  target[k++]=source[j2++];
      }
      if (j1==jj2) {
	while (j2<jj3)
	  target[k++]=source[j2++];
      }
      if (j2==jj3) {
	while (j1<jj2)
	  target[k++]=source[j1++];
      }
    }
    xchg=source;
    source=target;
    target=xchg;
  }
	watch.stop(); mc.stop(); pfc.stop();
}

int main(int argc, char *argv[])
{
	if( argc < 3 || argc > 4 )
		exit_usage();
	if( argv[1][0] != '-' || argv[1][1] != 'e' )
		exit_usage();
	std::cout << "\t<run buildtime=\"" << sys_names::buildtime() << "\" compiler=\"" << sys_names::compiler() << "\" execstatus=\"0\">" << std::endl;
	std::cout << "\t\t<algorithm>msort-c</algorithm>" << std::endl;
	try
	{
		switch( argv[1][2] )
		{
			case 'i':
				do_sort<int>(argv[2], argc==4? argv[3] : NULL);
				break;
			case 'p':
				do_sort<element>(argv[2], argc==4? argv[3] : NULL);
				break;
			case 'r':
				do_sort<record>(argv[2], argc==4? argv[3] : NULL);
				break;
			break;
		default:
			exit_usage();
		}
		std::cout << watch.xmltag(2) << std::endl;
		std::cout << pfc.xmltag(2) << std::endl;
		std::cout << mc.xmltag(2) << std::endl;
	}
	catch( std::exception& ex )
	{
		std::cout << "\t\t<error>Caught std::exception: " << ex.what() << "</error>" << std::endl;
		return 1;
	}
	std::cout << "\t</run>" << std::endl;
	return 0;
}
