/*
 * Copyright (C) 2003 Kristoffer Vinther
 * This code is based largely on code developed by
 *
 *   Li Xiao, Xiaodong Zhang, and Stefan A. Kubricht
 *
 * available at
 * http://www.jea.acm.org/2000/XiaoMemory/
 */

#include "../../Bench/aarray.h"
#include "../../Bench/elements.h"
#include "../../Bench/stopwatch.h"
#include <exception>
#include <iostream>

#ifdef WIN32
#define OUT_FILE "C:\\Documents and Settings\\kv\\Local Settings\\Temp\\sorted.img"
#else // WIN32

#ifndef IMAGEDIR
#define IMAGEDIR "/tmp/"
#endif // IMAGEDIR
#define OUT_FILE IMAGEDIR "sorted.img"

#endif // WIN32

template<class T>
struct PQUEUE {
  T c;
  int index0;
  int index;
};

void exit_usage()
{
	std::cerr << "msort-m -e{i|p|r} input [output]" << std::endl;
	exit(1);
}

stopwatch watch;
pf_counter pfc;
miss_counter mc;

template<class T>
void do_sort(const char *in, const char *out)
{
	typedef PQUEUE<T> QUEUE;

	QUEUE *queue, tmp;
	typename auto_array::map_array<T> source,target,xchg;
	size_t n,cachesize,pagesize;
	size_t i,j1,j2,jj2,jj3,k,kk,nn,m,w,it=0,min,change;
	size_t current,parent, childl, childr;

	typename auto_array::map_array<T> a(in, auto_array::MAP_EXISTING | auto_array::MAP_READ | auto_array::MAP_WRITE);
	n = a.end()-a.begin();
	std::cout << "\t\t<size>" << n << "</size>" << std::endl;
#ifdef _MIPS_ISA
	cachesize = 1024*1024/sizeof(T)/2;
	pagesize = 8192;
#else // _MIPS_ISA
	cachesize = 256*1024/sizeof(T)/2;
	pagesize = 4096;
#endif // _MIPS_ISA
	nn = n/cachesize;
	typename auto_array::map_array<T> b(out, auto_array::MAP_CREATE | auto_array::MAP_READ | auto_array::MAP_WRITE, n);
	queue = new QUEUE[nn];

	pfc.start(); mc.start(); watch.start();
	min = cachesize;
	if( n < min )
		min = n;
  kk=0;
  do {
    jj2=kk;
    jj3=jj2+min;
    for (i=jj2; i<jj3; i=i+4)
      for (j1=i; j1<i+3;j1++)
	for (j2=j1+1; j2<i+4; j2++) {
	  if (a[j1]>a[j2]) {
	    T xt=a[j1];
	    a[j1]=a[j2];
	    a[j2]=xt;
	  }
	}

    source=a;
    target=b;
    w=2;
    m=min/(2*w);
    while (2*w<min) {
      k=kk;
      w=w*2;
      m=m/2;
      j2=kk;
      for (i=0;i<m;i++){
	j1=j2;
	j2=j2+w;
	jj2=j2;
	jj3=j2+w;
	while ((j1<jj2)&&(j2<jj3)) {
	  if (source[j1]<source[j2])
	    target[k++]=source[j1++];
	  else
	    target[k++]=source[j2++];
	}
	if (j1==jj2) {
	  while (j2<jj3)
	    target[k++]=source[j2++];
	}
	if (j2==jj3) {
	  while (j1<jj2)
	    target[k++]=source[j1++];
	}
      }
      xchg=source;
      source=target;
      target=xchg;
    }
    kk=kk+min;
    it++;
  }while(it<nn);

  if((2*w==n)&&(target==a))
    for (i=0; i<n; i++)
      a[i] = b[i];

  // second phase: multi-mergesort

  if(2*w<n) {
    if(target==b)
      for (i=0; i<n; i++)
	b[i] = a[i];

    for (i=0;i<nn;i++)
      queue[i].index=i*cachesize;

    // build a priority queue

    queue[0].c=b[0];
    for (i=1; i<nn;i++) {
      queue[i].c=b[queue[i].index];
      parent=i;
      while(parent>0) {
	current=parent;
	parent=(parent-1)/2;
	if(queue[current].c<queue[parent].c) {
	  tmp=queue[parent];
	  queue[parent]=queue[current];
	  queue[current]=tmp;
	}
      }
    }

    // remove-min, add new element and  keep heap

    k=0;
    while(k<n) {
      a[k]=queue[0].c;
      k++;
      queue[0].index++;
      if(queue[0].index%cachesize==0)
		  queue[0].c=std::numeric_limits<T>::max();
      else
	queue[0].c=b[queue[0].index];
      current=0;
      change=1;
      while(change==1) {
	change=0;
	childl=2*current+1;
	if (childl>=nn) break;
	childr=childl+1;
	min=childl;
	if ((queue[childr].c<queue[childl].c)&&(childr<nn))
	  min=childr;
	if (queue[current].c>queue[min].c) {
	  tmp=queue[current];
	  queue[current]=queue[min];
	  queue[min]=tmp;
	  current=min;
	  change=1;
	}
      }
    }
  }   
	watch.stop(); mc.stop(); pfc.stop();
}

int main(int argc, char *argv[])
{
	if( argc < 3 || argc > 4 )
		exit_usage();
	if( argv[1][0] != '-' || argv[1][1] != 'e' )
		exit_usage();
	std::cout << "\t<run buildtime=\"" << sys_names::buildtime() << "\" compiler=\"" << sys_names::compiler() << "\" execstatus=\"0\">" << std::endl;
	std::cout << "\t\t<algorithm>msort-m</algorithm>" << std::endl;
	try
	{
		switch( argv[1][2] )
		{
			case 'i':
				do_sort<int>(argv[2], argc==4? argv[3] : NULL);
				break;
			case 'p':
				do_sort<element>(argv[2], argc==4? argv[3] : NULL);
				break;
			case 'r':
				do_sort<record>(argv[2], argc==4? argv[3] : NULL);
				break;
			break;
		default:
			exit_usage();
		}
		std::cout << watch.xmltag(2) << std::endl;
		std::cout << pfc.xmltag(2) << std::endl;
		std::cout << mc.xmltag(2) << std::endl;
	}
	catch( std::exception& ex )
	{
		std::cout << "\t\t<error>Caught std::exception: " << ex.what() << "</error>" << std::endl;
		return 1;
	}
	std::cout << "\t</run>" << std::endl;
	return 0;
}
