
#include "../../Bench/aarray.h"
#include "../../Bench/elements.h"
#include "../../Bench/stopwatch.h"
#include <exception>
#include <iostream>
#include <algorithm>

void exit_usage()
{
	std::cerr << "stdsort -e{i|p|r} input [output]" << std::endl;
	exit(1);
}

stopwatch w;
pf_counter pfc;
miss_counter mc;

template<class T>
void do_sort(const char *in, const char *out)
{
	typename auto_array::map_array<T> A(in, auto_array::MAP_EXISTING | auto_array::MAP_READ | auto_array::MAP_WRITE);
	std::cout << "\t\t<size>" << A.end()-A.begin() << "</size>" << std::endl;
	pfc.start(); mc.start(); w.start();
	std::sort(A.begin(), A.end());
	w.stop(); mc.stop(); pfc.stop();
/*	if( out )
	{
		typename auto_array::map_array<T> B(out, auto_array::MAP_CREATE | auto_array::MAP_READ | auto_array::MAP_WRITE, A.end()-A.begin());
		pfc.start(); mc.start(); w.start();
		std::copy(A.begin(), A.end(), B.begin());
		w.stop(); mc.stop(); pfc.stop();
	}
*/}

int main(int argc, char *argv[])
{
	if( argc < 3 || argc > 4 )
		exit_usage();
	if( argv[1][0] != '-' || argv[1][1] != 'e' )
		exit_usage();
	std::cout << "\t<run buildtime=\"" << sys_names::buildtime() << "\" compiler=\"" << sys_names::compiler() << "\" execstatus=\"0\">" << std::endl;
	std::cout << "\t\t<algorithm>stdsort</algorithm>" << std::endl;
	try
	{
		switch( argv[1][2] )
		{
			case 'i':
				do_sort<int>(argv[2], argc==4? argv[3] : NULL);
				break;
			case 'p':
				do_sort<element>(argv[2], argc==4? argv[3] : NULL);
				break;
			case 'r':
				do_sort<record>(argv[2], argc==4? argv[3] : NULL);
				break;
			break;
		default:
			exit_usage();
		}
		std::cout << w.xmltag(2) << std::endl;
		std::cout << pfc.xmltag(2) << std::endl;
		std::cout << mc.xmltag(2) << std::endl;
	}
	catch( std::exception& ex )
	{
		std::cout << "\t\t<error>Caught std::exception: " << ex.what() << "</error>" << std::endl;
		return 1;
	}
	std::cout << "\t</run>" << std::endl;
	return 0;
}
