#define BTE_COLLECTION_IMP_MMAP
#include <ami.H>
#include <ami_stream.H>
#include <ami_scan.H>
#include <ami_sort.H>

#include "elements.h"
#include "stopwatch.h"
#include <iostream.h>
#include <time.h>
#include <stdlib.h>

#ifdef WIN32
#define OUT_FILE "C:\\Documents and Settings\\kv\\Local Settings\\Temp\\sorted.img"
#else // WIN32

#ifndef IMAGEDIR
#define IMAGEDIR "/tmp/"
#endif // IMAGEDIR
#define OUT_FILE IMAGEDIR "sorted.img"

#endif // WIN32

void exit_usage()
{
	cerr << "ami_sort -e{i|p|r} input [output]" << std::endl;
	exit(1);
}

stopwatch w;
pf_counter pfc;
miss_counter mc;

template<class T>
void do_sort(const char *in, const char *out)
{
	MM_manager.set_memory_limit (196*1024*1024);
	cout << "\t\t<algorithm>ami_sort</algorithm>" << endl;
	AMI_STREAM<T> ins(in);
	if( !ins )
	{
		cout << "\t\t<error>!instream</error>" << endl;
		exit(1);
	}
	cout << "\t\t<size>" << ins.stream_len() << "</size>" << endl;
	if( out )
	{
		AMI_STREAM<T> outs(out);
		if( !outs )
		{
			cout << "\t\t<error>!outstream</error>" << endl;
			exit(1);
		}
		pfc.start(); mc.start(); w.start();
		if( AMI_sort(&ins, &outs) != AMI_ERROR_NO_ERROR )
			cout << "\t\t<error>ae != AMI_ERROR_NO_ERROR</error>" << endl;
		w.stop(); mc.stop(); pfc.stop();
	}
	else
	{
		AMI_STREAM<T> outs(OUT_FILE);
		if( !outs )
		{
			cout << "\t\t<error>!outstream</error>" << endl;
			exit(1);
		}
		pfc.start(); mc.start(); w.start();
		if( AMI_sort(&ins, &outs) != AMI_ERROR_NO_ERROR )
			cout << "\t\t<error>ae != AMI_ERROR_NO_ERROR</error>" << endl;
//		std::copy(B.begin(), B.end(), A.begin());
		w.stop(); mc.stop(); pfc.stop();
	}
}

class scan_random : AMI_scan_object
{
private:
    unsigned int max, remaining;
public:
	scan_random(unsigned int count = 1000, int seed = 17) : max(count), remaining(count) { }
	virtual ~scan_random() { }
    AMI_err initialize()
	{
		srand48((long int)time(NULL));
		return AMI_ERROR_NO_ERROR;
	}
    AMI_err operate(int *out1, AMI_SCAN_FLAG *sf)
	{
		if( (*sf = (remaining-- != 0)) )
		{
			*out1 = random();
			return AMI_SCAN_CONTINUE;
		}
		else
			return AMI_SCAN_DONE;
	}
    AMI_err operate(element *out1, AMI_SCAN_FLAG *sf)
	{
		if( (*sf = (remaining-- != 0)) )
		{
			*out1 = element(random(),static_cast<char*>(NULL)+0x600DF00D);
			return AMI_SCAN_CONTINUE;
		}
		else
			return AMI_SCAN_DONE;
	}
    AMI_err operate(record *out1, AMI_SCAN_FLAG *sf)
	{
		if( (*sf = (remaining-- != 0)) )
		{
//			*out1 = element(random(),static_cast<char*>(NULL)+0x600DF00D);
			return AMI_SCAN_CONTINUE;
		}
		else
			return AMI_SCAN_DONE;
	}
};

template<class T>
void gen_sort(const char *out, size_t test_size)
{
	AMI_STREAM<T> outs(out);
	if( !outs )
	{
		cout << "\t\t<error>sortgen outs</error>" << endl;
		return;
	}
	scan_random rnds(test_size);
	if( AMI_scan(&rnds, &outs) != AMI_ERROR_NO_ERROR )
			cout << "\t\t<error>ae != AMI_ERROR_NO_ERROR</error>" << endl;
}

int main(int argc, char *argv[])
{
	if( argc < 3 || argc > 5 )
		exit_usage();
	if( argv[1][0] != '-' ||  argv[1][1] != 'e' )
		exit_usage();
	cout << "\t<run buildtime=\"" << sys_names::buildtime() << "\" compiler=\"" << sys_names::compiler() << "\" execstatus=\"0\">" << endl;
//	try
	{
		switch( argv[1][2] )
		{
			case 'i':
				if( argv[2][0] == '-' && argv[2][1] == 'g' && argv[2][2] == '\0' )
					gen_sort<int>(argv[4],atoi(argv[3]));
				else
					do_sort<int>(argv[2], argc==4? argv[3] : NULL);
				break;
			case 'p':
				if( argv[2][0] == '-' && argv[2][1] == 'g' && argv[2][2] == '\0' )
					gen_sort<element>(argv[4],atoi(argv[3]));
				else
					do_sort<element>(argv[2], argc==4? argv[3] : NULL);
				break;
			case 'r':
				if( argv[2][0] == '-' && argv[2][1] == 'g' && argv[2][2] == '\0' )
					gen_sort<record>(argv[4],atoi(argv[3]));
				else
					do_sort<record>(argv[2], argc==4? argv[3] : NULL);
				break;
			break;
		default:
			exit_usage();
		}
		cout << w.xmltag(2) << endl;
		cout << pfc.xmltag(2) << endl;
		cout << mc.xmltag(2) << endl;
	}
/*	catch( std::exception& e )
	{
		std::cerr << "Exception in TPIE sort: " << e.what() << std::endl;
		return 1;
	}
*/	cout << "\t</run>" << endl;
	return 0;
}
