/*
 * Copyright (C) 2003 Kristoffer Vinther.
 * For contact information, please visit
 * http://www.brics.dk/~kv/thesis/.
 * If you use this code in commercial applications, please notify me.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * Free Software Foundation, Inc.
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <cmath>
#include <cassert>
#include <iostream>
#include "Test.h"

void checkSqrt()
{
	for( int k=4; k<27; k++ )
	{
		int kp = k;
		int i = 0;
		int check = 0;
		std::cout << k << ": ";
		int f = (int)floor(sqrt((double)k));
		int c = (int)ceil(sqrt((double)k));
		while( kp > 0 && (kp%f != 0 || kp/f+i>c) )
		{
			std::cout << c << ' ';
			kp -= c;
			check += c;
			i++;
		}
		while( kp > 0 )
		{
			std::cout << f << ' ';
			kp -= f;
			check += f;
			i++;
		}
		assert( k == check );
		assert( f <= i && i <= c );
		std::cout << "-- " << k << " = " << check << "; " << f << "<=" << i << "<=" << c << std::endl;
		if( k!=check || f>i || i>c )
			std::cout << "Fuckup at k=" << k << "!" << std::endl;
	}
}

void checkCub()
{
	for( int n=8; n<260; n++ )
	{
		int np = n;
		int i = 0;
		int check = 0;
		std::cout << n << ": ";
		int fi = (int)floor(pow((double)n, 1.0/3));
		int ci = (int)ceil(pow((double)n, 1.0/3));

		int k = (int)floor(pow((double)n, 1.0/3)+.5);
		int f = n/k;
		int c = f+1;
		while( np > 0 && (np%f != 0 || np/f+i>k) )
		{
			std::cout << c << ' ';
			np -= c;
			check += c;
			i++;
		}
		while( np > 0 )
		{
			std::cout << f << ' ';
			np -= f;
			check += f;
			i++;
		}
		assert( n == check );
		assert( fi <= i && i <= ci );
		std::cout << "-- " << n << " = " << check << "; " << fi << "<=" << i << "<=" << ci << std::endl;
		if( n!=check || fi>i || i>ci )
			std::cout << "Fuckup at k=" << n << "!" << std::endl;
	}
}


#include "FArray.h"

using namespace auto_array;

#ifndef NDEBUG

#include <memory>

#define _CUBE_BUFFERS

#include "Sort/Sort.h"
#include "Sort/PointerBMT.h"
#include "Sort/ImplicitBMT.h"
#include "Sort/LazyMerger.h"
#include "Sort/HeapMerger.h"
#include "Sort/FStream.h"

#include "Sort/funnel/merge_tree.h"
#include "Sort/funnel/merge_tree_types.h"

template<class T>
class sorted_stream
{
public:
	class iterator : public std::iterator<std::output_iterator_tag, T, int>
	{
		friend class sorted_stream<T>;
	public:
		inline iterator& operator++()
		{
			if( cnt )
			{
				if( s->prev > s->current )
					std::cerr << "Elements " << static_cast<int>(cnt) << " and " << static_cast<int>(cnt+1) << " out of order: " << s->prev << '>' << s->current << std::endl;
				s->prev = s->current;
				s->chksum += s->current;
			}
			else
			{
				s->prev = s->current;
				s->chksum = s->current;
			}
			++cnt;
			return *this;
		}
		inline T& operator*()
			{ return s->current; }
		inline const T& operator*() const
			{ return s->current; }
		inline bool operator==(const iterator& rhs) const
			{ return cnt == rhs.cnt; }
		inline bool operator!=(const iterator& rhs) const
			{ return cnt != rhs.cnt; }
		inline int operator-(const iterator& rhs) const
			{ return static_cast<int>(cnt-rhs.cnt); }
	protected:
		iterator(sorted_stream *s) : s(s), cnt(0) { }
		iterator(sorted_stream *s, size_t c) : s(s), cnt(c) { }
	private:
		sorted_stream *s;
		size_t cnt;
	};

	sorted_stream(size_t s) : size(s) { }
	iterator begin()
		{ return iterator(this); }
	iterator end()
		{ return iterator(this, size); }
	inline T checksum() const
		{ return chksum; }
protected:
	T prev;
	T current;
	T chksum;
	size_t size;
};

using namespace funnel;

class Refill42 { public: template<class Merger> bool operator()(const Merger*, int* b, int* e)
{
	for( int* p=b; p!=e; p++ )
		*p = 42;
	return false;
} };

template<class Merger>
void check_heap_merger(const char* v, bool balanced = false)
{
	order_t k = static_cast<order_t>((rand() & 63) + Merger::order + 1);
	size_t N = k*k*k/2;
	size_t ss = k*k;
	if( balanced )
		k = pow_of_order<Merger::order>(logc<Merger::order>(k));
	N = N+(size_t)((float)rand()/RAND_MAX*N);
	std::cout << "Checking " << k << "-merger (" << v << '-' << Merger::order << " variant)..." << std::flush;
	typedef typename Merger::value_type T;
	typedef typename Merger::stream stream;
	T *input = new T[N];
	T chksum = RandArray::InitArray<T*,T>(input,input+N);
	T *p;
	std::cout << '.' << std::flush;
	std::allocator<T> alloc;
	{
		Merger merge_tree(k,alloc);
		std::cout << '.' << std::flush;
		for( p=input; p+ss<=input+N; p+=ss )
			std::sort(p, p+ss);
		if( p < input+N )
			std::sort(p, input+N);
		std::cout << '.' << std::flush;
		for( int i=0; i<3; ++i )
		{
			sorted_stream<T> output(N);
			for( p=input; p+ss<=input+N; p+=ss)
				merge_tree.add_stream(stream(p,p+ss));
			if( p < input+N )
				merge_tree.add_stream(stream(p,input+N));
			typename sorted_stream<T>::iterator e = merge_tree(output.begin(), output.end());
			assert( e == output.end() );
			assert( output.checksum() == chksum );
			merge_tree.reset();
		}
	}
	{
		sorted_stream<T> output(N);
		typename Merger::template rebind<Refill42>::other merge_tree42(k,alloc);
		for( p=input; p+ss<=input+N; p+=ss)
			merge_tree42.add_stream(stream(p,p+ss));
		if( p < input+N )
			merge_tree42.add_stream(stream(p,input+N));
		typename sorted_stream<T>::iterator e = merge_tree42(output.begin(), output.end());
		assert( e == output.end() );
		assert( output.checksum() == chksum );
		for( p=input; p<input+N; ++p )
			assert( *p == 42 );
		std::cout << '.' << std::endl;
	}
	delete[] input;
}

template<class Merger>
void check_stack_merger(const char* v, bool balanced = false)
{
	funnel::order_t k = static_cast<order_t>((rand() & 63) + Merger::order + 1);
	size_t N = k*k*k/2;
	size_t ss = k*k;
	if( balanced )
		k = pow_of_order<Merger::order>(logc<Merger::order>(k));
	N = N+(size_t)((float)rand()/RAND_MAX*N);
	std::cout << "Checking " << k << "-merger (" << v << '-' << Merger::order << " variant)..." << std::flush;
	typedef typename Merger::value_type T;
	typedef typename Merger::stream stream;
	T *input = new T[N];
	T chksum = RandArray::InitArray<T*,T>(input,input+N);
	T *p;
	std::cout << '.' << std::flush;
	stack_allocator<T> alloc(Merger::size_of(k));
	{
		Merger merge_tree(k,alloc);
		std::cout << '.' << std::flush;
		for( p=input; p+ss<=input+N; p+=ss )
			std::sort(p, p+ss);
		if( p < input+N )
			std::sort(p, input+N);
		std::cout << '.' << std::flush;
		for( int i=0; i<3; ++i )
		{
			sorted_stream<T> output(N);
			for( p=input; p+ss<=input+N; p+=ss)
				merge_tree.add_stream(stream(p,p+ss));
			if( p < input+N )
				merge_tree.add_stream(stream(p,input+N));
			typename sorted_stream<T>::iterator e = merge_tree(output.begin(), output.end());
			assert( e == output.end() );
			assert( output.checksum() == chksum );
			merge_tree.reset();
		}
	}
	{
		sorted_stream<T> output(N);
		typename Merger::template rebind<Refill42>::other merge_tree42(k,alloc);
		for( p=input; p+ss<=input+N; p+=ss)
			merge_tree42.add_stream(stream(p,p+ss));
		if( p < input+N )
			merge_tree42.add_stream(stream(p,input+N));
		typename sorted_stream<T>::iterator e = merge_tree42(output.begin(), output.end());
		assert( e == output.end() );
		assert( output.checksum() == chksum );
		for( p=input; p<input+N; ++p )
			assert( *p == 42 );
		std::cout << ". free: " << (unsigned int)alloc.max_size() << std::endl;
	}
	delete[] input;
}

template<int order>
void checkMergers()
{
//	int aA[3] = { 87, 503, 512 };
//	int aB[13] = { 61, 154, 170, 275, 426, 509, 612, 653, 677, 703, 765, 897, 908 };
/*	for( ;; )
	{
		int aA[1024]; RandArray::InitArray<int*,int>(aA,aA+1024); std::sort(aA,aA+1024);
		int aB[1024]; RandArray::InitArray<int*,int>(aB,aB+1024); std::sort(aB,aB+1024);
		int aout[16];
		int *out = aout;
		buffer<int*> A1(aB,aB+1024);
		buffer<int*> B1(aA,aA+1024);
		buffer<int*> A2(aB,aB+1024);
		buffer<int*> B2(aA,aA+1024);
		hl_merger<2,buffer<int*>,int> m1;
		two_merger<2,buffer<int*>,int> m2;
		m1.add_stream(&A1,0);
		m1.add_stream(&B1,1);
		m2.add_stream(&A2,0);
		m2.add_stream(&B2,1);
		sorted_stream<int> output1(1023);
		sorted_stream<int> output2(1023);
		sorted_stream<int>::iterator b1 = output1.begin();
		sorted_stream<int>::iterator b2 = output2.begin();
		int t1 = m1(b1,output1.end(),2);
		int t2 = m2(b2,output2.end(),2);
		assert( t1 == t2 );
		assert( b1 == b2 );
		assert( A1.begin() == A2.begin() );
		assert( B1.begin() == B2.begin() );
		std::cout << " done" << std::endl;
	}
*/
	if( order == 2 )
	{
		check_heap_merger <typename merge_tree_types<int,order,buffer<int*>,simple_merger<2,buffer<int*>,int> >::rec_heap_mveb>("simple_merger");
		check_heap_merger <typename merge_tree_types<int,order,buffer<int*>,hl_merger<2,buffer<int*>,int> >::rec_heap_mveb>("hl_merger");
		check_heap_merger <typename merge_tree_types<int,order,buffer<int*>,two_merger<2,buffer<int*>,int> >::rec_heap_mveb>("two_merger");
	}
/*	check_heap_merger <typename merge_tree_types<int,order,buffer<int*>,for_merger<order,buffer<int*>,int> >::rec_heap_mveb>("for_merger");
	check_heap_merger <typename merge_tree_types<int,order,buffer<int*>,for_val_merger<order,buffer<int*>,int> >::rec_heap_mveb>("for_val_merger");
	check_heap_merger <typename merge_tree_types<int,order,buffer<int*>,loser_merger<order,buffer<int*>,int> >::rec_heap_mveb>("loser_merger");
*/
/*	check_heap_merger <typename merge_tree_types<int,order,buffer<int*> >::rec_heap_mveb>("rec_heap_mveb");
	check_stack_merger<typename merge_tree_types<int,order,buffer<int*> >::rec_stack_mveb>("rec_stack_mveb");
	check_heap_merger <typename merge_tree_types<int,order,buffer<int*> >::rec_heap_veb>("rec_heap_veb");
	check_stack_merger<typename merge_tree_types<int,order,buffer<int*> >::rec_stack_veb>("rec_stack_veb");
	check_heap_merger <typename merge_tree_types<int,order,buffer<int*> >::rec_heap_mdf>("rec_heap_mdf");
	check_stack_merger<typename merge_tree_types<int,order,buffer<int*> >::rec_stack_mdf>("rec_stack_mdf");
	check_heap_merger <typename merge_tree_types<int,order,buffer<int*> >::rec_heap_df>("rec_heap_df");
	check_stack_merger<typename merge_tree_types<int,order,buffer<int*> >::rec_stack_df>("rec_stack_df");
	check_heap_merger <typename merge_tree_types<int,order,buffer<int*> >::rec_heap_mbf>("rec_heap_mbf");
	check_stack_merger<typename merge_tree_types<int,order,buffer<int*> >::rec_stack_mbf>("rec_stack_mbf");
	check_heap_merger <typename merge_tree_types<int,order,buffer<int*> >::rec_heap_bf>("rec_heap_bf");
	check_stack_merger<typename merge_tree_types<int,order,buffer<int*> >::rec_stack_bf>("rec_stack_bf");

	check_heap_merger <typename merge_tree_types<int,order,buffer<int*> >::pb_heap_mveb>("pb_heap_mveb");
	check_stack_merger<typename merge_tree_types<int,order,buffer<int*> >::pb_stack_mveb>("pb_stack_mveb");
*/	//check_stack_merger<typename merge_tree_types<int,order,buffer<int*> >::impl_mveb>("impl_mveb", true);
/*	check_heap_merger <typename merge_tree_types<int,order,buffer<int*> >::pb_heap_veb>("pb_heap_veb");
	check_stack_merger<typename merge_tree_types<int,order,buffer<int*> >::pb_stack_veb>("pb_stack_veb");
*/	//check_stack_merger<typename merge_tree_types<int,order,buffer<int*> >::impl_veb>("impl_veb");
/*	check_heap_merger <typename merge_tree_types<int,order,buffer<int*> >::pb_heap_mdf>("pb_heap_mdf");
	check_stack_merger<typename merge_tree_types<int,order,buffer<int*> >::pb_stack_mdf>("pb_stack_mdf");
	check_heap_merger <typename merge_tree_types<int,order,buffer<int*> >::pb_heap_df>("pb_heap_df");
	check_stack_merger<typename merge_tree_types<int,order,buffer<int*> >::pb_stack_df>("pb_stack_df");
	check_stack_merger<typename merge_tree_types<int,order,buffer<int*> >::impl_df>("impl_df");
	check_heap_merger <typename merge_tree_types<int,order,buffer<int*> >::pb_heap_mbf>("pb_heap_mbf");
	check_stack_merger<typename merge_tree_types<int,order,buffer<int*> >::pb_stack_mbf>("pb_stack_mbf");
	check_heap_merger <typename merge_tree_types<int,order,buffer<int*> >::pb_heap_bf>("pb_heap_bf");
	check_stack_merger<typename merge_tree_types<int,order,buffer<int*> >::pb_stack_bf>("pb_stack_bf");
	check_stack_merger<typename merge_tree_types<int,order,buffer<int*> >::impl_bf>("impl_bf");
*/
//	check_heap_merger <LazyMerger<PBBinMergeTree<int>, InputStream<int*> > >("pb_lazy", true);
//	check_heap_merger <LazyMerger<ImplBinMergeTree<int>, InputStream<int*> > >("impl_lazy", true);
//	check_heap_merger <HeapMerger<int, 20000000, 3500, InputStream<int*> > > ("bin_heap");
}

void checkMerger()
{
	for( size_t SIZE=4; SIZE<12000; SIZE++ )
	{
		checkMergers<2>();
		checkMergers<3>();
		checkMergers<4>();
		checkMergers<5>();
		checkMergers<7>();
	}
}

using namespace iosort;

void checkQSort()
{
	for( size_t SIZE=4; SIZE<12000; SIZE++ )
	{
		unsigned long sum = 0;
//		SafeArray<int> src(SIZE);
		MultiMapArray<int> src("C:\\Documents and Settings\\kv\\Local Settings\\Temp\\checkQSort.img", MAP_CREATE | MAP_READ | MAP_WRITE | MAP_TEMPORARY, SIZE);
		for( size_t i=0; i<SIZE; i++ )
		{
			src[i] = rand();
			sum += src[i];
		}

		std::cout << "Sorting " << (int)SIZE << " elements..." << std::flush;
		//qsort(src, SIZE, sizeof(int), __qsort_comp);
		//QuickSort<int>(src, SIZE);
		std::sort(src.begin(), src.end(SIZE));

		Verify<int> dest(src[0]);
		for( size_t i=0; i<SIZE; i++ )
			dest << src[i];
		assert( dest.insertionCount() == SIZE );
		assert( dest.checkSum() == sum );
		std::cout << " done" << std::endl;
	}
}

void checkFSort()
{
	typedef SafeArray<int> Array;
//	typedef HeapMerger<int, 200, 16, InputStream<Array::iterator>, Refill42> Merger;
//	typedef LazyMerger<ImplBinMergeTree<int>, InputStream<Array::iterator>, Refill42> Merger;
	typedef funnel::merge_tree_types<int,5,funnel::buffer<Array::iterator> >::pb_stack_mveb Merger;
	typedef funnel::splitter<5> Splitter;
	for( size_t SIZE=29; SIZE<19000; SIZE++ )
	{
		unsigned long sum = 0;
		Array src(SIZE);
		Array dest(SIZE);
		for( size_t i=0; i<SIZE; i++ )
		{
			src[i] = rand();
			sum += src[i];
		}
		std::cout << "Sorting " << (int)SIZE << " elements..." << std::flush;
		merge_sort<Merger,Splitter>(src.begin(), src.end(SIZE), dest.begin());
		Verify<int> vrf1(dest[0]);
		for( size_t i=0; i<SIZE; i++ )
			vrf1 << dest[i];
//		for( size_t i=0; i<SIZE; i++ )
//			assert(src[i] == 42);
		assert( vrf1.insertionCount() == SIZE );
		assert( vrf1.checkSum() == sum );

		std::cout << " and back again..." << std::flush;
		merge_sort<Merger,Splitter>(dest.begin(), dest.end(SIZE), src.begin());
		Verify<int> vrf2(src[0]);
		for( size_t i=0; i<SIZE; i++ )
			vrf2 << src[i];
		assert( vrf2.insertionCount() == SIZE );
		assert( vrf2.checkSum() == sum );

		std::cout << " done" << std::endl;
	}
}

void checkIMSort()
{
	typedef SafeArray<int> Array;
//	typedef HeapMerger<int, 200, 16, InputStream<Array::iterator> > Merger;
//	typedef LazyMerger<PBBinMergeTree<int>, InputStream<Array::iterator> > Merger;
	typedef funnel::merge_tree_types<int,5,funnel::buffer<Array::iterator> >::pb_stack_mveb Merger;
	typedef funnel::splitter<5> Splitter;
	for( size_t SIZE=2; SIZE<19000; SIZE++ )
	{
		unsigned long sum = 0;
		Array src(SIZE);
		for( size_t i=0; i<SIZE; i++ )
		{
			src[i] = rand();
			sum += src[i];
		}
		std::cout << "Sorting " << (int)SIZE << " elements..." << std::flush;
//		if( SIZE == 1317 )
//			std::cout << "Sorting " << (int)SIZE << " elements..." << std::flush;
//		InplaceMergeSort<Merger>(src.begin(), src.end(SIZE));
		inplace_merge_sort<Merger,Splitter>(src.begin(), src.end(SIZE));
		Verify<int> vrf1(src[0]);
		for( size_t i=0; i<SIZE; i++ )
			vrf1 << src[i];
		assert( vrf1.insertionCount() == SIZE );
		assert( vrf1.checkSum() == sum );

		std::cout << " done" << std::endl;
	}
}

#else // NDEBUG

void checkMerger()
{
}
void checkQSort()
{
}
void checkFSort()
{
}
void checkIMSort()
{
}

#endif // NDEBUG

#include <functional>
#include <algorithm>
#include "Benchmark.h"

template<typename _Tp, typename _Compare>
inline const _Tp& median(const _Tp& __a, const _Tp& __b, const _Tp& __c, _Compare __comp)
{
	if( __comp(__a, __b) )
		if( __comp(__b, __c) )
			return __b;
		else if( __comp(__a, __c) )
			return __c;
		else
			return __a;
	else if( __comp(__a, __c) )
		return __a;
	else if( __comp(__b, __c) )
		return __c;
	else
		return __b;
}

template<typename _RandomAccessIter, typename _Tp, typename _Compare>
inline _RandomAccessIter thin_unguarded_partition_aux(_RandomAccessIter __first, _RandomAccessIter __last, _Tp __pivot, _Compare __comp)
{
	for( ;; )
	{
		while( __comp(*__first, __pivot) )
			++__first;
		--__last;
		while( __comp(__pivot, *__last) )
			--__last;
		if(!( __first < __last ))
			return __first;
		std::iter_swap(__first, __last);
		++__first;
	}
}

template<typename _RandomAccessIter, typename _Compare>
_RandomAccessIter thin_unguarded_partition(_RandomAccessIter __first, _RandomAccessIter __last, _Compare __comp)
{
	return thin_unguarded_partition_aux(__first, __last, median(*__first, *(__first+(__last-__first)/2), *(__last-1), __comp), __comp);
}

template<class _RanIt, class _Pr>
inline void Med3(_RanIt _First, _RanIt _Mid, _RanIt _Last, _Pr _Pred)
{
	if( _Pred(*_Mid, *_First) )
		std::iter_swap(_Mid, _First);
	if( _Pred(*_Last, *_Mid) )
		std::iter_swap(_Last, _Mid);
	if( _Pred(*_Mid, *_First) )
		std::iter_swap(_Mid, _First);
}

template<class _RanIt, class _Pr>
std::pair<_RanIt, _RanIt> fat_unguarded_partition(_RanIt _First, _RanIt _Last, _Pr _Pred)
{
	_RanIt _Mid = _First+(_Last-_First)/2;
	Med3(_First, _Mid, _Last-1, _Pred);
	_RanIt _Pfirst = _Mid;
	_RanIt _Plast = _Pfirst+1;

	while( _First < _Pfirst && !_Pred(*(_Pfirst-1), *_Pfirst) && !_Pred(*_Pfirst, *(_Pfirst-1)) )
		--_Pfirst;
	while( _Plast < _Last && !_Pred(*_Plast, *_Pfirst) && !_Pred(*_Pfirst, *_Plast) )
		++_Plast;

	_RanIt _Gfirst = _Plast;
	_RanIt _Glast = _Pfirst;

	for( ;; )
	{	// partition
		for( ; _Gfirst < _Last; ++_Gfirst )
			if( _Pred(*_Pfirst, *_Gfirst) )
				;
			else if( _Pred(*_Gfirst, *_Pfirst) )
				break;
			else
				std::iter_swap(_Plast++, _Gfirst);
		for( ; _First < _Glast; --_Glast )
			if( _Pred(*(_Glast-1), *_Pfirst) )
				;
			else if( _Pred(*_Pfirst, *(_Glast-1)) )
				break;
			else
				std::iter_swap(--_Pfirst, _Glast - 1);
		if( _Glast == _First && _Gfirst == _Last )
			return std::pair<_RanIt, _RanIt>(_Pfirst, _Plast);

		if( _Glast == _First )
		{	// no room at bottom, rotate pivot upward
			if( _Plast != _Gfirst )
				std::iter_swap(_Pfirst, _Plast);
			++_Plast;
			std::iter_swap(_Pfirst++, _Gfirst++);
		}
		else if( _Gfirst == _Last )
		{	// no room at top, rotate pivot downward
			if( --_Glast != --_Pfirst )
				std::iter_swap(_Glast, _Pfirst);
			std::iter_swap(_Pfirst, --_Plast);
		}
		else
			std::iter_swap(_Gfirst++, --_Glast);
	}
}

using namespace cmd_switch;

template<class It>
std::pair<double,unsigned long> ScanThis(scan sce, It begin, It end, size_t order)
{
	typedef std::pair<It,It> Stream;
	static It min = begin;
	Stream *streams = NULL;
	It i, j;
	BeginPFCounter();
	switch( sce )
	{
	case f:
		std::cout << "\t\t<algorithm>Forward scan</algorithm>" << std::endl;
		StartTimer();
		for( i=begin; i<end; i++ )
			if( *i < *min )
				min = i;
		break;
	case b:
		std::cout << "\t\t<algorithm>Backward scan</algorithm>" << std::endl;
		StartTimer();
		for( i=end-1; i>=begin; i-- )
			if( *i < *min )
				min = i;
		break;
	case o:
		std::cout << "\t\t<algorithm>Outward scan</algorithm>" << std::endl;
		StartTimer();
		for( i=begin+(end-begin)/2, j=i+1; j<end; i--, j++ )
		{
			if( *i < *min )
				min = i;
			if( *j < *min )
				min = j;
		}
		if( i == begin )
			if( *i < *min )
				min = i;
		break;
	case in:
		std::cout << "\t\t<algorithm>Inward scan</algorithm>" << std::endl;
		StartTimer();
		for( i=begin, j=end-1; i<j; i++, j-- )
		{
			if( *i < *min )
				min = i;
			if( *j < *min )
				min = j;
		}
		if( i == j )
			if( *i < *min )
				min = i;
		break;
	case sc:
		{
			std::cout << "\t\t<algorithm>Cyclic stream scan</algorithm>" << std::endl;
			std::cout << "\t\t<order>" << (unsigned int)order << "</order>" << std::endl;
			streams = new Stream[order];
			size_t run_size = (end-begin)/order;
			size_t rem = (size_t)((end-begin) % order);
			run_size += rem/order;
			size_t big_runs = rem % order;
			It last = begin;
			{
				size_t i;
				for( i=0; i!=order-big_runs; i++, last+=run_size )
					streams[i] = Stream(last, last+run_size);
				run_size++;
				for( ; i!=order; i++, last+=run_size )
					streams[i] = Stream(last, last+run_size);
			}

			StartTimer();
			Stream *p=streams;
			while( p->first != p->second )
			{
				i = p->first;
				p->first++;
				if( *i < *min )
					min = i;
				p++;
				if( p == streams+order )
					p = streams;
			}
		}
		break;
	case sr:
		std::cout << "\t\t<algorithm>Random stream scan</algorithm>" << std::endl;
		break;
	default:
		assert( false );
	}
	double t = StopTimer();
	delete[] streams;
	return std::pair<double, unsigned long>(t, EndPFCounter());
}

void scanTest(int argc, char *argv[], cmd_switch::scan sc, _int64_t n, size_t streams, const char *file)
{
	std::cout << "\t<run buildtime=\"" << buildtime() << "\" compiler=\"" << compilername() << "\" commandline=\"";
	std::cout << argv[0];
	for( int i=1; i<argc; i++ )
		std::cout << ' ' << argv[i];
	std::cout << "\" execstatus=$>" << std::endl;
	std::cout << "\t\t<size>" << n << "</size>" << std::endl;
	if( sc == no )
	{
		assert( file );
		MapArray<long> src(file, MAP_CREATE | MAP_READ | MAP_WRITE, (MapArray<long>::size_type)n);
		RandArray::InitArray<long*,long>(src.begin(), src.end((MapArray<long>::size_type)n));
	}
	else
	{
		std::pair<double, unsigned long> timing;
		if( sc == pr || sc == pf || sc == pb || sc == Pr || sc == Pf || sc == Pb )
		{
			MapArray<long>::size_type N = (MapArray<long>::size_type)n;
			MapArray<long> src(SRC_FILE, MAP_CREATE | MAP_READ | MAP_WRITE | MAP_TEMPORARY, N);
			if( sc == pr || sc == Pr )
				for( MapArray<long>::iterator p=src.begin(); p!=src.end(N); p++ )
					*p = rand();
			else if( sc == pf || sc == Pf )
				for( MapArray<long>::iterator p=src.begin(); p!=src.end(N); p++ )
					*p = (long)(p-src.begin());
			else
				for( MapArray<long>::iterator p=src.begin(); p!=src.end(N); p++ )
					*p = (long)(src.end(N)-p);
			BeginPFCounter();
			if( sc == pr || sc == pf || sc == pb )
			{
				std::cout << "\t\t<algorithm>Thin Unguarded Partition</algorithm>" << std::endl;
				StartTimer();
				thin_unguarded_partition(src.begin(), src.end(N), std::less<long>());
			}
			else
			{
				std::cout << "\t\t<algorithm>Fat Unguarded Partition</algorithm>" << std::endl;
				StartTimer();
				fat_unguarded_partition(src.begin(), src.end(N), std::less<long>());
			}
			timing.first = StopTimer();
			timing.second = EndPFCounter();
		}
		else if( file )
		{
			MapArray<long> src(file, MAP_EXISTING | MAP_READ | MAP_WRITE | MAP_TEMPORARY, (MapArray<long>::size_type)n);
			timing = ScanThis(sc, src.begin(), src.end((MapArray<long>::size_type)n), streams);
		}
		else
		{
			MapArray<long> src(SRC_FILE, MAP_CREATE | MAP_READ | MAP_WRITE | MAP_TEMPORARY, (MapArray<long>::size_type)n);
			RandArray::InitArray<long*,long>(src.begin(), src.end((MapArray<long>::size_type)n));
			timing = ScanThis(sc, src.begin(), src.end((MapArray<long>::size_type)n), streams);
		}
		std::cout << "\t\t<time unit=\"s\">" << timing.first << "</time>" << std::endl
				  << "\t\t<pagefaults>" << timing.second << "</pagefaults>" << std::endl;
	}
	std::cout << "\t</run>" << std::endl;
}
