/*
 * Copyright (C) 2003 Kristoffer Vinther.
 * For contact information, please visit
 * http://www.brics.dk/~kv/thesis/.
 * If you use this code in commercial applications, please notify me.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * Free Software Foundation, Inc.
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <iostream>
#include "Test.h"
#include "cmd.h"

extern void fun();

// Converts a string to a 64bit integer
_int64_t a2i64(const char* a)
{
	_int64_t i = 0;
	for( i=0; *a>='0' && *a<='9'; a++)
		i = 10*i+(*a)-'0';
	return i;
}

// Prints help on usage of the program to stderr and exits with failure status
void exit_usage()
{
	std::cerr << "Usage: sortmonkey [-t{2|3|Q|F}] [-s{q|s|m|i|Q|F[i|e]}] [-a{c|d|m|M|S}] [-n size]" << std::endl
			  << "       sortmonkey [-c{f|b|o|i|p{r|f|b}|P{r|f|b}|s{r|c} order}] [-n size] [-f file]" << std::endl
			  << "       sortmonkey fun" << std::endl
			  << std::endl
			  << "Tests:\t\t2: Split of n in sqrt of n" << std::endl
			  << "\t\t3: Split of n in cubic root of n" << std::endl
			  << "\t\tQ: QuickSort on arrays of different sizes containing random elements" << std::endl
			  << "\t\tF: FunnelSort on arrays of different sizes containing random elements" << std::endl
			  << std::endl
			  << "Sorting algs.:\tq:  C standard library qsort()" << std::endl
			  << "\t\ts:  C++ standard library std::sort()" << std::endl
			  << "\t\tm:  mergesort based on C++ std::merge()" << std::endl
			  << "\t\ti:  inplace mergesort based on C++ std::inplace_merge()" << std::endl
			  << "\t\tMc: I/O optimal mergesort based on C++ std::make_heap() and optimized for L2/Memory" << std::endl
			  << "\t\tMC: I/O optimal inplace mergesort based on C++ std::make_heap() and optimized for L2/Memory" << std::endl
			  << "\t\tMm: I/O optimal mergesort based on C++ std::make_heap() and optimized for Memory/Disk" << std::endl
			  << "\t\tMM: I/O optimal inplace mergesort based on C++ std::make_heap() and optimized for Memory/Disk" << std::endl
			  << "\t\tQ:  custom QuickSort()" << std::endl
			  << "\t\tF:  FunnelSort; funnel variants:" << std::endl
			  << "\t\t       i: implicit" << std::endl
			  << "\t\t       e: explicit" << std::endl
			  << "\t\t       E: inplace explicit" << std::endl
			  << std::endl
			  << "Arrays:\t\tc: C standard array allocated on heap using new" << std::endl
			  << "\t\td: automatic array allocated on heap using new" << std::endl
			  << "\t\tm: memorymapped automatic array" << std::endl
			  << "\t\tM: Paging memorymapped automatic array" << std::endl
			  << "\t\tS: Sequentially memorymapped automatic array" << std::endl
			  << std::endl
			  << "Other:" << std::endl
			  << "\t\tc:   Scan an array" << std::endl
			  << "\t\tf:   Specify file to open/create" << std::endl
			  << "\t\t     Use -f to simply create a file with random data" << std::endl
			  << "\t\t     Use -c to scan an array:" << std::endl
			  << "\t\t       f: forward" << std::endl
			  << "\t\t       b: backward" << std::endl
			  << "\t\t       o: from middle and out" << std::endl
			  << "\t\t       i: from start and end and in" << std::endl
			  << "\t\t       p/P: do a thin/fat partition complete with exchanges on" << std::endl
			  << "\t\t          r: random integers" << std::endl
			  << "\t\t          f: ascendiong integers" << std::endl
			  << "\t\t          b: decending integers" << std::endl
			  << "\t\t       s: split array in order subarrays and scan each of them in" << std::endl
			  << "\t\t          r: random order" << std::endl
			  << "\t\t          c: cyclic order" << std::endl
			  << "\t\t     Use -c and -f to to scan the file specified with -f" << std::endl
			  << "\t\tUse fun to run the funny function." << std::endl;
	exit(EXIT_FAILURE);
}

using namespace cmd_switch;

void parseargs(int argc, char *argv[], cmd_switch::algorithm& alg, cmd_switch::array_type& arr, _uint64_t& n)
{
	scan sce = no;
	char *file = NULL;
	alg = FE;
	arr = c;
	n = 16388616;
	size_t order = 0;
	if( argc == 2 && strcmp("fun", argv[1]) == 0 )
	{
		fun();
		exit(EXIT_SUCCESS);
	}
	for( int i=1; i<argc; i++ )
	{
		if( argv[i][0] == '-' )
		{
			switch( argv[i][1] )
			{
			case 't':
				if( argv[i][2] == 0 )
					exit_usage();
				switch( argv[i][2] )
				{
				case '2':
					checkSqrt();
					break;
				case '3':
					checkCub();
					break;
				case 'Q':
					checkQSort();
					break;
				case 'F':
					checkFSort();
					break;
				case 'I':
					checkIMSort();
					break;
				case 'M':
					checkMerger();
					break;
				default:
					exit_usage();
				}
				exit(EXIT_SUCCESS);
			case 'c':
				switch( argv[i][2] )
				{
				case 'f':
					sce = f;
					break;
				case 'b':
					sce = b;
					break;
				case 'o':
					sce = o;
					break;
				case 'i':
					sce = in;
					break;
				case 'p':
					switch( argv[i][3] )
					{
					case 'r':
						sce = pr;
						break;
					case 'f':
						sce = pf;
						break;
					case 'b':
						sce = pb;
						break;
					default:
						if( argv[i][3] )
							exit_usage();
						sce = pr;
					}
					break;
				case 'P':
					switch( argv[i][3] )
					{
					case 'r':
						sce = Pr;
						break;
					case 'f':
						sce = Pf;
						break;
					case 'b':
						sce = Pb;
						break;
					default:
						if( argv[i][3] )
							exit_usage();
						sce = Pr;
					}
					break;
				case 's':
					switch( argv[i][3] )
					{
					case 'r':
						sce = sr;
						break;
					case 'c':
						sce = sc;
						break;
					default:
						if( argv[i][3] )
							exit_usage();
						sce = sc;
					}
					i++;
					if( i == argc )
						exit_usage();
					order = atoi(argv[i]);
					break;
				case '\0':
					sce = f;
					break;
				default:
					exit_usage();
				}
				break;
			case 'f':
				if( argv[i][2] )
					exit_usage();
				i++;
				if( i == argc )
					exit_usage();
				file = argv[i];
				break;
			case 'n':
				i++;
				if( i == argc )
					exit_usage();
				n = a2i64(argv[i]);
				break;
			case 'a':
				if( argv[i][2] == 0 )
					exit_usage();
				switch( argv[i][2] )
				{
				case 'c':
					arr = c;
					break;
				case 'd':
					arr = d;
					break;
				case 'm':
					arr = m;
					break;
				case 'M':
					arr = M;
					break;
				case 'S':
					arr = S;
					break;
				default:
					exit_usage();
				}
				break;
			case 's':
				if( argv[i][2] == 0 )
					exit_usage();
				switch( argv[i][2] )
				{
				case 'q':
					alg = q;
					break;
				case 's':
					alg = s;
					break;
				case 'h':
					alg = h;
					break;
				case 'm':
					alg = mg;
					break;
				case 'i':
					alg = img;
					break;
				case 'M':
					switch( argv[i][3] )
					{
					case '\0':
						alg = Mg;
						break;
					case 'c':
						alg = Mg;
						break;
					case 'C':
						alg = MIg;
						break;
					case 'm':
						alg = MMg;
						break;
					case 'M':
						alg = MMIg;
						break;
					default:
						exit_usage();
					}
					break;
				case 'Q':
					alg = Q;
					break;
				case 'F':
					switch( argv[i][3] )
					{
					case 'i':
                        alg = FI;
						break;
					case 'e':
                        alg = FE;
						break;
					case 'E':
                        alg = FEI;
						break;
					case 'I':
                        alg = FII;
						break;
					default:
						if( argv[i][3] )
							exit_usage();
                        alg = FE;
					}
					break;
				default:
					exit_usage();
				}
				break;
			default:
				exit_usage();
			}
		}
		else
			exit_usage();
	}
	if( sce != no || file )
	{
		scanTest(argc, argv, sce, n, order, file);
		exit(EXIT_SUCCESS);
	}
	if( file && sce == no )
		exit_usage();
}
