/*
 * Copyright (C) 2003 Kristoffer Vinther.
 * For contact information, please visit
 * http://www.brics.dk/~kv/thesis/.
 * If you use this code in commercial applications, please notify me.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * Free Software Foundation, Inc.
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __CMD_H_INCLUDED_
#define __CMD_H_INCLUDED_

#include "Sort/int64.h"

#ifdef WIN32
#define SRC_FILE "C:\\Documents and Settings\\kv\\Local Settings\\Temp\\FunnelUnSorted.img"
#define DST_FILE "C:\\Documents and Settings\\kv\\Local Settings\\Temp\\FunnelSorted.img"
#else // WIN32

#ifndef IMAGEDIR
#define IMAGEDIR "/tmp/"
#endif // IMAGEDIR

#define SRC_FILE IMAGEDIR "FunnelUnSorted.img"
#define DST_FILE IMAGEDIR "FunnelSorted.img"
#endif // WIN32

_int64_t a2i64(const char* a);

namespace cmd_switch
{
	enum algorithm { q, s, h, mg, img, Mg, MMg, MIg, MMIg, Q, FE, FI, FEI, FII };
	enum array_type { c, d, m, M, S };
	enum scan { no, f, b, in, o, pr, pf, pb, Pr, Pf, Pb, sc, sr };
}

void exit_usage();
void parseargs(int argc, char** argv, cmd_switch::algorithm& alg, cmd_switch::array_type& arr, _uint64_t& n);

template<class T>
class Counting
{
public:
	Counting() { }
	Counting(const T& e) : e(e) { }
	Counting(const Counting<T>& src) : e(src.e)
		{ moves++; }

	inline Counting<T> operator=(const T& rhs)
	{
		e = rhs;
		return *this;
	}
	inline Counting<T> operator=(const Counting<T>& rhs)
	{
		e = rhs.e;
		moves++;
		return *this;
	}
	inline bool operator<(const Counting<T>& rhs) const
	{
		comparisons++;
		return e < rhs.e;
	}
	inline bool operator<=(const Counting<T>& rhs) const
	{
		comparisons++;
		return e <= rhs.e;
	}

	static inline _uint64_t getMoves()
		{ return moves; }
	static inline _uint64_t getComps()
		{ return comparisons; }
	static inline void reset()
		{ moves = 0; comparisons = 0; }

private:
	static _uint64_t moves;
	static _uint64_t comparisons;
	T e;
};

template<class T>
_uint64_t Counting<T>::moves;
template<class T>
_uint64_t Counting<T>::comparisons;

#include <stdlib.h>
#include <time.h>

class RandArray
{
public:
	template<class FwdIt, class T>
	static T InitArray(FwdIt begin, FwdIt end)
	{
		if( begin == end )
			return T();
#ifdef WIN32
#ifndef _DEBUG
		srand((unsigned)time(NULL));
#endif// _DEBUG
		*begin = rand()^(rand()<<8)^(rand()<<16)^(rand()<<24);
		T chksum = *begin;
		for( begin++; begin!=end; begin++ )
		{
			*begin = rand()^(rand()<<8)^(rand()<<16)^(rand()<<24);
			chksum += *begin;
		}
#else // WIN32
#ifndef _DEBUG
		srand48((long int)time(NULL));
#endif// _DEBUG
		*begin = lrand48();
		T chksum = *begin;
		for( begin++; begin!=end; begin++ )
		{
			*begin = lrand48();
			chksum += *begin;
		}
#endif // WIN32
		return  chksum;
	}
};

#endif // __CMD_H_INCLUDED_

